/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <math.h>
#include "aodv6.h"
#include "rt_table.h"
#include "messages.h"
#include "tqueue.h"
#include "pktqueue.h"
#include "util.h"
#include "config.h"

int helloInterval = HELLO_INTERVAL;
int activeRouteTimeout = ACTIVE_ROUTE_TIMEOUT;
int deletePeriod = DELETE_PERIOD;
int myRouteTimeout = MY_ROUTE_TIMEOUT;
int reverseRouteLife = REV_ROUTE_LIFE;
              
struct resend_data {
	unsigned long timeout;
	unsigned long key;
	int hoplimit;
	int resend;
	struct aodv6_rreq *rreq;
	size_t len;
};

struct rreq_dupe {
	struct rreq_dupe *next;
	struct rreq_dupe *prev;
	u_int32_t rreq_id;
	struct in6_addr originator;
};

struct rreq_dupe dupelist = { &dupelist, &dupelist, 0 };

extern FILE *ulogd;
extern int sockfd;
extern int xdbg;
extern int xtr;

u_int32_t rreq_id;

extern void rt_dump(void);

u_int32_t get_rreq_id(void)
{
	return ++rreq_id;
}

void printHelloParams( void )
{
	fprintf(stdout, "\tNODE_TRAVERSAL_TIME    = %d\n",NODE_TRAVERSAL_TIME);
	fprintf(stdout, "\tNET_TRAVERSAL_TIME     = %d\n",NET_TRAVERSAL_TIME);
	fprintf(stdout, "\tNET_DIAMETER           = %d\n",NET_DIAMETER);
	fprintf(stdout, "\tPATH_DISCOVERY_TIME    = %d\n",PATH_DISCOVERY_TIME);
	fprintf(stdout, "\thelloInterval          = %d [msec]\n", helloInterval);
	fprintf(stdout, "\tactiveRouteTimeout     = %d [msec]\n", activeRouteTimeout);
	fprintf(stdout, "\tdeletePeriod           = %d [msec]\n", deletePeriod);
	fprintf(stdout, "\tmyRouteTimeout         = %d [msec]\n", myRouteTimeout);
	fprintf(stdout, "\treverseRouteLife       = %d [msec]\n", reverseRouteLife);
}
	
void resetHelloParams(int new)
{
	int path =  PATH_DISCOVERY_TIME;
	int art = ACTIVE_ROUTE_TIMEOUT;
	int axh = ALLOWED_HELLO_LOSS * new;
	int b = 0;
	int c = 0;
	
	helloInterval = new;
	activeRouteTimeout =  (art < axh)? axh : art;
	deletePeriod = (K * activeRouteTimeout);
	b = (activeRouteTimeout < path)? path : activeRouteTimeout; 
	c = (2 * NET_TRAVERSAL_TIME);
	reverseRouteLife = ( activeRouteTimeout < c) ? c : activeRouteTimeout ;
	
	//Next gives a very bad performance this way :(
	//myRouteTimeout = 2*b; 
	myRouteTimeout = reverseRouteLife;

	printHelloParams();

}

int get_host_addr(struct in6_addr *addr)
{
	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	if (addr == NULL) return -1;
	memcpy(addr, &Config.HostAddrGlobal, sizeof(struct in6_addr));
	return 0;
}

int get_own_addr(struct in6_addr *addr)
{
	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (addr == NULL) 
		return -1;

	memcpy(addr, &Config.NodeAddrGlobal, sizeof(struct in6_addr));
	return 0;
}

/**
 * rreq_buffer_dupe - buffer RREQ ID and Originator
 * @dupe: RREQ ID and Originator
 *
 * Buffers Originator address and RREQ ID sent in a RREQ for
 * %PATH_TRAVERSAL_TIME milliseconds.
 **/
static void rreq_buffer_dupe(struct rreq_dupe *dupe)
{
	struct rreq_dupe *last = &dupelist;
	if (dupe == NULL) return;

	/* This is original code but I don't seem to work 
	dupe->prev = &dupelist;
	dupe->next = dupelist.next;
	dupelist.next->prev = dupe;
	dupelist.next = dupe;
	*/

	if (xdbg) 
	{
		fprintf(stdout,"INFO: Adding dupe with ID %u for ", dupe->rreq_id); 
		print_in6addr(stdout,&dupe->originator);
	}

	/* add dupe at the end of dupelist */
	while (last->next != &dupelist) 
		last = last->next;
	/* got last dupe in dupelist */
	last->next = dupe;
	dupe->prev = last;
	dupe->next = &dupelist;
}

/**
 * rreq_check_dupe - check for duplicate
 * @addr: Originator address
 * @rreq_id: RREQ ID
 *
 * Checks if this RREQ ID/Originator pair is already in buffer.  If
 * duplicate exists, returns 1.  Otherwise, returns 0.
 **/
static int rreq_check_dupe(struct in6_addr *addr, u32 rreq_id)
{
	struct rreq_dupe *n;

	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	if (addr == NULL) return 0;

	if (xdbg) 
	{
		fprintf(stdout,"INFO: Checking dupe for "); 
		print_in6addr(stdout,addr);
	}

	if (dupelist.next == &dupelist)
	{
		if (xdbg) fprintf(stdout,"\nWARNING: Dupelist is empty\n");
		return 0;
	}

	for (n = dupelist.next; n != &dupelist; n = n->next) 
	{
		if (rreq_id == n->rreq_id &&
		    memcmp(addr, &n->originator, sizeof(struct in6_addr)) == 0)
		{
			if(xdbg) fprintf(stdout,"\nINFO: Found dupe w/ RREQ_ID %u for ", rreq_id);
			return 1;
		}
	}

	if(xdbg) fprintf(stdout,"\nFound NO dupe\n");
	return 0;
}

/**
 * rreq_no_dupe - expire duplicate entry
 * @p: pointer to expired entry
 *
 * Deletes entry from duplicate buffer.  Called on SIGALRM.
 **/
static void rreq_no_dupe(void *p)
{
	struct rreq_dupe *dupe = (void *)p;

	if (p == NULL) return;

	dupe->next->prev = dupe->prev;
	dupe->prev->next = dupe->next;
	dupe->next = dupe->prev = NULL;
}

/**
 * hello_missed - expire neighbor
 * @data: pointer to unreachable address
 *
 * When %ALLOWED_HELLO_LOSS * %HELLO_INTERVAL milliseconds has passed
 * without receiving a Hello message, this function is called.
 * Destination is invalidated.
 **/
void hello_missed(void *data)
{
	struct in6_addr *dest = (struct in6_addr *)data;
	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	fprintf(stdout,"Hello message missed from IP ");
	print_in6addr(stdout, dest);
	
	rt_del(dest, 128);
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
}

/**
 * aodv6_rrep_send - send Route Reply message
 * @dst: IPv6 header destination address
 * @rrep: RREP message to send
 * @len: length of RREP message
 *
 * Sends a RREP message @rrep to @dst.  Caller must supply length of
 * message @len including any extensions.
 **/
int aodv6_rrep_send(struct in6_addr *dst, struct aodv6_rrep *rrep, size_t len)
{
	struct sockaddr_in6 daddr;
	struct iovec iov[1];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	int cmsglen;
	struct in6_pktinfo pinfo;
	int ret = 0, on = 1;

	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);

	if (dst == NULL) return -ENODATA;
	if (rrep == NULL) return -ENODATA;
	if (len < sizeof(struct aodv6_rrep)) return -EINVAL;

	memset(&daddr, 0, sizeof(struct sockaddr_in6));
	daddr.sin6_family = AF_INET6;
	memcpy(&daddr.sin6_addr, dst, sizeof(struct in6_addr));
	daddr.sin6_port = htons(AODVPORT);

	memset(&pinfo, 0, sizeof(pinfo));
	//get_own_addr(&pinfo.ipi6_addr);
	get_host_addr(&pinfo.ipi6_addr);

	cmsglen = CMSG_SPACE(sizeof(pinfo));
	cmsg = malloc(cmsglen);
	if (cmsg == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}
	cmsg->cmsg_len = CMSG_LEN(sizeof(pinfo));
	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	memcpy(CMSG_DATA(cmsg), &pinfo, sizeof(pinfo));

	iov[0].iov_base = (unsigned char *)rrep;
	iov[0].iov_len = len;

	msg.msg_control = cmsg;
	msg.msg_controllen = cmsglen;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (void *)&daddr;
	msg.msg_namelen = sizeof(daddr);

	setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(int));

	ret = sendmsg(sockfd, &msg, 0);

	if ((ntohs(((u16 *)dst)[0]) != 0xff02) && (ntohs(((u16 *)dst)[7]) != 0x1))
	{
		fprintf(stdout, 
			"INF-RREP: sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x[%x:%x:%x:%x:%x:%x:%x:%x] -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	    	 ret,
			 NIP6ADDR(&rrep->orig), NIP6ADDR(&pinfo.ipi6_addr), 
			 NIP6ADDR(dst));
		if(xdbg) dump_mux(stdout, rrep, MSG_RREP);
	} 

	ulog(L_SEND_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_SEND_INFO, 
			"sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	     	ret, NIP6ADDR(&pinfo.ipi6_addr), NIP6ADDR(dst));
	ulog_if(L_SEND_PACKET) 
	{
		dump_mux(ulogd, rrep, MSG_RREP);
	}
	
	free(cmsg);

	return 0;
}

/**
 * aodv6_rreq_send - send Route Request message
 * @dst: IPv6 header destination address
 * @hoplimit: IPv6 header hoplimit
 * @rreq: RREQ message to send
 * @len: length of RREQ message
 *
 * Sends a RREQ message @rreq to @dst with hops limited to @hoplimit.
 * Caller must supply length of message @len including any extensions.
 **/
int aodv6_rreq_send(struct in6_addr *dst, int hoplimit, 
		     struct aodv6_rreq *rreq, size_t len)
{
	struct sockaddr_in6 daddr;
	struct iovec iov[1];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	int cmsglen;
	struct in6_pktinfo pinfo;
	struct rreq_dupe *dupe;
	int ret = 0, on = 1;

	if(xtr) fprintf(stdout,"==== %s ===\n",__FUNCTION__);
	
	if (rreq == NULL) return -ENODATA;
	if (dst == NULL) return -ENODATA;

	if (len < sizeof(struct aodv6_rreq)) return -EINVAL;

	if (rtnl_rt_exist(&rreq->dest) == 0)
	{
		fprintf(stdout,"Have route, will return!\n");
		//return -1;
		return 0;
	}
	
	dupe = (struct rreq_dupe *)malloc(sizeof(struct rreq_dupe));
	
	if (dupe == NULL) 
	{
		perror(__FUNCTION__);
		fprintf(stdout,"Can't malloc\n");
		return -ENOMEM;
	}

	dupe->rreq_id = rreq_id;
	memcpy(&dupe->originator, dst, sizeof(struct in6_addr));
	
	if(xdbg) fprintf(stdout,"Calling rreq_buffer_dupe()...\n");
	rreq_buffer_dupe(dupe);
	if(xdbg) fprintf(stdout,"Adding alarm...\n");

	/* remove the dupe after PATH_DISCOVERY_TIME 
	 * -------------------------------------------*/
	add_alarm(PATH_DISCOVERY_TIME, ntohl(rreq->rreq_id), 
		  RREQ_DUPE, rreq_no_dupe, dupe);

	memset(&daddr, 0, sizeof(struct sockaddr_in6));
	daddr.sin6_family = AF_INET6;
	memcpy(&daddr.sin6_addr, dst, sizeof(struct in6_addr));
	daddr.sin6_port = htons(AODVPORT);

	/* Set socket and send message
	 * ----------------------------*/
	setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, 
		   &hoplimit, sizeof(int));

	memset(&pinfo, 0, sizeof(pinfo));
	//get_own_addr(&pinfo.ipi6_addr);
	get_host_addr(&pinfo.ipi6_addr);

	cmsglen = CMSG_SPACE(sizeof(pinfo));
	cmsg = malloc(cmsglen);
	if (cmsg == NULL) 
	{
		fprintf(stdout, "ERROR: cmsg==NULL\n");
		perror(__FUNCTION__);
		return -ENOMEM;
	}
	cmsg->cmsg_len = CMSG_LEN(sizeof(pinfo));
	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	memcpy(CMSG_DATA(cmsg), &pinfo, sizeof(pinfo));

	iov[0].iov_base = (unsigned char *)rreq;
	iov[0].iov_len = sizeof(struct aodv6_rreq);

	msg.msg_control = cmsg;
	msg.msg_controllen = cmsglen;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (void *)&daddr;
	msg.msg_namelen = sizeof(daddr);

	setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(int));

	//MI - SO_DONTROUTE or SO_BROADCAST?
	ret = sendmsg(sockfd, &msg, SO_DONTROUTE);
	free(cmsg);
	
	//MI -Should we check if rreq->orig is own AODV address?
	fprintf(stdout, 
			"\nINF-RREQ: sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x[%x:%x:%x:%x:%x:%x:%x:%x] -> %x:%x:%x:%x:%x:%x:%x:%x [hoplim=%d]\n",
	     	ret, NIP6ADDR(&rreq->orig), 
			NIP6ADDR(&pinfo.ipi6_addr), 
			NIP6ADDR(&rreq->dest), 
			hoplimit);
	if(xdbg) dump_mux(stdout,rreq,MSG_RREQ);

	ulog(L_SEND_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_SEND_INFO, "sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	     ret, NIP6ADDR(&pinfo.ipi6_addr), NIP6ADDR(dst));
	ulog_if(L_SEND_PACKET) {
		dump_mux(ulogd, rreq, MSG_RREQ);
	}

	return ret;
}

/**
 * aodv6_rreq_resend - resend Route Request message
 * @data: pointer to &resend_data structure
 *
 * Resends a previously generated and sent RREQ stored in &resend_data
 * structure @data.  This function is called from the task queue.
 **/
void aodv6_rreq_resend(void *data)
{
	struct resend_data *rd;
	struct in6_addr allnodes = IPV6_ALL_MANET_NODES;
	int timeout = 0, ret = 0;

	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (data == NULL) 
	{
		fprintf(stderr, "No data passed in, nothing to resend!");
		return;
	}

	rd = (struct resend_data *)malloc(sizeof(struct resend_data));

	if(rd == NULL)
	{
		fprintf(stderr,"ERROR: Malloc failed!!!\n");
		return;
	}	
	memcpy(rd, data, sizeof(struct resend_data));
	
	if (rtnl_rt_exist(&rd->rreq->dest) == 0) 
	{
		if(xdbg) 
		{
			fprintf(stdout,"rtnl_rt_exist() returned 0: Have route, DO NOT resend rreq ");
			if(&rd->rreq->dest)
				fprintf(stdout," for %x:%x:%x:%x:%x:%x:%x:%x", 
				NIP6ADDR(&rd->rreq->dest));
			fprintf(stdout," Will return\n ");
		}
		free(rd);
		return;
	}

	if (rd->resend >= RREQ_RETRIES) 
	{
		if(xdbg) 
		{
			fprintf(stdout,"No more resends");
			if(&rd->rreq->dest)
				fprintf(stdout," for %x:%x:%x:%x:%x:%x:%x:%x", 
				NIP6ADDR(&rd->rreq->dest));
			fprintf(stdout," Remove packet and return\n ");
		}

		/* timeout, drop all packets for destination */
		/*pktqueue_apply(&rd->rreq->dst, ip6q_drop);*/
		pktqueue_remove(&rd->rreq->dest);
		free(rd);
		return;
	}
	rd->hoplimit += TTL_INCREMENT;
	if (rd->hoplimit > TTL_THRESHOLD) 
	{
		rd->hoplimit = NET_DIAMETER;
		rd->resend++;
	}

	/* Update RREQ ID and Orig seq #*/
	rd->rreq->rreq_id = htonl(get_rreq_id());
	rd->rreq->orig_seq = htonl(inc_get_seqnum());

#ifdef _SECURE_BOOT_
	/* Re-sign the message */
	{
		int ret = sign_msg(rd->rreq, MSG_RREQ);
		
		if(ret<0)
			exit(1);
	}
#endif
	
	//MI - This is logically correct, but if reaches NET size it stops sending requests
	timeout = 2 * rd->hoplimit * NODE_TRAVERSAL_TIME;
	timeout = (timeout > 1000)? 1000: timeout;

	//ulog(L_SEND_INFO, __FUNCTION__ ": resend = %d\n", rd->resend);
	
	ret = aodv6_rreq_send(&allnodes, rd->hoplimit, rd->rreq, rd->len);
	
	if (ret < 0) 
	{
		if(xdbg) fprintf(stderr,"WARNING-%s():%d: rreq_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
		free(rd);
		return;
	}
	else if(&rd->rreq->dest)
		fprintf(stdout,
				"RREQ: Resend #%d %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x [hlim=%d, reqID=%u, seq#=%u]\n",
					rd->resend,	NIP6ADDR(&rd->rreq->dest), 
					NIP6ADDR(&allnodes), rd->hoplimit, 
					ntohl(rd->rreq->rreq_id), 
					ntohl(rd->rreq->orig_seq));	
	//if (xdbg) 
		fprintf(stdout,"RREQ-SEND: Set alarm to wait %d [msec] \
				\n\t\tbefore resending the request for hoplim = %d\n", timeout, rd->hoplimit);
	add_alarm(timeout, rd->key, ROUTE_DISC, aodv6_rreq_resend, rd);
}

/**
 * aodv6_rreq_repair - send local repair RREQ
 * @data: destination address
 *
 **/
void aodv6_rreq_repair(void *data)
{
	struct in6_addr *dst = (struct in6_addr *)data;

	if(xtr) fprintf(stdout,"==== %s ===\n",__FUNCTION__);

	if (rtnl_rt_exist(dst) == 0)
	{
		if(xdbg) fprintf(stdout,"Have route, will return!\n");
		return;
	}

	/* timeout, drop all packets for destination */
	pktqueue_remove(dst);
	if(xdbg) fprintf(stdout,
			"Timeout! All packets dropped for %x:%x:%x:%x:%x:%x:%x:%x\n",
			NIP6ADDR(dst));
	return;
}

/**
 * aodv6_rreq_fwd - forwards Route Request message
 * @dst: IPv6 header destination address
 * @rrep: RREq message to send
 * @len: length of RREQ message
 *
 * Fwd (unicast) the RREQ message @rreq to @dst.  
 * Caller must supply length of
 * message @len including any extensions.
 * If route to @dest does not exist, a route request 
 * will be generated. 
 **/
int aodv6_rreq_fwd(struct in6_addr *dst, 
						struct aodv6_rreq *rreq, 
						size_t len)
{
	struct sockaddr_in6 daddr;
	struct iovec iov[1];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	int cmsglen;
	struct in6_pktinfo pinfo;
	int ret = 0, on = 1;

	if (dst == NULL) return -ENODATA;

	if (rreq == NULL) {
		fprintf(stderr,
				"WARNNING === %s ===\n Asked to fwd with NULL rreq. Abandon!",
				__FUNCTION__);
		return -ENODATA;
	}
	
	if (len < sizeof(struct aodv6_rreq)) {
		fprintf(stderr,"ERROR === %s ===\nWrong len",__FUNCTION__);
		return -EINVAL;
	}
	
	if(xtr) 
		fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	if(xdbg) dump_mux(stdout,rreq,MSG_RREQ);
	
	memset(&daddr, 0, sizeof(struct sockaddr_in6));
	daddr.sin6_family = AF_INET6;

	/* Unicast to @dst 
	 * ---------------- */
	memcpy(&daddr.sin6_addr, dst, sizeof(struct in6_addr));
	daddr.sin6_port = htons(AODVPORT);

	memset(&pinfo, 0, sizeof(pinfo));
	//get_own_addr(&pinfo.ipi6_addr);
	get_host_addr(&pinfo.ipi6_addr);

	cmsglen = CMSG_SPACE(sizeof(pinfo));
	cmsg = malloc(cmsglen);
	if (cmsg == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}
	cmsg->cmsg_len = CMSG_LEN(sizeof(pinfo));
	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	memcpy(CMSG_DATA(cmsg), &pinfo, sizeof(pinfo));

	iov[0].iov_base = (unsigned char *)rreq;
	iov[0].iov_len = len;

	msg.msg_control = cmsg;
	msg.msg_controllen = cmsglen;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (void *)&daddr;
	msg.msg_namelen = sizeof(daddr);

	setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(int));

	ret = sendmsg(sockfd, &msg, 0);

	fprintf(stdout, 
			"\nINF-RREQ: fwdmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x[%x:%x:%x:%x:%x:%x:%x:%x] -> %x:%x:%x:%x:%x:%x:%x:%x \n",
	     	ret, NIP6ADDR(&rreq->orig), 
			NIP6ADDR(&pinfo.ipi6_addr), 
			NIP6ADDR(&rreq->dest));
	if(xdbg) dump_mux(stdout,rreq,MSG_RREQ);

	ulog(L_SEND_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_SEND_INFO, "sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	     ret, NIP6ADDR(&pinfo.ipi6_addr), NIP6ADDR(dst));
	ulog_if(L_SEND_PACKET) {
		dump_mux(ulogd, rreq, MSG_RREP);
	}
	free(cmsg);

	if(xtr) 
		fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	
	return 0;
}

/**
 * aodv6_rreq_rebroadcast - rebroadcast Route Request message
 * @rreq: RREQ to forward
 * @hoplimit: received hoplimit minus 1
 * @len: length of RREQ received
 *
 * Forwards a RREQ if @hoplimit has not reached zero.
 **/
int aodv6_rreq_rebroadcast(struct aodv6_rreq *rreq, int hoplimit, size_t len)
{
	/* use all manet nodes multicast */
	struct sockaddr_in6 daddr;
	struct in6_addr allnodes = IPV6_ALL_MANET_NODES;
	struct iovec iov[1];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	int cmsglen;
	struct in6_pktinfo pinfo;
	int ret = 0, on = 1;

	if(xtr) 
		fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (rreq == NULL) {
		fprintf(stderr, 
				"WARNNING === %s ===\n Asked to rebroadcast with NULL rreq. Abandon!",
				__FUNCTION__);

		return -ENODATA;
	}
	
	if (len < sizeof(struct aodv6_rreq)) {
		fprintf(stderr,"ERROR === %s ===\nWrong len",__FUNCTION__);
		return -EINVAL;
	}
	if (hoplimit <= 0) 
	{
		fprintf(stderr,"ERROR === %s ===\nWrong hoplimit",__FUNCTION__);
		return -EINVAL;
	}

	memset(&daddr, 0, sizeof(struct sockaddr_in6));
	daddr.sin6_family = AF_INET6;

	/* Use broadcasting address ff02::01 
	 * ----------------------------------*/
	memcpy(&daddr.sin6_addr, &allnodes, sizeof(struct in6_addr));
	daddr.sin6_port = htons(AODVPORT);

	//setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, 
	//	   &hoplimit, sizeof(int));

	memset(&pinfo, 0, sizeof(pinfo));
	//get_own_addr(&pinfo.ipi6_addr);
	get_host_addr(&pinfo.ipi6_addr);

	cmsglen = CMSG_SPACE(sizeof(pinfo));
	cmsg = malloc(cmsglen);
	if (cmsg == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}
	cmsg->cmsg_len = CMSG_LEN(sizeof(pinfo));
	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	memcpy(CMSG_DATA(cmsg), &pinfo, sizeof(pinfo));

	iov[0].iov_base = (unsigned char *)rreq;
	iov[0].iov_len = sizeof(struct aodv6_rreq);

	msg.msg_control = cmsg;
	msg.msg_controllen = cmsglen;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (void *)&daddr;
	msg.msg_namelen = sizeof(daddr);

	setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(int));

	ret = sendmsg(sockfd, &msg, 0);

	fprintf(stdout, 
			"INF-RREQ rbcast(%d) %x:%x:%x:%x:%x:%x:%x:%x[%x:%x:%x:%x:%x:%x:%x:%x] -> %x:%x:%x:%x:%x:%x:%x:%x [hlim=%d]\n",
	     	ret, 
			NIP6ADDR(&rreq->orig),   
			NIP6ADDR(&pinfo.ipi6_addr),
			NIP6ADDR(&daddr.sin6_addr),
			hoplimit);
	if(xdbg) dump_mux(stdout,rreq, MSG_RREQ);

	ulog(L_SEND_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_SEND_INFO, "sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	     ret, NIP6ADDR(&pinfo.ipi6_addr), NIP6ADDR(&daddr.sin6_addr));
	ulog_if(L_SEND_PACKET) {
		dump_mux(ulogd, rreq, MSG_RREQ);
	}

	if(xtr) 
		fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return ret;
}

/**
 * aodv6_rerr_send - send Route Error message
 * @dst: IPv6 header destination address for RERR
 * @rerr: RERR to send
 * @len: length of RERR message
 *
 * Sends a Route Error message @rerr to @dst.  Invalidates all
 * unreachable routes in routing table.
 **/
int aodv6_rerr_send(struct in6_addr *dst, struct aodv6_rerr *rerr, size_t len)
{
	struct sockaddr_in6 daddr;
	struct iovec iov[1];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	int cmsglen;
	struct in6_pktinfo pinfo;
	struct rt_entry *rte;
	int ret = 0, on = 1;

	if (dst == NULL) return -ENODATA;
	if (rerr == NULL) return -ENODATA;
	if (len < sizeof(struct aodv6_rerr)) return -EINVAL;

	memset(&daddr, 0, sizeof(struct sockaddr_in6));
	daddr.sin6_family = AF_INET6;

	/* Send error to @dst 
	 * ------------------ */
	memcpy(&daddr.sin6_addr, dst, sizeof(struct in6_addr));
	daddr.sin6_port = htons(AODVPORT);

	rte = rt_get(dst, 128, RT_ANY);
	
	//rt_get returns NULL if no route found
	if(rte)
		ret=rt_entry_precursor_destroy(rte);

	if(ret != 0)
		fprintf(stdout,"rt_entry_precursor_destroy returned %d\n", ret);

	memset(&pinfo, 0, sizeof(pinfo));
	//get_own_addr(&pinfo.ipi6_addr);
	get_host_addr(&pinfo.ipi6_addr);

	cmsglen = CMSG_SPACE(sizeof(pinfo));
	cmsg = malloc(cmsglen);
	if (cmsg == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}
	cmsg->cmsg_len = CMSG_LEN(sizeof(pinfo));
	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	memcpy(CMSG_DATA(cmsg), &pinfo, sizeof(pinfo));

	iov[0].iov_base = (unsigned char *)rerr;
	iov[0].iov_len = sizeof(struct aodv6_rerr);

	msg.msg_control = cmsg;
	msg.msg_controllen = cmsglen;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (void *)&daddr;
	msg.msg_namelen = sizeof(daddr);

	setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(int));

	ret = sendmsg(sockfd, &msg, 0);

	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	fprintf(stdout, 
			"INFO-RERR(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	      	ret, 
			NIP6ADDR(&rerr->dests[0].addr), //NIP6ADDR(&pinfo.ipi6_addr), 
			NIP6ADDR(dst));

	ulog(L_SEND_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_SEND_INFO, "sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	     ret, NIP6ADDR(&pinfo.ipi6_addr), NIP6ADDR(dst));
	ulog_if(L_SEND_PACKET) {
		dump_mux(ulogd, rerr, MSG_RERR);
	}

	return ret;
}

/**
 * aodv6_rrep_ack_send - send Route Reply Acknowledgment message
 * @dst: destination for RREP-ACK
 *
 * Sends a Route Reply Acknowledgment message to @dst in response to a
 * RREP with F_ACK flag set.
 **/
int aodv6_rrep_ack_send(struct in6_addr *dst)
{
	struct sockaddr_in6 daddr;
	struct iovec iov[1];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	int cmsglen;
	struct in6_pktinfo pinfo;
	struct aodv6_rrep_ack *ack;
	int ret = 0, on = 1;
	struct in6_addr localaddr;

	if (dst == NULL) return -ENODATA;

	ack = create_rrep_ack();

	memset(&daddr, 0, sizeof(struct sockaddr_in6));
	daddr.sin6_family = AF_INET6;
	memcpy(&daddr.sin6_addr, dst, sizeof(struct in6_addr));
	daddr.sin6_port = htons(AODVPORT);

	memset(&pinfo, 0, sizeof(pinfo));
	get_own_addr(&localaddr);
	get_host_addr(&pinfo.ipi6_addr);

	cmsglen = CMSG_SPACE(sizeof(pinfo));
	cmsg = malloc(cmsglen);
	if (cmsg == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}
	cmsg->cmsg_len = CMSG_LEN(sizeof(pinfo));
	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	memcpy(CMSG_DATA(cmsg), &pinfo, sizeof(pinfo));

	iov[0].iov_base = (unsigned char *)ack;
	iov[0].iov_len = sizeof(struct aodv6_rrep_ack);

	msg.msg_control = cmsg;
	msg.msg_controllen = cmsglen;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	msg.msg_name = (void *)&daddr;
	msg.msg_namelen = sizeof(daddr);

	setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &on, sizeof(int));

	ret = sendmsg(sockfd, &msg, 0);

	if (xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	fprintf(stdout, 
			"send-ackmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
			ret, 
			NIP6ADDR(&localaddr), //NIP6ADDR(&pinfo.ipi6_addr), 
			NIP6ADDR(dst));

	ulog(L_SEND_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_SEND_INFO, "sendmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x -> %x:%x:%x:%x:%x:%x:%x:%x\n",
	     ret, NIP6ADDR(&pinfo.ipi6_addr), NIP6ADDR(dst));
	ulog_if(L_SEND_PACKET) {
		dump_mux(ulogd, ack, MSG_RREP_ACK);
	}

	return ret;
}

/**
 * aodv6_hello_send - send Hello message
 * @setint: Hello Interval
 *
 * Sends a special Route Reply message called Hello message
 * advertising local node.  If @setint is set, adds Hello Interval
 * Extension to the message.  Extension value is controlled by
 * @setint.
 **/
int aodv6_hello_send(int setint)
{
	struct aodv6_rrep *rrep;
	struct in6_addr allnodes = IPV6_ALL_MANET_NODES;
	struct in6_addr own_ip;
	struct rt_entry *rte = NULL;
	int len = 0;
	u32 interval = htonl(setint);

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	get_own_addr(&own_ip);
	
	if(&own_ip == NULL)
	{
		fprintf(stderr,"ERROR === %s ===\nCan't determine own_ip\n",__FUNCTION__);
		return -1;
	}

	rte = rt_get(&own_ip, 128, RT_ACTIVE);
	//if ((xdbg) & (rte!= NULL)) fprintf(stdout,
	//		"\nGot RT to own_ip with lifetime=%u, expires=%d, seq_num=%d\n",
	//			rte->lifetime, (int)(rte->lifetime-timems()), rte->sequence);
	if ((rte == NULL) || (rte->lifetime<=timems()) )
	{
		struct addr_list *a, *list;

		if(xdbg)
		{
			fprintf(stdout,"\tRoute to own_ip %x:%x:%x:%x:%x:%x:%x:%x expired!\n",
						NIP6ADDR(&own_ip));
			fprintf(stdout, "\tUpdate it!\n");
		}

		ulog(L_WARN | L_SEND, "No route found\n");

		list = rtnl_local_addrs();
		for (a = list; a; a = a->next) 
		{
			if(memcmp(&own_ip,&a->addr,sizeof(struct in6_addr))==0)
			{
				if (rt_set_local(a)< 0) 
				{
					fprintf(stderr, "ERROR: Could not update AODV RT with local addr.\n");
					exit(-1);
				}
				
				if(xdbg) fprintf(stdout,
						"Updated local addr %x:%x:%x:%x:%x:%x:%x:%x to AODV RT (ifindex=%d)\n", 
					NIP6ADDR(&a->addr), a->ifindex);
			}
			
		}

		rte = rt_get(&own_ip, 128, RT_ACTIVE);
		if(rte==NULL)
		{
			fprintf(stderr,
					"ERROR: Route to own_ip %x:%x:%x:%x:%x:%x:%x:%x failed to updated\n", 
					NIP6ADDR(&own_ip));
			
			ulog(L_WARN | L_SEND, "No route found\n");
			return -1;
		}
	}
	
	//fprintf(stdout, __FUNCTION__" : Found rt_entry for %x:%x:%x:%x:%x:%x:%x:%x via %x:%x:%x:%x:%x:%x:%x:%x exiration(%u)\n", 
	//		NIP6ADDR(&rte->destination), 
	//		NIP6ADDR(&rte->nexthop),
	//		rte->lifetime);

	/* Use RREP msg for Hello Msgs, with destination=own_ip */
	rrep = create_rrep(0, rte->plen, 0, 
				&own_ip,
				inc_get_seqnum(), //rte->sequence, <- NOT working with it. 
				&own_ip, 
			  // ALLOWED_HELLO_LOSS * HELLO_INTERVAL);
			  ALLOWED_HELLO_LOSS * helloInterval
#ifdef _SECURE_BOOT_
				  , 1
#endif
				  );

	free(rte);
	len = sizeof(struct aodv6_rrep);

	if (setint > 0) {
		len = add_ext((void *)&rrep, len, EXT_HELLO_INTERVAL, 
			       sizeof(int), (void *)&interval);
		//ulog(L_SEND_INFO, "%s: interval = %d %d\n",__FUNCTION__, ntohl(interval), setint);
	}

	//fprintf(stdout,__FUNCTION__ ": Sending hello-rrep of len %d to allnodes at ff02::1\n", len);
	
	aodv6_rrep_send(&allnodes, rrep, len);
	free(rrep);
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return 0;
}

/**
 * is_hello - check if RREP is Hello message
 * @rrep: route reply to check
 *
 * Determines whether @rrep is a Hello message.  Returns true
 * (non-zero) if RREP is Hello, otherwise zero.
 **/
int is_hello(struct aodv6_rrep *rrep)
{
	if (rrep == NULL) return -ENODATA;

	return !ipv6_addr_cmp(&rrep->orig, &rrep->dest);
}

/**
 * aodv6_rrep_recv - handle Route Reply message
 * @msg: message data buffer
 * @len: length of data in buffer
 *
 * Parses RREP message.  Adds or updates entry in routing table and
 * sends back RREP-ACK if necessary.
 **/
int aodv6_rrep_recv(struct msghdr *msg, size_t len)
{
	struct aodv6_rrep *rrep = (struct aodv6_rrep *)msg->msg_iov->iov_base;
	struct sockaddr_in6 *ss_addr = (struct sockaddr_in6 *)msg->msg_name;
	struct in6_addr *saddr = (struct in6_addr *)&ss_addr->sin6_addr;
	struct in6_addr daddr;
	struct rt_entry *fwd_rte = NULL;
	struct cmsghdr *cmsg;
	struct in6_pktinfo *infop;
	u32 flags = 0;
	int update = 0, has_exts = 0, ifindex = 0;
	int isForMe =0, isHello = 0, isFromMe = 1, updateKernel=1;
	int rc = 0;
	
	//unsigned long t1 = timems();
	
	isHello = is_hello(rrep);
	isFromMe = (ipv6_addr_cmp(&Config.NodeAddrGlobal, &rrep->dest) == 0)?1:0;  
	
	if ((!isHello)&(xtr)) fprintf(stdout, "=== %s ===\n",__FUNCTION__);

	if (msg == NULL) 
	{
		fprintf(stderr,"ERROR: NULL message\n");
		return -ENODATA;
	}
	if ((len < sizeof(struct aodv6_rrep)) && (!isHello))
	{ 
		fprintf(stderr,"ERROR: Invalid message of len %d vs. %d from"\
				" %x:%x:%x:%x:%x:%x:%x:%x\n", len, sizeof(struct aodv6_rrep), 
					NIP6ADDR(saddr));		
		/* invalid message length */
		return -E_MSGLEN;
	}

	ulog(L_RECV_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_RECV_INFO, "recvmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(saddr));
	ulog_if(L_RECV_PACKET) {
		dump_mux(ulogd, rrep, MSG_RREP);
	}
	
	if(!isHello)
	{
		fprintf(stdout,
				"\n---> RREP-recvmsg(%d) From (rrep->dest) %x:%x:%x:%x:%x:%x:%x:%x\n",
				len, NIP6ADDR(&rrep->dest));
		fprintf(stdout,"\t\tthrough %x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(saddr));
		fprintf(stdout,"\t\tfor (rrep->orig) %x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&rrep->orig));
	}

	isForMe = rt_is_local_addr(&rrep->orig);

	if(isHello & xdbg)
	{
		fprintf(stdout,
				"HELLO-reply: from rrep->dest[saddr] -> rrep->orig:\n");
		fprintf(stdout,"\t%x:%x:%x:%x:%x:%x:%x:%x[%x:%x:%x:%x:%x:%x:%x:%x]",
				NIP6ADDR(&rrep->dest),NIP6ADDR(saddr));
		fprintf(stdout," -> %x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&rrep->orig));
	}
	
	/* 1. Check if this is our RREP that the neighbor is forwarding 
	 * Needed??
	 * */
	if(isFromMe & xdbg)
	{
		fprintf(stdout, "\n\n--->After %lu [msec] I CAN SEE MY RREP FORWARDED \
				\n\tBY %x:%x:%x:%x:%x:%x:%x:%x \n\n",rrep->lifetime - timems(),
				NIP6ADDR(saddr));

		return 0;
	}
	
	/* 2. Verify IP and Signature */
#ifdef _SECURE_IP_
	if(!isFromMe) //Check for all messages, routed or hellos, because the RT gets updated.
	{
		/* For a Hello message we want to check the IP and the signature before
		 * updating our Routing Tables
		 */
		/* Check the IP and signature for Hello msg*/
		if(rrep->keyLen == 0) 
		{
			fprintf(stderr,"NO KEY PROVIDED IN THE MESSAGE!\n");
#ifndef _HACK_TEST_
			return -1;
#endif
		}
		
		if(!verify_secIP( SCOPE_GLOBAL, rrep->pubKey, rrep->keyLen, &rrep->dest ))
		{
#ifdef _SECURE_BOOT_
			
			//IP is OK, check the signature
			if(verify_sign(rrep, MSG_RREP))
			{
				fprintf(stderr,"ERROR: Signature NOT verified for HELLO MSG!\n");
				return -1;
			}
#endif
			
		} else
		{
			fprintf(stderr,"ERROR: IP NOT verified for RREP!\n");
			return -1;
		} 
		
	}
#endif

	if (!isHello & xdbg)
	{
		fprintf(stdout,"recvmsg[len=%d] from %x:%x:%x:%x:%x:%x:%x:%x\n",
			len,NIP6ADDR(saddr));
		dump_mux(stdout, rrep, MSG_RREP);
	}

	for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) 
	{
		if (cmsg->cmsg_type == IPV6_PKTINFO) 
		{
			infop = (struct in6_pktinfo *) CMSG_DATA(cmsg);
			memcpy(&daddr, &infop->ipi6_addr, sizeof(struct in6_addr));
			ifindex = infop->ipi6_ifindex;
			if(!isHello && xdbg)
			{
				fprintf(stdout,
						"RREP-recv: cmsg_type=IPV6_PKTINFO\n"); 
				fprintf(stdout, "\t sent by infop->ipi6_addr: %x:%x:%x:%x:%x:%x:%x:%x\n", 
						NIP6ADDR(&infop->ipi6_addr));
			}
			break;
		}
	}
	
	//Check if saddr is diff or same with rrep->dest and if not store the RT
	//REMEMBER: rrep->dest is the addr of the dest for 
	//which route is supplied and rrep->orig is the originator of the initial RREQ
	//Even if the sender node is one hop away saddr should be node's Ether addr
	//and the rrep->dest (if routing dest is this neighbor node) should be its AODVaddr 
	//We need this RT.... 
	if (memcmp(saddr, &rrep->dest, sizeof(struct in6_addr)) != 0) 
	{
		//MI - HERE nexthop set to NULL will error! 
		//Gets set to NULL if addr==nexthop
		//rtnl_rt_update(saddr, NULL, ifindex, ntohl(rrep->lifetime));
		//rtnl_rt_update(saddr, saddr,ifindex, ntohl(rrep->lifetime));
		//MI - when do I add this if not here?
		
		 if (!isHello)	//Do not print for Hello msgs.
		 {
			fprintf(stdout,"INF-RREP: Update kernel RT: \n");
		 	fprintf(stdout,"\t%x:%x:%x:%x:%x:%x:%x:%x/%d nexthop %x:%x:%x:%x:%x:%x:%x:%x (lifetime=%d)\n",
					NIP6ADDR(&rrep->dest), rrep->prefix_len,
					NIP6ADDR(saddr), ntohl(rrep->lifetime));
		 }

		 //Update Kernel's RT. Returns 0 for success
		 rc = rtnl_rt_update(&rrep->dest,saddr,ifindex,ntohl(rrep->lifetime));
		 if(rc) 
			fprintf(stderr,"WARNING: Kernel RT not updated. Err=%d\n",rc);
	}

	if (len > sizeof(struct aodv6_rrep))
		has_exts = 1;

#ifndef _SECURE_IP_
	if (rrep->flags & RREP_F_REPAIR) flags |= F_REPAIR;
#endif

	if (rrep->flags & RREP_F_ACK) flags |= F_ACK;

	if((!isHello) & (xdbg))
	{
		fprintf(stdout, "Looking in RT for rrep->dest: ");
		fprintf(stdout,"%x:%x:%x:%x:%x:%x:%x:%x/%d\n",
				NIP6ADDR(&rrep->dest), rrep->prefix_len);
	}
	
	//I believe I should only be interested in prefix_len=128 and in the
	//original design was always 0!
	if ((fwd_rte = rt_get(&rrep->dest, 
					rrep->prefix_len<128?128:rrep->prefix_len, 
					RT_ANY)) != NULL) 
	{
		/* Route is in routing table, 
		 * --------------------------*/
		int ret = 0;
 
		/* Update the lifetime 
		 * --------------------*/
		fwd_rte->lifetime = timems() + ntohl(rrep->lifetime);
		//rt_update(fwd_rte, updateKernel, 1);
		update = 1;
		
		/*  Update seq # if 
		 *  -----------------
		 * - RREP dest seq is greater than that in routing table 
		 * - seq is equal but route is not active 
		     (hopcount == INFINITE_HOPS) 
		 * - seq is equal but RREP hopcount is smaller */
		if((!isHello) & (fwd_rte->hopcount != rrep->hopcount+1) & (xdbg))
			fprintf(stdout,"Found route with fwd_rte->hopcount=%d, \
					\n\t\tbut rrep->hopcount=%d\n",
					fwd_rte->hopcount, rrep->hopcount);
		if ((ntohl(rrep->dest_seq) > fwd_rte->sequence) || 
		    ((ntohl(rrep->dest_seq) == fwd_rte->sequence) && 
		     //(rrep->hopcount < fwd_rte->hopcount))) 
			 //See RFC 3561, Section 6.2
			 (rrep->hopcount + 1 < fwd_rte->hopcount))) 
		{
			if((!isHello) & (xdbg)) 
				fprintf(stdout,"Update above route entry.\n");
			update = 1; 
		}

		/* MI: - RFC 3561
		 * The the precursors list of the 
		 * RT to destination must store: 
		 * 1) originator of the route request (can not be done here - on RREQ-fwd) 
		 * 2) the next hop to which the RREP will be forwarded
		 * 3) Also the RT to the next hop to destnation must contain the 
		 * next hop to originator as precursor - See 3) bellow
		 */
		/* 2) */
		/* MI: NOTE: Hello msg are RREPs with orig == dest. Skip them.
		 */
		if(!isHello)
		{
			struct in6_addr nexthop;
			if (rt_get_nexthop(&rrep->orig, &nexthop) == 0)
				ret = rt_entry_precursor_add(fwd_rte, &nexthop);

		}
		/* 3) can not be achieved here because we are using the
		 * wireless interface address as next hop and have no 
		 * RT for neighbors on this addresses*/

	} else {

		/* Route not in routing table, create 
		 * ----------------------------------*/
		//MI-don't print info for Hellos
		if(!isHello)
			fprintf(stdout,"No existing route in RT. Create one\n");
		fwd_rte = malloc(sizeof(struct rt_entry));
		if (fwd_rte == NULL) {
			perror(__FUNCTION__);
			return -ENOMEM;
		}
		memset(fwd_rte, 0, sizeof(struct rt_entry));
		update = 1;
	}

	/* Route have to be updated 
	 * -------------------------*/
	if (update) 
	{
		memcpy(&fwd_rte->destination, &rrep->dest, sizeof(struct in6_addr));
		if (IN6_IS_ADDR_LINKLOCAL(&saddr) == 1) 
		{
			//MI-don't print info for Hellos
			if(!isHello)
				fprintf(stdout,
						"Sender is link-local addr: %x:%x:%x:%x:%x:%x:%x:%x Set nexthop to ::\n", 
					NIP6ADDR(saddr) );
			memset(&fwd_rte->nexthop, 0, sizeof(struct in6_addr));
		} else
			memcpy(&fwd_rte->nexthop, saddr, sizeof(struct in6_addr));
		
		fwd_rte->plen = rrep->prefix_len;
		fwd_rte->interface = ifindex;
		fwd_rte->sequence = ntohl(rrep->dest_seq);
		fwd_rte->hopcount = rrep->hopcount + 1;
		fwd_rte->last_hopcount = rrep->hopcount + 1;
		fwd_rte->lifetime = timems() + ntohl(rrep->lifetime);
		fwd_rte->flags = 0;

		//MI-don't print info for Hellos
		if((!isHello) && (xdbg))
		{
			fprintf(stdout,
					"\nSet RT: %x:%x:%x:%x:%x:%x:%x:%x/%d nexthop=%x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&rrep->dest), rrep->prefix_len, NIP6ADDR(saddr));
			fprintf(stdout,"lifetime=%u dest-seq=%d hopcount=%d\n",
					ntohl(rrep->lifetime),fwd_rte->sequence,fwd_rte->hopcount);
		}
		
		if (has_exts) 
		{
			struct extension *ext;
			int rlen = len - sizeof(struct aodv6_rrep);

			while (rlen > 0) {
				ext = (struct extension *)rrep + 1;
				if (rlen < ext->length) {
					/* packet too small */
					ulog(L_WARN | L_RECV, "packet too small to hold extension\n");
					goto failed;
				}
				switch (ext->type) {
				case EXT_HELLO_INTERVAL:
					if (ext->length != 4) {
						/* invalid length */
						ulog(L_WARN | L_RECV, "invalid extension length\n");
						rlen = -1;
						goto failed;
					}
					fwd_rte->hello_interval = 
						((struct aodv6_ext_hello_int *)ext)->interval;
					rlen -= sizeof(struct aodv6_ext_hello_int);
					break;
				case EXT_MANET_GATEWAY:
					//if (ext_gw_rcv((struct aodv6_ext_gw *)ext) < 0)
					if (ext->length == 0) {
						/* invalid length */
						ulog(L_WARN | L_RECV, "invalid extension length\n");
						rlen = -1;
						goto failed;
					}
					rlen -= sizeof(struct aodv6_ext_gateway);
					break;
				default:
					/* unknown extension */
					break;
				}
			}
		}
		if((!isHello) & (xdbg))
			fprintf(stdout,"Calling rt_update() with kernel=1, create=1\n" );
		rt_update(fwd_rte, updateKernel, 1);
		ulog(L_SEND_INFO, "send queued packets\n");
		if((!isHello) & (xdbg))
			fprintf(stdout,"Sending queue packets...\n");
		pktqueue_remove(&fwd_rte->destination);
	}

	if (is_hello(rrep)) 
	{
		struct in6_addr *data;
		data = (struct in6_addr *)malloc(sizeof(struct in6_addr));
		memcpy(data,&rrep->dest,sizeof(struct in6_addr));

		//add_alarm(ALLOWED_HELLO_LOSS * HELLO_INTERVAL, in6_hash(&rrep->dest),
		add_alarm(ALLOWED_HELLO_LOSS * helloInterval, in6_hash(&rrep->dest),
			  HELLO_MISS, hello_missed, (void *)data);
	}
	
	if((!isHello) && (xdbg))
		fprintf(stdout, "RREP orig %x:%x:%x:%x:%x:%x:%x:%x/128 isHello=%d isForMe=%d\n", 
			NIP6ADDR(&rrep->orig), isHello, isForMe);

	if (!is_hello(rrep) && !isForMe && update) 
	{
		/* we were not original requester for this RREP,
                   forward RREP to the next hop towards the source */
		struct in6_addr nexthop;
		int ret = 0;

		fprintf(stdout,
				"INF-RREP: Forwarding rrep to orig %x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&rrep->orig));

		//if (rt_get_nexthop(&rrep->orig, &nexthop) < 0) 
		if (rt_get_nexthop(&rrep->orig, &nexthop) == 0)
		{
			struct aodv6_rrep *fwd_rrep = NULL;
			fprintf(stdout,"INF-RREP: For rreq->orig: %x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&rrep->orig));
			fprintf(stdout,"\t\tnexthope is : %x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&nexthop));
			fwd_rrep = create_rrep(
				flags, rrep->prefix_len, rrep->hopcount + 1,
				&rrep->dest, ntohl(rrep->dest_seq),
				&rrep->orig, ntohl(rrep->lifetime)
#ifdef _SECURE_BOOT_
				, 0
#endif
				);

#if defined _SECURE_IP_ || defined _SECURE_BOOT_
			copy_secData(fwd_rrep, rrep, MSG_RREP);
#endif
			
			ret = aodv6_rrep_send(&nexthop, fwd_rrep, len);
			if((ret) & (xdbg))
				fprintf(stdout,"WARNING-%s:%d: rrep_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
			free(fwd_rrep);
		}
		else
		{
			if(xdbg) fprintf(stdout,"No nexthop for rrep->orig ");
			print_in6addr(stdout,&rrep->orig);
		}
			
	}

	if (flags & F_ACK)
		aodv6_rrep_ack_send(saddr);
	if (fwd_rte)
		free(fwd_rte);

	//fprintf(stdout,"\n---> TIMER rrep : %lu - %lu = %lu\n", timems(), t1, timems()-t1);

	return 0;

 failed:
	if (fwd_rte)
		free(fwd_rte);

	return -1;
}

/**
 * aodv6_rreq_recv - handle Route Request message
 * @msg: message data buffer
 * @len: length of data in buffer
 *
 * Parses RREQ message.  Determines if node should reply to or forward
 * the request.
 **/
int aodv6_rreq_recv(struct msghdr *msg, size_t len)
{
	struct aodv6_rreq *rreq = (struct aodv6_rreq *)msg->msg_iov->iov_base;
	struct sockaddr_in6 *ss_addr = (struct sockaddr_in6 *)msg->msg_name;
	struct in6_addr *saddr = (struct in6_addr *)&ss_addr->sin6_addr;
	struct in6_addr daddr;
	struct rt_entry *fwd_rte = NULL, *rev_rte = NULL;
	int *hopptr, hops = -1, ret = 0, ifindex = 0;
	struct in6_pktinfo *infop;
	struct cmsghdr *cmsg;
	u32 flags = 0, rflags = 0;
	int local = 0;
	int newRT = 0;
#ifdef _eAODVD_
	int _dropIt = 0, _chgIt = 0;
#endif

	//unsigned long t1 = timems();
	
	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);

	fprintf(stdout, "\n---> RREQ-recvmsg(%d) From (rreq->orig) %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(&rreq->orig));
	fprintf(stdout,"\tthrough (saddr): %x:%x:%x:%x:%x:%x:%x:%x\n",
			NIP6ADDR(saddr));
	fprintf(stdout, "\tfor (rreq->dest): %x:%x:%x:%x:%x:%x:%x:%x\n",
			NIP6ADDR(&rreq->dest));	

	if (msg == NULL) return -ENODATA;

	if (len < sizeof(struct aodv6_rreq)) {
		/* invalid message length */
		return -E_MSGLEN;
	}

	local = rt_is_local_addr(&rreq->dest);
	
	if(local) {
		fprintf(stdout,"INF-RREQ: We are the rreq->dest.\n");
		dump_mux(stdout,rreq, MSG_RREQ);
	}

#ifdef _eAODVD_
	{
		/* Any IP from which we were instructed to drop packets?*/
		if(dropAddr) 
		{
			struct in6_addr dropIP6;
	
			if (inet_pton(AF_INET6, dropAddr, &dropIP6) <= 0)
					return -1;
			
			if(ipv6_addr_cmp(&rreq->orig, &dropIP6) == 0)
			{
				_dropIt = 1;
				/* do we have to drop packets randomly? */
				if(_dropRnd)
				{
					/* Lucky packet ... */
					srand(in6_hash(&dropIP6));
					/* drop ~ 1/2, randomly */
					if(rand()%2 != 0)
						_dropIt = 0;
				}

				if(_alterPkt)
				{
					/* Don't need to drop, only change it */
					_dropIt = 0;
					_chgIt = 1;
				}
			}
		}
		/* Drop packets coming fron any address randomly */
		else if (_dropRnd)
		{
			srand(987456);
			if(rand()%3==0)
				_dropIt = 1;
		}
		else if (_alterPkt)
		{
			_chgIt = 1;
		}
		
		if(_dropIt)
		{
			if(xdbg) 
			{
				fprintf(stdout, "\t---> Dropping packet from ");
				print_in6addr(stdout, &rreq->orig);
			}
			return 0;
		}

	}
#endif

	/* 0. Check if received RREQ with this src IP and RREQ ID
     *     within PATH_DISCOVERY_TIME. If so silently discard. 
	 */
	if (rreq_check_dupe(&rreq->orig, rreq->rreq_id)) 
	{
		ulog(L_RECV_INFO, "duplicate message received\n");
		//if(xdbg) 
			fprintf(stdout, "---> Duplicate RREQ rcvd whithin PATH_DISCOVERY_TIME\n");
		return 0;
	} 
	/* The above check (original code) should prevent us from getting here
	 * but IT DOES NOT, so we get bounced back messages...
	 * Let's see if it works.
	 * Check if RREQ is not a bounced back message. 
	 *		a) Are we the source of the message?
	 *		b) Have we seen this already? <- How can we check other than dupelist?
	 *		If yes drop it and wait
	 */
	/* a) Are we the originator of the message? */
	else if((ipv6_addr_cmp(&rreq->orig, &Config.NodeAddrGlobal) == 0) || 
			(ipv6_addr_cmp(&rreq->orig, &Config.HostAddrGlobal) == 0))	
	{
		if(xdbg)
		{
			fprintf(stdout,
					"Own RREQ for %x:%x:%x:%x:%x:%x:%x:%x ", 
					NIP6ADDR(&rreq->dest));
			fprintf(stdout,
					"bounced back from %x:%x:%x:%x:%x:%x:%x:%x\n",
	      			NIP6ADDR(saddr));
			dump_mux(stdout, rreq, MSG_RREQ);
		}

		return 0;	
	} 
	/* b) Did we just forward the request?
	 * Is the seq # of the originator same with the one in request?
	 */
	if ((rev_rte = rt_get(&rreq->orig, 128, RT_ANY)))
	{
		if(rev_rte->sequence == rreq->orig_seq)
		{
			fprintf(stdout,
					"RREQ for %x:%x:%x:%x:%x:%x:%x:%x bounced back from %x:%x:%x:%x:%x:%x:%x:%x.\n", 
					NIP6ADDR(&rreq->dest), NIP6ADDR(saddr));
			return 0;
		}
	}
	/* c) Do we still have an older route to destination for this RREQ, but
	 * nexthop is the neighbor that has no more route to this destination?
	 * Invalidate the route through this neigbor and return.
	 */
	if ((fwd_rte = rt_get(&rreq->dest, 128, RT_ANY)))
	{
		if((ipv6_addr_cmp(saddr, &fwd_rte->nexthop) == 0))
		{
			fprintf(stdout,
					"RREQ-bounced back: RT for %x:%x:%x:%x:%x:%x:%x:%x now obsolete.\n", 
					NIP6ADDR(&rreq->dest));
			/* Invalidate route from AODV RT and delete from Kernel*/
			rt_del(&fwd_rte->destination,fwd_rte->plen);
			return 0;
		}
	}
	
#ifdef _SECURE_BOOT_

	/* 1. Check the IP and signature of the RREQ
	 * even if we only rebroadcast the message
	 * because we do not want to update RT with invalid info
	 */
	//if(local)
	{
		if (xdbg) dump_mux(stdout, rreq, MSG_RREQ);

		/* Verify IP 
		 * ----------*/
		if (xdbg) fprintf(stdout,"\t--->\tVerify IP !\n");
	
		//verify_secIP returns 0 on success
		if(!verify_secIP( SCOPE_GLOBAL, rreq->pubKey, rreq->keyLen, &rreq->orig ))
		{
			/* IP is OK, check the signature
			 * ------------------------------*/
			if(verify_sign(rreq, MSG_RREQ))
			{
				fprintf(stderr,"ERROR: Signature NOT verified for RREQ from ");
				print_in6addr(stdout, &rreq->orig);
				return -1;
			}
			
		} else
		{
			fprintf(stderr,"ERROR: IP NOT verified for RREQ!\n");
			return -1;
		}
	}
	
#endif

	/* RFC : 
	 * 	"When used to forward, update activeRouteTimeout for
	 * 		source, 
	 * 		destination,
	 * 		nexthop" 
	 */
	/* Update active route lifetime to previous hop 
	 * --------------------------------------------*/
	/* This will always be true because saddr is wireless interface address. 
	 * I don't think the IP of my neightbor's interface 
	 * is needed in my routing table, so I'll change the logic a bit.
	 * WARNING: If my own rreq bounced back and I don't catch it, this cause
	 * wrong entry in the Kernel RT:
	 * 		my_aodv0_IP via neighbot_eth0_IP
	 */
	if (memcmp(saddr, &rreq->orig, sizeof(struct in6_addr)) != 0)
	{
		//int rc = 0;
		//int rlife = 0;
		/* This is the original code I changed ....
		 * If called with nh=NULL will return error. 
		 * It puts NULL if dest==nh 
		 */
		//rtnl_rt_update(saddr, NULL, ifindex, ACTIVE_ROUTE_TIMEOUT);
		//rtnl_rt_update(saddr, saddr, ifindex, 
		//		max((ALLOWED_HELLO_LOSS * helloInterval),activeRouteTimeout));
		/* Next part is moved bellow 
		rlife = (reverseRouteLife < myRouteTimeout)? 
		           myRouteTimeout: reverseRouteLife; //-not RFC but better prformance
		//rlife = 2*(reverseRouteLife - (NODE_TRAVERSAL_TIME * rreq->hopcount));
		if (xdbg)
			fprintf(stdout,"Update RT for %x:%x:%x:%x:%x:%x:%x:%x/128\
					\n\twith nexthop: %x:%x:%x:%x:%x:%x:%x:%x and relative lifetime=%d\n",
					NIP6ADDR(&rreq->orig),
					NIP6ADDR(saddr), rlife);
		
		//Update Kernel's RT. Returns 0 for success
		//rlife = timems() + reverseRouteLife
		rc = rtnl_rt_update(&rreq->orig,saddr,ifindex,rlife);
		*/
	}	


	if (rreq->flags & RREQ_F_JOIN) flags |= F_JOIN;
#ifndef _SECURE_BOOT_
	if (rreq->flags & RREQ_F_REPAIR) flags |= F_REPAIR;
	if (rreq->flags & RREQ_F_GRATUITOUS) flags |= F_GRATUITOUS;
#endif

	//Dump the request for debugging...
	//dump_mux(stdout,rreq, MSG_RREQ);
	
	ulog(L_RECV_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_RECV_INFO, "recvmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(saddr));
	
	ulog_if(L_RECV_PACKET) {
		dump_mux(ulogd, rreq, MSG_RREQ);
	}

	for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) 
	{
		if (cmsg->cmsg_type == IPV6_HOPLIMIT) 
		{
			hopptr = (int *) CMSG_DATA(cmsg);
			hops = *hopptr;
#ifdef _eAODVD_
			/* Force the neighbor to drop the packets by resetting the 
			 * number of hops the packet can travel 
			 */
			if(_chgIt)
			{
				fprintf(stdout, "---> Changing # of hops to 200 :-}. Gotcha!\n");
				//hops = 2;
				hops = 200;
			}
#endif
			if(xdbg) fprintf(stdout, 
						"cmsg_type is IPV6_HOPLIMIT and got hops=%d\n", 
						hops); 
		}
		if (cmsg->cmsg_type == IPV6_PKTINFO) 
		{
			infop = (struct in6_pktinfo *) CMSG_DATA(cmsg);
			memcpy(&daddr, &infop->ipi6_addr, sizeof(struct in6_addr));
			ifindex = infop->ipi6_ifindex;
			//if (xdbg) 
			//	fprintf(stdout, 
			//"cmsg_type = IPV6_PKTINFO, infop->addr: %x:%x:%x:%x:%x:%x:%x:%x\n", 
			//			NIP6ADDR(&daddr));
		}
		if (cmsg->cmsg_type == IPV6_PKTINFO) 
		{
			infop = (struct in6_pktinfo *) CMSG_DATA(cmsg);
			memcpy(&daddr, &infop->ipi6_addr, sizeof(struct in6_addr));
			ifindex = infop->ipi6_ifindex;
			//if (xdbg) 
			//	fprintf(stdout, 
			//"cmsg_type = IPV6_PKTINFO, infop->addr: %x:%x:%x:%x:%x:%x:%x:%x\n", 
			//			NIP6ADDR(&daddr));
		}
	}

	/* 2. Check for reverse route to originator
	 * - copy src seq from RREQ
	 * - next hop from IPv6 hdr src addr
	 * - hopcount from RREQ
	 * - lifetime max(current, time+REV_ROUTE_LIFE)
	 */
	fprintf(stdout,"INF-RREQ: Check and save reverse RT to rreq->orig\n");
	if ((rev_rte = rt_get(&rreq->orig, 128, RT_ANY))) 
	{
		if((ipv6_addr_cmp(saddr, &rev_rte->nexthop) == 0))
		{
		/* Reverse route exists.
         ------------------------*/
		if(xdbg) 
		fprintf(stdout,
				"\tRev_rt exists for rreq->orig %x:%x:%x:%x:%x:%x:%x:%x.\n", 
				NIP6ADDR(&rreq->orig));
		
		/* Update lifetime if same nexthop */
		/* ------------------------------- */
		{ 
			//RFC definition is next but it gives bad performance
			////use same as newRT case (bellow)
			int ltime = max(reverseRouteLife, activeRouteTimeout)+ 
						(rev_rte->hopcount* (NODE_TRAVERSAL_TIME + NODE_SEC_VERIFY_TIME)*2);
			ltime = timems()+ltime; 
			rev_rte->lifetime = (rev_rte->lifetime < ltime)? ltime : rev_rte->lifetime;
			
			/* //RFC: see also 6.5
			rev_rte->lifetime = max(rev_rte->lifetime,
						//(timems() + 2 * (NET_TRAVERSAL_TIME - //RFC: see akso 6.2
					  (timems() + reverseRouteLife - 
						(2 * rreq->hopcount * NODE_TRAVERSAL_TIME)));
			*/
		}
		
		if(xdbg) 
			fprintf(stdout, "\tRev_rt relative lifetime updated to %lu [msec]\n",
				rev_rte->lifetime - timems());
			
		/* Update seq # if
		 * -----------------
		 * - RREQ src seq is higher than rt's
		 * - seq is equal but RREQ hopcount is smaller */
		if ((ntohl(rreq->orig_seq) > rev_rte->sequence) || 
		    (ntohl(rreq->orig_seq) == rev_rte->sequence && 
		     // rreq->hopcount + 1 < rev_rte->hopcount)) //hopcount is not stored incremented
		     rreq->hopcount < rev_rte->hopcount)) 
		{
			rev_rte->sequence = ntohl(rreq->orig_seq);
			if(xdbg) 
				fprintf(stdout,
					"\tRev_rt updated with nh=saddr and rev_hopcount=%d\n", rreq->hopcount);
			memcpy(&rev_rte->nexthop, saddr, 
			       sizeof(struct in6_addr));
			rev_rte->hopcount = rreq->hopcount;
		}

		ulog(L_INFO, "updated reverse route\n");
		
		 /* Update the Routing Tables with new lifetime and hopcount 
		  * --------------------------*/
		/* rt_update calls rtnl_rt_update() for Kernel */
		//ret = rt_update(rev_rte, 0, 1);
		//ret = rt_update(rev_rte, 1, 1);
		printf("HERE1 update lifetime to %lu\n",rev_rte->lifetime-timems());
		ret = rt_update(rev_rte,1,0);
		}		
		else //the message came on a different RT
			newRT = 1;
	} else 
		newRT = 1;
		
	if(newRT) 
	{
		/* No reverse route. Create new. 
		 * -----------------------------*/
		fprintf(stdout,"INF-RREQ: No reverse route found. Create new one\n");
		rev_rte = malloc(sizeof(struct rt_entry));
		if (rev_rte == NULL) 
		{
			fprintf(stderr,"Can't malloc\n");
			perror(__FUNCTION__);
			return -ENOMEM;
		}
		memset(rev_rte, 0, sizeof(struct rt_entry));
		memcpy(&rev_rte->destination, &rreq->orig, sizeof(struct in6_addr));
		rev_rte->interface = ifindex;
		rev_rte->plen = 128;
		rev_rte->sequence = ntohl(rreq->orig_seq);
		rev_rte->hopcount = rreq->hopcount;
		memcpy(&rev_rte->nexthop, saddr, sizeof(struct in6_addr));
		rev_rte->precursors = NULL;
		//rev_rte->lifetime = reverseRouteLife;  //see sect 6.2 too;
		{ 
			int ltime = reverseRouteLife + 
				(rev_rte->hopcount* (NODE_TRAVERSAL_TIME + NODE_SEC_VERIFY_TIME)*2);
			/* No less than activeRouteTimeout - RFC section 6.2 */
			ltime = (ltime<activeRouteTimeout)? activeRouteTimeout:ltime; 
			//ltime = (ltime<reverseRouteLife)?reverseRouteLife:ltime;
			printf("HERE2 lifetime rev_rte = %d\n", ltime);
			rev_rte->lifetime = timems()+ltime; 
		}
		rev_rte->flags = 0;
		ulog(L_INFO, "created reverse route\n");
		ret = rt_update(rev_rte, 1, 1);
		if(ret != 0)
		{
			fprintf(stderr,"ERROR: RT update failed for rreq->orig: ");
			print_in6addr(stderr, &rreq->orig);
		}
		free(rev_rte);
	}

	/* 3. Check if we have active route to the destination.  
	 * 	  Or if we are the destination. If
     *    not rebroadcast RREQ to all interfaces. 
	 */
	//xdbg = 1;
	if (local) 
	{
		/* - we are the destination 
		 * ------------------------*/
		struct aodv6_rrep *new_rrep = NULL;
		u32 seq = 0;  
		if(xdbg) fprintf(stdout,"\t--->\tWe are the destination!\n");
		
		//Update local sequence_number
		seq = upd_get_seqnum(ntohl(rreq->dest_seq));
		
		new_rrep = create_rrep(
			rflags, 128, 0, &rreq->dest, seq,
			&rreq->orig, 
			(myRouteTimeout + 
			 	(rreq->hopcount*(NODE_TRAVERSAL_TIME + NODE_SEC_VERIFY_TIME)*2)) //MY_ROUTE_TIMEOUT 
#ifdef _SECURE_BOOT_
			, 1
#endif
			);
		
		ret = aodv6_rrep_send(saddr, new_rrep,
					 sizeof(struct aodv6_rrep));
		if(xdbg)
		{ 
			if(ret)
				fprintf(stdout,"WARNING-%s:%d: rrep_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
			else
				fprintf(stdout,"\t---> RREP-SENT to saddr %x:%x:%x:%x:%x:%x:%x:%x\n",
						NIP6ADDR(saddr));
		}
		free(new_rrep);
	} else 
	{
		int noRT = 0;
		int oldRT = 0; 
		/* MI: If there is no RT from HelloMsg =>
		 * we need hoplimit > 1. No reason to
		 * waist time and start with hoplimit 1
		 * out of 1,3,5,20...
		 */
		int h = max(hops -1,3);
		//struct in6_addr allnodes = IPV6_ALL_MANET_NODES;
		//
		/* We are not the destination
		 * ----------------------------*/
		fprintf(stdout,"INF-RREQ: Check for active RT to rreq->dest.\n");
		
		/* lookup for route to destination in the routing table 
		 * ---------------------------------------------------*/
		if ((fwd_rte = rt_get(&rreq->dest, 128, RT_ACTIVE)) == NULL)
		{
			noRT = 1;
			if (xdbg) fprintf(stdout, "No route in AODV-RT to %x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&rreq->dest));
		}
		else
		{
			/* MI: rt_get checks if rt is valid, but it happens that 
			 * route is valid here but almost ready to expire, and 
			 * it will be invalidated by the time is needed 
			 * and the message will inherite this short lifetime 
			 * If we UNICAST the request must give it a chance to 
			 * get there. Should we allow 2*NODE_TRAVERSAL_TIME*fwd_rte->hopcount */
			if (fwd_rte->lifetime <= timems()) //2*NODE_TRAVERSAL_TIME*fwd_rte->hopcount)
			{
				oldRT = 1;
				if (xdbg) fprintf(stdout, "Route to %x:%x:%x:%x:%x:%x:%x:%x is old (life=%lu)\n",
					NIP6ADDR(&rreq->dest),fwd_rte->lifetime - timems());
			}
		}
		
		if((noRT) || (oldRT))
		{
			struct aodv6_rreq *new_rreq = NULL;

			fprintf(stdout,
				"\tno active route found to rreq->dest %x:%x:%x:%x:%x:%x:%x:%x (hoplim=%d)\n",
					NIP6ADDR(&rreq->dest), hops);

			if (hops < 1)//(hops <= 1) 
			{

				if (xdbg) fprintf(stdout,
						"\tWARNING: Return(0) because Hop=1 so NO FORWARDING.\n");
				return 0;
			}
		
			/* no route, rebroadcast request
			 * --------------------------------
			 * - use own IP as source in IPv6 header
			 * - dest seq max(rreq.seq, rt.seq(dest))
			 * - IPv6 hdr hoplimit decreased by one
			 * - RREQ hopcount incremented by one 
			 */
			fprintf(stdout,
					"INF-RREQ: Reb-cast(1) rt_req to %x:%x:%x:%x:%x:%x:%x:%x with (hops-1)=%d\n",
					 NIP6ADDR(&rreq->dest),hops-1);
			
			new_rreq = create_rreq(flags, rreq->hopcount + 1,
				       ntohl(rreq->rreq_id),
				       &rreq->dest, ntohl(rreq->dest_seq),
				       &rreq->orig, ntohl(rreq->orig_seq)
#ifdef _SECURE_BOOT_
					   , 0
#endif
					   );
			
#ifdef _SECURE_BOOT_
			/* This is a rebroadcast
			 * Copy the original signature 
			 */
			if(xdbg) fprintf(stdout, 
					"\tCopy original signature for rebroadcast 1\n");
			ret = copy_secData(new_rreq, rreq, MSG_RREQ);
			if(ret<0)
			{
				fprintf(stderr, "ERROR: Could not copy signature\n");
				return ret;
			}
#endif
			/* If we have an expired RT try to use hopcount 
			 * so re-bcasting saves time */
			/* hops is the max # of hops allowed */
			if((oldRT) && (fwd_rte != NULL)) //Just to be sure :)
				h = max(hops-1, fwd_rte->hopcount);
			else
				h = ((hops-1) <=1)? 3 : hops-1;
				
			if(xdbg) fprintf (stdout,"Reb-cast with hoplim h = %d\n",h);
			ret = aodv6_rreq_rebroadcast(new_rreq, h, 
						sizeof(struct aodv6_rreq));

			if((ret) & (xdbg))
				fprintf(stdout,"WARNING-%s:%d: rreq_rebroadcast() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
		}  
		else  
		{
			/* Have route (fwd_rte) to rreq->dest 
			 * -----------------------------------*/
			local = rt_is_local(fwd_rte);
			fprintf(stdout,
					"INF-RREQ: Found RT to dest %x:%x:%x:%x:%x:%x:%x:%x \n\twith hopcount=%d and lifetime=%lu\n",
					NIP6ADDR(&rreq->dest) ,fwd_rte->hopcount,fwd_rte->lifetime-timems());
		
			//fprintf(stdout, "DBG-DBG-DBG: rt-seq = %d req-seq = %d\n",
			//		fwd_rte->sequence, ntohl(rreq->dest_seq));
			
			/* reply if ...
			 * -------------*/
			if (fwd_rte->sequence >= ntohl(rreq->dest_seq) && !local) 
			{

#ifndef _SECURE_IP_
				/* If RREQ comes with a dest seq# > seq# from RT (fwd_rte), route is
				 * consider old.
				 * If dest seq# in RT (fwd_rte) >= than dest seq # from the RREQ message
				 * and NOT in Secure Mode, send RREP to Originator
				 */
				struct aodv6_rrep *new_rrep = NULL;
#else
				/* Intermediat nodes DO NOT reply in secure mode */
				struct aodv6_rreq *fwd_rreq = NULL;
#endif

#ifndef _SECURE_IP_ //if _SECURE_BOOT_ is defined => _SECURE_IP_ is defined too
				/* MI-Changed logic here:
				 * Old version:
				 * -------------
				 * 1) send rrep to originator with 
				 * lifetime = fwd_rte->lifetime - timems()
				 * 2) send Gratuitos message (no RT lifetime update)
				 * New version:
				 * -------------
				 * 1) Send first gratuitos msg and, if successful, 
				 * update Rt lifetime to activeRouteLife per RFC
				 * 2) If not successful rebroadcast rreq
				 * 3) send rrep to originator if 1) was successful
				 */
				u32 lifetime = max(activeRouteTimeout,
						(fwd_rte->lifetime - timems()));
				int sendRep = 0;

				if (flags | F_GRATUITOUS) 
				{
					/* MI: Per RFC - Section 6.6 - 
					 * increment hopcount first
					 */				
					struct aodv6_rrep *grat_rrep = NULL;
					grat_rrep = create_rrep(
						rflags, 128, rreq->hopcount + 1, 
						&rreq->orig, ntohl(rreq->orig_seq),
						&rreq->dest, lifetime);
					
					if (xdbg) fprintf(stdout,
							"\t---> RREP-SENT to rreq->dest because of F_GRATUITOUS flags\n");
					
					ret = aodv6_rrep_send(&fwd_rte->nexthop, grat_rrep,
						 sizeof(struct aodv6_rrep));

					if(ret) 
					{
						/* it did not send it successfully *
						 * keep rebroadcasting it then 
						 */
						struct aodv6_rreq *new_rreq = NULL;
						if (xdbg)
							fprintf(stdout,"WARNING-%s:%d: rrep_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
						fprintf(stdout,
							"INF-RREQ: Reb-cast(1) rt_req to %x:%x:%x:%x:%x:%x:%x:%x with (hops-1)=%d\n",
							 NIP6ADDR(&rreq->dest),hops-1);
			
						new_rreq = create_rreq(flags, rreq->hopcount + 1,
				       						ntohl(rreq->rreq_id),
				     						&rreq->dest, ntohl(rreq->dest_seq),
				       						&rreq->orig, ntohl(rreq->orig_seq));
			
						/* If we have an expired RT try to use hopcount 
			 			* so re-bcasting saves time */
						/* hops is the max # of hops allowed */
						if(fwd_rte != NULL) //Just to be sure :)
							h = max(hops-1, fwd_rte->hopcount);
				
						ret = aodv6_rreq_rebroadcast(new_rreq, h, 
									sizeof(struct aodv6_rreq));

						if((ret) & (xdbg))
							fprintf(stdout,
										"WARNING-%s:%d: rreq_rebroadcast() ret=%d\n",
										__FUNCTION__,__LINE__,ret);
						
						if(new_rreq != NULL)
							free(new_rreq);
					} 
					else // Success is 0
					{
						sendRep = 1;
						
						if(grat_rrep != NULL)
							free(grat_rrep);

						/* MI: RFC 3561: a route should be updated anytime
						* it is succesfully used
						* Update lifetime with ACTIVE_ROUTE_TIMEOUT
	 					*/
						if(activeRouteTimeout > ( fwd_rte->lifetime - timems()))
						{
							fwd_rte->lifetime = activeRouteTimeout + timems();
							ret = rt_update(fwd_rte,1,0);

							if(ret)
								if(xdbg) fprintf(stdout,
									"WARNING: Could not update RT to %x:%x:%x:%x:%x:%x:%x:%x\n",
									NIP6ADDR(&rreq->dest));
						}
					}
					
				}

				if(sendRep)
				{
					lifetime = fwd_rte->lifetime - timems();	
					if (xdbg) fprintf(stdout,
							"---> fwd_rte->seq >= rreq->dest_seq && !local.\n");

					/* MI: Per RFC - Section 6.6 - 
					 * increment hopcount first
					 */		
					new_rrep = create_rrep(
						rflags, fwd_rte->plen, fwd_rte->hopcount + 1, //fwd_rte->hopcount,
						&rreq->dest, fwd_rte->sequence,
						&rreq->orig, lifetime);
				
					if(xdbg) fprintf(stdout,
							"\t---> RREP-SENT (no-security) to saddr\n");
				
					ret = aodv6_rrep_send(saddr, new_rrep, 
					 	sizeof(struct aodv6_rrep));
				
					if((ret) & (xdbg))
						fprintf(stdout,"WARNING-%s:%d: rrep_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
				
					if(xdbg) fprintf(stdout,
							"\tAdd precursor to %x:%x:%x:%x:%x:%x:%x:%x rt_entry\n",
												NIP6ADDR(&rreq->dest));
					/* MI: - RFC 3561
					 * The the precursors list of the 
					 * RT to destination must store: 
					 * 1) originator of the request and 
					 * 2) the next hop to which the RREP will be forwarded.
					 * 3) Also the RT to the next hop to destnation must contain the 
					 * next hop to originator as precursor
					 */
					rt_entry_precursor_add(fwd_rte, &rreq->orig);
					/* We only have the wireless interface address not 
					 * the aodv address? 
					 * It should be OK as long as they are one hop away
					 * */
					rt_entry_precursor_add(fwd_rte, saddr);
					/* 3) can not be achieved here because we are using the
					 * wireless interface address as next hop and have no 
					 * RT for neighbors on this addresses*/

					free(new_rrep);
				}
#else
				/* When secure bootstrapping, intermediat nodes should not
				 *  reply; should forward the request to destination
				 */
				fwd_rreq = create_rreq(flags, rreq->hopcount + 1,
					       ntohl(rreq->rreq_id),
					       &rreq->dest, ntohl(rreq->dest_seq),
					       &rreq->orig, ntohl(rreq->orig_seq)
#ifdef _SECURE_BOOT_
						   , 0
#endif
						   );
				
				if (xdbg) fprintf(stdout,"\tCopy original signature for fwd rreq 1\n");
				ret = copy_secData(fwd_rreq, rreq, MSG_RREQ);
				if(ret<0)
				{
					fprintf(stderr,"ERROR: Could not copy signature for fwd req 1!\n");
					return ret;
				}

				ret = aodv6_rreq_fwd(&fwd_rte->nexthop, fwd_rreq, 
						sizeof(struct aodv6_rreq));

				if((ret) & (xdbg))
					fprintf(stdout,"WARNING-%s:%d: rreq_fwd() ret=%d\n",
							__FUNCTION__,__LINE__,ret);
				
				if(!ret)	// 0 is success
				{
					if(xdbg) fprintf(stdout,
							"\tUpdate RT & add precursor (saddr) to this rt_entry\n");

					/* MI: RFC 3561: a route should be updated anytime
					 * it is succesfully used
					 * Update lifetime with ACTIVE_ROUTE_TIMEOUT
	 				 */
					printf("HERE3 Before update lifetime %lu\n",fwd_rte->lifetime-timems());
					fwd_rte->lifetime = max(fwd_rte->lifetime,(activeRouteTimeout + timems()));
					printf("HERE3 update lifetime to %lu\n",fwd_rte->lifetime-timems());
					ret = rt_update(fwd_rte,1,0);

					if(ret)
						if(xdbg) fprintf(stdout,
								"WARNING: Could not update RT to %x:%x:%x:%x:%x:%x:%x:%x\n",
								NIP6ADDR(&rreq->dest));
					/* MI: - RFC 3561
					 * The the precursors list of the 
					 * RT to destination must store: 
					 * 1) originator of the request and 
					 * 2) the next hop to which the RREP will be forwarded.
					 * 3) Also the RT to the next hop to destnation must contain the 
					 * next hop to originator as precursor
					 */
					rt_entry_precursor_add(fwd_rte, &rreq->orig);
					/* We only have the wireless interface address not 
					 * the aodv address? 
					 * It should be OK as long as they are one hop away
					 * */
					rt_entry_precursor_add(fwd_rte, saddr);
					/* 3) can not be achieved here because we are using the
					 * wireless interface address as next hop and have no 
					 * RT for neighbors on this addresses*/
				}
				free(fwd_rreq);
#endif
			} else if (local) 
			{
				/* - we are the destination 
				 * -------------------------*/
				struct aodv6_rrep *new_rrep = NULL;
				u32 seq = 0;
				
				if(xdbg) fprintf(stdout,"INF-RREQ: ---> We are the destination.\n");
				
				seq = upd_get_seqnum(ntohl(rreq->dest_seq));

				new_rrep = create_rrep(
					rflags, fwd_rte->plen, 
					0, &rreq->dest, seq,
					&rreq->orig, 
					(myRouteTimeout + 
			 			(rreq->hopcount*(NODE_TRAVERSAL_TIME + NODE_SEC_VERIFY_TIME)*2)) //MY_ROUTE_TIMEOUT <- RTable has revRoute +2*hoplimit*traveTime
#ifdef _SECURE_BOOT_
					, 1
#endif
					);

				fprintf(stdout,"INFO-RREQ: RREP sent to rreq->dest via saddr\n");

				ret = aodv6_rrep_send(saddr, new_rrep,
						 sizeof(struct aodv6_rrep));

				if((ret) & (xdbg))
					fprintf(stdout,"WARNING-%s:%d: rrep_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
				free(new_rrep);
			} else {
				/* - Not local but dest seq in rt is less than
				 * the rreq's dest seq
				 */
				/* IMPORTANT NOTE: If seq # for dest in your RT < seq # from RREQ, 
				 * keep rebroadcasting, your RT is too old! */
				/* - is it OK to rebroadcast??  */
				struct aodv6_rreq *fwd_rreq = NULL;

				if(xdbg) fprintf(stdout,"\tfwd_rte->seq < rreq->dest_seq => rebroadcast.\n");
			
				if (hops == 1)
				{
					if(xdbg) fprintf(stdout,"\tBUT hop = 1. return 0!\n");
					return 0;
				}
				fwd_rreq = create_rreq(
					flags, rreq->hopcount + 1, ntohl(rreq->rreq_id), 
					&rreq->dest, max(ntohl(rreq->dest_seq), 
						fwd_rte->sequence),
					&rreq->orig, ntohl(rreq->orig_seq)
#ifdef _SECURE_BOOT_
					, 0
#endif
					);

#ifdef _SECURE_BOOT_
				if(xdbg) fprintf(stdout, "\tCopy original signature for fwd rreq 2\n");
				ret = copy_secData(fwd_rreq, rreq, MSG_RREQ);
if(ret<0)
				{
					fprintf(stderr,"ERROR: Could not copy signature for fwd req 2!\n");
					return ret;
				}
#endif

				fprintf(stdout,
					"INF-RREQ: Reb-cast(2) rt_req to %x:%x:%x:%x:%x:%x:%x:%x with (hops-1)=%d\n",
					 NIP6ADDR(&rreq->dest),hops-1);

				/* Rt is old, but can save some time on 
				 * re-bcasting with hops = 1 */ 
				if(fwd_rte != NULL)
					h = max(h, fwd_rte->hopcount);
				
				ret = aodv6_rreq_rebroadcast(fwd_rreq, h, 
						sizeof(struct aodv6_rreq));

				
				if((ret) & (xdbg))
					fprintf(stdout,"WARNING-%s:%d: rreq_rebroadcast() ret=%d\n",
								__FUNCTION__,__LINE__,ret);
				
				free(fwd_rreq);
			}
		}
		free(fwd_rte);
	}

	//xdbg = 0;

	//fprintf(stdout,"\n---> TIMER rreq : %lu - %lu = %lu\n", timems(), t1, timems()-t1);

	return ret;
}

/**
 * aodv6_rerr_recv - handle Route Error message
 * @msg: message data buffer
 * @len: length of data in buffer
 *
 * Parses RERR message and invalidates unreachable destinations in the
 * routing table.
 **/
int aodv6_rerr_recv(struct msghdr *msg, size_t len)
{
	struct aodv6_rerr *rerr = (struct aodv6_rerr *)msg->msg_iov->iov_base;
	struct destination *dst = (struct destination *)(rerr + 1);
	struct sockaddr_in6 *ss_addr = (struct sockaddr_in6 *)msg->msg_name;
	struct in6_addr *saddr = (struct in6_addr *)&ss_addr->sin6_addr;
	struct rt_entry *rte = NULL;
	int dests = rerr->destcount;
	u32 flags = 0;

	/* See draft section 6.11 (iii) */
	if (msg == NULL) return -ENODATA;

	if (dests < 1)
		return -1;

	if (len != sizeof(struct aodv6_rerr) + dests * 
	    sizeof(struct destination)) {
		/* invalid message length */
		return -E_MSGLEN;
	}

	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	fprintf(stdout, "recv-errmsg(%d) from %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(saddr));
	
	ulog(L_RECV_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_RECV_INFO, "recvmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(saddr));
	ulog_if(L_RECV_PACKET) {
		dump_mux(ulogd, rerr, MSG_RERR);
	}

	if (rerr->flags & RERR_F_NODELETE) flags |= F_NODELETE;

	/* if nodelete flag is set, node is doing local repair */
	
	for ( ; dests > 0; dests--) {
		rte = rt_get(&dst->addr, 128, RT_ANY);
		if (rte == NULL) {
			/* destination not in active use */
			dst++;
			continue;
		}
		if (!ipv6_addr_cmp(saddr, &rte->nexthop)) {
			/* source node is next hop on this destination */
			rte->sequence = ntohl(dst->seq);
			rte->hopcount = INFINITE_HOPS;
			
		}
		dst++;
	}	
	
	return 0;
}

/**
 * aodv6_rrep_ack_recv - handle Route Reply Acknowledgment message
 * @msg: message data buffer
 * @len: length of data in buffer
 *
 * Parse RREP-ACK message and note ack received.
 **/
int aodv6_rrep_ack_recv(struct msghdr *msg, size_t len)
{
	struct aodv6_rerr *ack = (struct aodv6_rerr *)msg->msg_iov->iov_base;
	struct sockaddr_in6 *ss_addr = (struct sockaddr_in6 *)msg->msg_name;
	struct in6_addr *saddr = (struct in6_addr *)&ss_addr->sin6_addr;

	if (msg == NULL) return -ENODATA;

	if (len != sizeof(struct aodv6_rrep_ack)) {
		/* invalid message length */
		return -E_MSGLEN;
	}

	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	fprintf(stdout, "recv-ackmsg(%d) from %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(saddr));
	
	ulog(L_RECV_INFO, "=== %s ===\n",__FUNCTION__);
	ulog(L_RECV_INFO, "recvmsg(%d) %x:%x:%x:%x:%x:%x:%x:%x\n",
	     len, NIP6ADDR(saddr));
	ulog_if(L_RECV_PACKET) {
		dump_mux(ulogd, ack, MSG_RREP_ACK);
	}

	return 0;
}

struct blacklist {
	struct in6_addr addr;
	struct blacklist *next;
	struct blacklist *prev;
};

struct blacklist *blacklist_set = NULL;

/**
 * aodv6_blacklist - add node to blacklist
 * @addr: node address
 *
 * Adds node to blacklist.  RREQs from blacklisted nodes are ignored.
 * Nodes are automatically removed from blacklist after
 * %BLACKLIST_TIMEOUT period.
 **/
void aodv6_blacklist(struct in6_addr *addr)
{
	struct blacklist *new, *node;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	if (addr == NULL) return;

	new = malloc(sizeof(struct blacklist));
	if (new == NULL) {
		/* out of memory */
		perror(__FUNCTION__);
		return;
	}
	memset(new, 0, sizeof(struct blacklist));

	//Set the address
	ipv6_addr_copy(&new->addr, addr);
	
	/* MI: Why adding it secind node, and not last. 
	 * If blacklist_set is NULL/empty this algo brakes.
	 * Change it... 
	new->prev = blacklist_set;
	new->next = blacklist_set->next;
	blacklist_set->next->prev = new;
	blacklist_set->next = new;
	*/

	//The "new" node will be the last one
	new->next = NULL;
		
	//Search for last node in the blacklist_set
	if(blacklist_set == NULL)
	{
		blacklist_set = new;
		new->prev = NULL;
	} else {
		node = blacklist_set;

		while(node->next != NULL)
			node = node->next;

		//Got last node in the linked list, add the new node here
		node->next = new;
		new->prev = node;
	}
	
	fprintf(stdout,"-> Added to blacklist: %x:%x:%x:%x:%x:%x:%x:%x\n",
	     NIP6ADDR(addr));
	
	//ulog(L_INFO, __FUNCTION__ ": add %x:%x:%x:%x:%x:%x:%x:%x\n",
	//     NIP6ADDR(addr));
}

/**
 * aodv6_blacklist_check - check whether node is blacklisted
 * @addr: node address
 *
 * Checks if @addr is in this node's blacklist set.  Returns 0 if
 * @addr is blacklisted and negative if not.
 **/
int aodv6_blacklist_check(struct in6_addr *addr)
{
	struct blacklist *cur = blacklist_set;

	if (addr == NULL) return -EINVAL;

	// MI- change it. Blacklist_set could be NULL/empty
	//for (cur = blacklist_set->next; cur != NULL; cur = cur->next) 
	for( cur = blacklist_set; cur != NULL; cur = cur->next)
	{
		if(ipv6_addr_cmp(&cur->addr, addr) == 0) 
		{
			//Found the address on the blacklist
			cur = NULL;
			return 0;
		} 
	}
	
	//The adress is not on the blacklist
	return -1;
}

/**
 * aodv6_blacklist_remove - remove node from blacklist set
 * @addr: node address
 *
 * Removes address from blacklist set after %BLACKLIST_TIMEOUT.
 * Called by alarm timer.
 **/
void aodv6_blacklist_remove(struct in6_addr *addr)
{
	struct blacklist *cur;

	if (addr == NULL) return;

	//MI -blacklist can be NULL/empty. Change the logic
	//for (cur = blacklist_set->next; cur != NULL; cur = cur->next) {
	for (cur = blacklist_set; cur != NULL; cur = cur->next)
	{
		if (ipv6_addr_cmp(&cur->addr, addr) == 0) 
		{
			cur->prev->next = cur->next;
			cur->next->prev = cur->prev;
			free(cur);
			cur = NULL;
			return;
		}
	}
	cur = NULL;
	//ulog(L_INFO, __FUNCTION__ ": remove %x:%x:%x:%x:%x:%x:%x:%x\n",
	//     NIP6ADDR(addr));
}

/**
 * request_route - activate Route Request process
 * @requester: packet id for route requesting packet
 * @addr: destination for which route is requested
 * @type: type of connection required
 *
 * Requests a route for destination @addr.  Handles resends if
 * necessary.  Caller can select %DIREC_UNI or %DIREC_BI for
 * connection type.  @type controls whether the G-bit is set for the
 * RREQ.  If anything goes, select @type %DIREC_ANY.  This function is
 * called when node generates a packet to a destination for which it
 * has no route.
 **/
int request_route(unsigned long requester, struct in6_addr *addr, int type)
{
	struct rt_entry *rte = NULL;
	struct aodv6_rreq *rreq = NULL;
	u32 flags = 0, seq = 0;
	u8 hoplimit = TTL_START;
	struct in6_addr allnodes = IPV6_ALL_MANET_NODES;
	struct in6_addr my_ip;
	struct resend_data *data;
	int timeout = 0, ret = 0;
	unsigned long t1 = timems();

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (addr == NULL) return -EINVAL;

	if (config_getaddr(SCOPE_GLOBAL, &my_ip) < 0) {
		/* failed to get own ip */
	}

	/* see if we have a stale entry for this destination */
	rte = rt_get(addr, 128, RT_ANY);
	if (rte) {
		/* if we do, we use last valid hopcount as initial
                   hoplimit and last known sequence number */
		hoplimit = rte->last_hopcount + TTL_INCREMENT;
		if (hoplimit > TTL_THRESHOLD)
			hoplimit = NET_DIAMETER;
		seq = rte->sequence;
	}

	/* if caller requests bidirectional communication, set
           Gratuitous flag in RREQ */
	if (type == DIREC_BI)
		flags |= F_GRATUITOUS;

#ifdef _SECURE_BOOT_
	fprintf(stdout,"\t--->\tCreate a secured/signed RREQ\n");
#else
	fprintf(stdout,"\t--->\tCreate a non-secured RREQ\n");
#endif
				
	rreq = create_rreq(flags, 0, get_rreq_id(), addr,
			seq, &my_ip, inc_get_seqnum()
			   //seq, &my_ip, get_seqnum(),
#ifdef _SECURE_BOOT_
			    , 1 
#endif
			);

	data = (struct resend_data *)malloc(sizeof(struct resend_data));
	if (data == NULL) {
		perror(__FUNCTION__);
		return -1;
	}

	timeout = 2 * hoplimit * NODE_TRAVERSAL_TIME;

	data->timeout = timeout;
	data->key = requester;
	data->hoplimit = hoplimit;
	data->resend = 0;
	data->rreq = rreq;
	data->len = sizeof(*rreq);

	//ulog(L_INFO, __FUNCTION__ ": %x:%x:%x:%x:%x:%x:%x:%x\n",
	//     NIP6ADDR(addr));

	ret = aodv6_rreq_send(&allnodes, hoplimit, rreq, sizeof(*rreq));
	if((ret) & (xdbg))
		fprintf(stdout,"WARNING-%s:%d: rreq_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);

	/* wait for 2 * TTL * NODE_TRAVERSAL_TIME milliseconds, retry
           RREQ_RETRIES times if no RREP */
	add_alarm(timeout, requester, ROUTE_DISC, aodv6_rreq_resend, data);

	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);

	fprintf(stdout,"\n---> TIMER: req_route: %lu\n", timems() - t1);
	return 0;
}

/**
 * send_hello - start sending Hello messages
 * @arg: Hello Interval extension value
 *
 * Starts sending Hello messages.  If @arg is set to a value greater
 * than zero, a Hello Interval extension is added to the Hello
 * message.
 **/
void send_hello(void *arg)
{
	int *setint = malloc(sizeof(int));

	*setint = *(int *)arg;
	aodv6_hello_send(*setint);
	add_alarm( helloInterval, //HELLO_INTERVAL, //expires in ms milliseconds
				0xffffffff, 	//key
				HELLO_SEND, 	//type
				send_hello, 	//task
				setint);		//data
	if(xdbg)
		fprintf(stdout, "HELLO-sent @ interval=%d\n", helloInterval);
}

/**
 * aodv6_hello_start - start periodic Hello send
 * @ifindex: interface to broadcast Hellos from
 * @interval: interval between consequent Hellos
 * @setopt: interval option value
 *
 * Starts sending periodic Hello messages.  Interface @ifindex is used
 * for sending out the messages.  If @setopt is greater than zero,
 * Hellos include a Hello Interval extension with the set value.
 **/
void aodv6_hello_start(int ifindex, int interval, int setopt)
{
	int *setint = malloc(sizeof(int));

	*setint = setopt;
	add_alarm(deletePeriod, 0xffffffff, HELLO_SEND, send_hello, setint);
	if(xdbg) fprintf(stdout,"HELLO-start-send @ %d interval\n",deletePeriod);
}

/**
 * local_repair - try repair link locally
 * @dst: unreachable destination
 * @seq: destination sequence number
 * @hopcount: last known hopcount
 * @sender: originator address
 *
 * When node detects link break during communication, it may attempt
 * local repair instead of sending a RERR to originator.  Local repair
 * RREQ is sent only once.  Sends a Route Request for @dst trying to
 * repair route locally.  See draft section 6.12.
 **/
int local_repair(struct in6_addr *dst, unsigned long seq,
		 int hopcount, struct in6_addr *sender)
{
	/* No local repair allowed in Secure Mode ! */
	
#ifndef _SECURE_BOOT_
	struct aodv6_rreq *rreq;
	struct rt_entry *rte;
	struct in6_addr allnodes = IPV6_ALL_MANET_NODES;
	struct in6_addr my_ip, *dest;
	int hops, hoplimit;
	unsigned long timeout;
	int ret = 0;

	if (hopcount >= MAX_REPAIR_TTL)
		return 0;

	rte = rt_get(sender, 128, RT_ACTIVE);
	if (rte != NULL)
		hops = rte->hopcount;
	else
		hops = 1;

	/* MIN_REPAIR_TTL = last hopcount */
	hoplimit = max(hopcount, hops / 2) + LOCAL_ADD_TTL;

	if (config_getaddr(SCOPE_GLOBAL, &my_ip) < 0) {
		/* failed to get own ip */
		return -1;
	}
	dest = malloc(sizeof(struct in6_addr));
	if (dest == NULL)
		return -ENOMEM;
	memset(dest, 0, sizeof(struct in6_addr));
	memcpy(dest, dst, sizeof(struct in6_addr));

	rreq = create_rreq(0, 0, get_rreq_id(), dest,
			seq, &my_ip, inc_get_seqnum()
#ifdef _SECURE_BOOT_
			, 0
#endif
			);

	ret = aodv6_rreq_send(&allnodes, hoplimit, rreq, sizeof(*rreq));
	if((ret) & (xdbg))
		fprintf(stdout,"WARNING-%s:%d: rreq_send() ret=%d\n",
								__FUNCTION__,__LINE__,ret);

	/* we use timeout period defined in expanding ring search
           rather than NET_TRAVERSAL_TIME */
	timeout = 2 * hoplimit * NODE_TRAVERSAL_TIME;
	add_alarm(timeout, in6_hash(dest), ROUTE_DISC, aodv6_rreq_repair, (void *)dest);

#endif

	return 0;
}

int beacon_received(struct in6_addr *dest)
{
	return 0;
}

/** 
 * link_breakage - notify neighbors of broken link
 * @dst: unreachable neighbor
 *
 * Link breakage detected.  Invalidate all routes using this neighbor
 * as next hop and notify precursors.  See draft section 6.11 (i). 
 **/
int link_breakage(struct in6_addr *dst)
{
	if (dst == NULL) return -EINVAL;
	/* find all routes with dst as next hop */
	rt_invalid_nexthop(dst);

	return 0;
}

/**
 * no_forward_route - send RERR in response to data packet
 * @dst: unreachable destination
 * @sender: node this packet came from
 *
 * Received data packet from @sender to @dst, but have no valid
 * forward route to @dst.  Sends a RERR to @sender.  See draft section
 * 6.11 (ii).
 **/
int no_forward_route(struct in6_addr *dst, struct in6_addr *sender)
{
	struct aodv6_rerr *rerr;
	struct dest *dsts;
	struct rt_entry *rte = NULL;
#ifndef _SECURE_IP_
	int do_local_repair = 0, hops = 0;
#endif
	int flags = 0, seq = 0;
	
	if (dst == NULL) {
		/* no destination */
		return -EINVAL;
	}
	if (sender == NULL) return -EINVAL;

	
	/* if we have a non-active route to dst, increment seq by one */
	rte = rt_get(dst, 128, RT_ANY);
	if (rte != NULL) 
	{
		/* Seq # must be incremented before trying local repair,
		 * otherwise if in case of S-->A--->B-[broken]->D
		 * when B tries local repair without increasing seq#, A will reply
		 * that it has route and we end up with an loop S-->A<-->B
		 */
		rte->sequence++;
		seq = rte->sequence;
#ifndef _SECURE_IP_
		rte->flags ^= RT_F_LOCAL_REPAIR;
		do_local_repair = rte->flags & RT_F_LOCAL_REPAIR;
		hops = rte->last_hopcount;
		rt_update(rte, 0, 0);
#endif
	}

#ifndef _SECURE_IP_
	/* attempt local repair here */
	if (do_local_repair) {
		return local_repair(dst, seq, hops, sender);
	}
#endif
	
	dsts = (struct dest *)malloc(sizeof(struct dest));
	
	if (dsts == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}

	//MI-FIXED BUG
	memset ( dsts, 0, sizeof(struct dest));

	ipv6_addr_copy(&dsts->s_dest.addr, dst);
	dsts->s_dest.seq = seq;

	//ulog(L_INFO, __FUNCTION__ ": sender = %x:%x:%x:%x:%x:%x:%x:%x\n",
	  //   NIP6ADDR(sender));
	rerr = create_rerr(flags, 1, dsts);
	aodv6_rerr_send(sender, rerr, sizeof(struct aodv6_rerr) + 
			 sizeof(struct destination));

	return 0;
}
