/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/types.h>
#include <fcntl.h>

#include "config.h"
#include "util.h"

/* Release information */
extern const char *description; // = "aodvd (AODV for IPv6 daemon)";
extern const char *version; // = "0.1";
extern const char *author; // = "Written by Antti J. Tuominen modified by Michaela Iorga";

extern int xdbg;
extern void resetHelloParams(int new);

struct nodeconfig Config;

#ifdef _eAODVD_
char *evilAddr = NULL;
char *evilKey = NULL;
char *dropAddr = NULL;
int _dropRnd = 0;
int _alterPkt = 0;
int _eIP = 0;
int _eKey = 0;
RSA *evilRSA = NULL;
unsigned char *evp_exp = NULL;
unsigned int evp_exp_len=0;
unsigned char *evp_mod = NULL;
unsigned int evp_mod_len=0;
#endif

int config_getaddr(int scope, struct in6_addr *addr)
{
	if (scope == SCOPE_GLOBAL) {
		memcpy(addr, &Config.NodeAddrGlobal, sizeof(struct in6_addr));
	} else if (scope == SCOPE_SITE) {
		memcpy(addr, &Config.NodeAddrSite, sizeof(struct in6_addr));
	} else if (scope == SCOPE_HOST) {
		memcpy(addr, &Config.HostAddrGlobal, sizeof(struct in6_addr));
	} else {
		memset(addr, 0, sizeof(struct in6_addr));
		return -1;
	}
	return 0;
}

int config_cmdline(int argc, char **argv)
{
	const char *cmd = *argv;
	char *configfile = NULL;
	char *defconfigfile = default_config_file;
	struct stat cstat;
	int dlevel = -1;

	argv++; argc--;
	while (argc) {
		argc--;
		if (strcmp(*argv, "-v") == 0 || strcmp(*argv, "--version") == 0) 
		{
			fprintf(stderr, "\n%s (version %s)\n  Authors: %s\n", 
					description, version, author);
#ifdef _eAODVD_
			fprintf(stderr, "  NOTE: This version can become malicious,\n\tif appropriate options are used.\n\tUse -h,--help option for more details.\n\n");
#endif
			exit(0);
		} else if (strcmp(*argv, "-h") == 0 || strcmp(*argv, "--help") == 0) 
		{
			fprintf(stdout, "\nUsage: \n\t%s [-h,--help] [-v,--version] [-d,--debug <debug level>] \
					\n\t\t[-C,--configfile <config file name>] \
					\n\t\t[-A,--globaladdr <aodv0 global IP>] \
					\n\t\t[-I,--hostaddr <host IP>] \
					\n\t\t[-a,--siteaddr <aodv0 site IP>]", cmd);
#ifdef _SECURE_IP_
			fprintf(stdout,"\n\t\t[-kl,--keylen <64|128 [bytes]>] ");
#endif
#ifdef _eAODVD_
			fprintf(stdout,	"\n\t\t[-eA,--eviladdr <aodv0 global IP>] \
					\n\t\t[-eK,--evilkey <base 64 encoded RSA public key>] \
					\n\t\t[-sA,--sabotageaddr <IP>] \
					\n\t\t[-alt,--alter <yes/true>] \
					\n\t\t[-rd,--randomdrop <yes/true>] ");
#endif
			fprintf(stdout, "\n\nwhere:\n\t-h,--help\t\tThis help.\n\t-v,--version\t\tVersion and description information.\n");
#ifdef _SECURE_IP_
			fprintf(stdout, "\t-kl,--keylen <value>\tOverrides the default RSA key size (128) in bytes.\n");
#endif
#ifdef _eAODVD_
			fprintf(stdout,"\t-eA,--eviladdr <IP>\tUse this <IP> instead of your secure IP.\n");
			fprintf(stdout,"\t-eK,--evilkey <Key>\tThe RSA public key (base 64 encoded) used to \n\t\t\t\tgenerate the IP given by the -eA option.\n");
			fprintf(stdout,"\t-sA,--sabotageaddr <IP>\tDrop or alter all packets originated by this IP.\n");
			fprintf(stdout,"\t-alt,--alter\t\tAlter packets. \
					\n\t\t\t\tWhen associated with -sA <sIP> option, \
					\n\t\t\t\tonly packet originated by sIP will be \
					\n\t\t\t\taltered. \
					\n\t\t\t\tWhen associated with -rd option\
					\n\t\t\t\tpackets will be altered randomly.\n");
			fprintf(stdout,"\t-rd,--randomdrop\tDrop packets randomly. \
					\n\t\t\t\tWhen associated with -sA <sIP> option, \
					\n\t\t\t\tonly packet originated by sIP will be \
					\n\t\t\t\tdropped randomly.\n");
#endif
			fprintf(stdout, "\n\tSee man page for more information.\n\n");
			exit(0);
		} else if (strcmp(*argv, "-d") == 0 || strcmp(*argv, "--debug") == 0) {
			if (!argc || !strncmp(*(argv + 1), "-", 1)) {
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			dlevel = atoi(*argv);
			if (dlevel == 0 && *argv[0] != '0') {
				fprintf(stderr, "Invalid debug level \"%s\"\n", *argv);
				exit(1);
			}
			Config.DebugLevel = dlevel;
		} else if (strcmp(*argv, "-C") == 0 || strcmp(*argv, "--configfile") == 0) {
			if (!argc || !strncmp(*(argv + 1), "-", 1)) {
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			configfile = malloc(strlen(*argv) + 1);
			strcpy(configfile, *argv);
		} else if (strcmp(*argv, "-f") == 0 || strcmp(*argv, "--logfile") == 0) {
			if (!argc || !strncmp(*(argv + 1), "-", 1)) {
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			strcpy(Config.LogFile, *argv);
		} else if (strcmp(*argv, "-l") == 0 || strcmp(*argv, "--log") == 0) {
			int lv;
			if (!argc || !strncmp(*(argv + 1), "-", 1)) {
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			lv = atoi(*argv);
			if (lv == 0 && *argv[0] != '0') {
				fprintf(stderr, "Invalid log verbosity \"%s\"\n", *argv);
				exit(1);
			}
			if (lv > 0) {
				Config.Log = 1;
				Config.LogVerbosity = lv;
			} else {
				Config.Log = 0;
				Config.LogVerbosity = 0;
			}
		} else if (strcmp(*argv, "-A") == 0 || strcmp(*argv, "--globaladdr") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
		}
#ifdef _SECURE_IP_
		else if (strcmp(*argv, "-kl") == 0 || strcmp(*argv, "--keylen") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) {
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			_RSAkeylen = atoi(*argv);
			if ((_RSAkeylen == 0 && *argv[0] != '0') ||
					( (_RSAkeylen % 64) != 0) || 
					(_RSAkeylen > 128) ) 
			{
				fprintf(stdout, "Unsupported RSA key length \"%s\"\n", *argv);
				fprintf(stdout, "Supported values : 64 bytes and 128 bytes (default value).\n");
				exit(1);
			}
			if (_RSAkeylen <= 0)
			{
				fprintf(stdout, "Invalid key size \"%s\"\n", *argv);
				exit(1);
			}
		}
#endif
#ifdef _eAODVD_		
		else if (strcmp(*argv, "-eA") == 0 || strcmp(*argv, "--eviladdr") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			evilAddr = malloc(strlen(*argv) + 1);
			strcpy(evilAddr, *argv);
			
			/* set _eIP global flag*/
			_eIP = 1;

		} else if (strcmp(*argv, "-eK") == 0 || strcmp(*argv, "--evilkey") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			evilKey = malloc(strlen(*argv) + 1);
			strcpy(evilKey, *argv);

			/* Set _eKey global flag*/
			_eKey = 1; 
			
		} else if (strcmp(*argv, "-sA") == 0 || strcmp(*argv, "--sabotageaddr") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			dropAddr = malloc(strlen(*argv) + 1);
			strcpy(dropAddr, *argv);
			
		} else if (strcmp(*argv, "-alt") == 0 || strcmp(*argv, "--alter") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			if(strcmp(*argv, "yes") == 0 || strcmp(*argv, "true") == 0)
				_alterPkt = 1;
		} else if (strcmp(*argv, "-rd") == 0 || strcmp(*argv, "--randomdrop") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
			if(strcmp(*argv, "yes") == 0 || strcmp(*argv, "true") == 0)
				_dropRnd = 1;
		}	
#endif
		else if (strcmp(*argv, "-I") == 0 || strcmp(*argv, "--hostaddr") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
		} else if (strcmp(*argv, "-a") == 0 || strcmp(*argv, "--siteaddr") == 0) 
		{
			if (!argc || !strncmp(*(argv + 1), "-", 1)) 
			{
				fprintf(stderr, "%s requires value\n", *argv);
				exit(1);
			}
			argv++; argc--;
		} else {
			fprintf(stderr, "Unknown argument \"%s\"\n", *argv);
			exit(1);
		}

		argv++;
	}

	if (configfile == NULL) {
		if (stat(defconfigfile, &cstat) >= 0) {
			configfile = malloc(strlen(defconfigfile));
			strcpy(configfile, defconfigfile);
		}
	} else if (stat(configfile, &cstat) < 0) {
		fprintf(stderr, "fatal: file %s not found\n", configfile);
		exit(1);
	}

	if (configfile)
		config_parse(configfile);

	return 0;
}

void config_parse(char *filename)
{
	FILE *config;
	int linecount = 0;
	int cur_ifi = 0;
#if defined _SECURE_IP_
	int setGlobal =0,setSite=0;
	int klb = 0;
	int fhPub = 0;
	int fhPrv = 0;
#endif

#ifdef _SECURE_IP_
	/* If secure bootsrapping is desired, only 
	 * network IDs are kept and
 	 * node's addresses are overriden
 	 */

	/* Generate RSA key pair */
	key = RSA_new();
	
	if(key == NULL)
	{
		exit(1);
	}
	
	/* Open the files where keys will be dumped */
	if(((fhPub = open("/tmp/aodvd.pub", O_RDWR | O_CREAT | O_TRUNC, 
				S_IREAD | S_IWRITE | S_IRGRP | S_IROTH)) == 0) | 
		((fhPrv = open("/tmp/aodvd.prv", O_RDWR | O_CREAT | O_TRUNC,
					   S_IREAD|S_IWRITE)) == 0))
	{
		fprintf(stderr,"*** ERROR: Could not open/create key files! ***\n");
		exit(1);
	}
	
	/* Generate a key pair */
	memset(key, 0, sizeof( key));

	/* can use RSA_3=3 or RSA_F4=65537 */
	fprintf(stdout,"\n---> Generating RSA key pair of length %d[bits] & p_exp=%d\n",
				_RSAkeylen*8, p_exp);
	key = RSA_generate_key(_RSAkeylen*8, p_exp, NULL, NULL);

	/* check the keys. If returns 0=> problems=> regenerate keys */
	if(!RSA_check_key(key))
	{
		memset(key, 0, sizeof( RSA));
		p_exp = RSA_F4;
		key = RSA_generate_key( _RSAkeylen*8, p_exp, NULL, NULL);
		if(!RSA_check_key(key))
		{
			/* Can't get good keys. Drop it. */
			exit(1);
		}
	}

	/*
	RSA_print_fp(stdout,key,0);
	
 	fprintf(stdout,"PublicKey-1(n) : \n%s\n",BN_bn2hex(key->n));
	fprintf(stdout,"PublicKey-1(e) : \n%s\n",BN_bn2hex(key->e));
	fprintf(stdout,"PrivatKey-1(d) : \n%s\n",BN_bn2hex(key->d));
	fprintf(stdout,"PrivatPrime-1(p) : \n%s\n",BN_bn2hex(key->p));
	fprintf(stdout,"PrivatPrime-1(q) : \n%s\n",BN_bn2hex(key->q));
	fprintf(stdout,"PrivatKey-1(dmp1) : \n%s\n",BN_bn2hex(key->dmp1));
	fprintf(stdout,"PrivatKey-1(dmq1) : \n%s\n",BN_bn2hex(key->dmq1));
	fprintf(stdout,"PrivatKey-1(iqmp) : \n%s\n",BN_bn2hex(key->iqmp));
	*/

	/*	 Store pubKey in a binary format */
	p_mod = (unsigned char *)malloc(_RSAkeylen);

	/*
	 struct rsa_st
	 {
		BIGNUM *n;              // public modulus = p*q
    	BIGNUM *e;              // public exponent, relativly prime with (p-1)*(q-1)
    	BIGNUM *d;              // private exponent
    	BIGNUM *p;              // secret prime factor
    	BIGNUM *q;              // secret prime factor
    	BIGNUM *dmp1;           // d mod (p-1)
    	BIGNUM *dmq1;           // d mod (q-1)
    	BIGNUM *iqmp;           // q^-1 mod p
		// ...
	} RSA ;

	===> IMPORTANT!!!
	If wrong header files are used, on the laptop, 
	n==NULL because it points to "engine"
	and all values are shifted down: e holds n val, 
	d holds e val and so on... 
	*/
	if(key->n != NULL)
	{
		klb = BN2bin(key->n, p_mod, RSA_size(key));
		if((klb > _RSAkeylen) || (*p_mod==0))
		{
			p_mod=(unsigned char*)realloc(p_mod, klb);
			if(( klb < BN2bin(key->n, p_mod, klb)) || (*p_mod == 0))
				exit(1);
		}
	}

	//fprintf(stdout,"\nMODULUS LEN=%d and RSA_size()=%d\n",klb, RSA_size(key));
	p_mod_len = klb;

	/* Dump the keys */
	{
		int i = 0;
		int p_exp_len = 0;
		int pubBufLen;
		unsigned char pubBuf[RSA_KEY_LENGTH + 16];
		unsigned char *encBuf;
		unsigned char *prvBuf;
		unsigned char *cipherPrv;
		unsigned int  kLen;	// length of plain/cipher key
		
		p_exp_len = BN_num_bytes(key->e);
		//Real length
		pubBufLen = _RSAkeylen + 4 + p_exp_len;

		//All buffer
		memset(pubBuf,0,RSA_KEY_LENGTH+16);
		pubBuf[i++] = 0x45;	//tag = 'E'
		pubBuf[i++] = p_exp_len;
		memcpy(&pubBuf[i],&p_exp,p_exp_len);
		i += p_exp_len;
		pubBuf[i++] = 0x4D; 	//tag = 'M'
		pubBuf[i++] = p_mod_len; //_RSAkeylen with max val RSA_KEY_LENGTH;
		memcpy(&pubBuf[i], p_mod, p_mod_len);
		i += p_mod_len;

		if((i > _RSAkeylen + 16) || (i > RSA_KEY_LENGTH + 16))
		{
			fprintf(stderr, "*** ERROR: Wrong data length %d for key size! \n", i-16);
			exit(1);
		}
		
		if(pubBufLen != i)
			pubBufLen = i;

		if(xdbg)
		{
			printf("\npubBuf[%d] = ",pubBufLen);
			printBin(stdout,pubBuf,pubBufLen);
		}

		/* Encode pub key */
		encBuf = base64_encode(pubBuf,pubBufLen);

		/* Write Pub Key: exp and modulus */
		if(xdbg) 
			fprintf(stdout, "\nFileHandle: %d encBuf-pub[%d] : %s\n", 
					fhPub, strlen(encBuf), encBuf);
		write(fhPub, encBuf, strlen(encBuf));

		free(encBuf);
		
		/* Encrypt pvKey 
		 * -------------*/
		/* Use ckLen to get the length */
		/* Encrypt and dump the entire RSA key pair because 
		 * in addition to the privateExponent which should be key->d
		 * we need the primes p,q and, precomputed values dmp1, 
		 * dmq1, ipmq for fast decryption and 
		 * signing operations
		 * (ALSO see IMPORTANT note above!) 
		 */
		kLen =  BN_num_bytes(key->d)+2+ 	//tag = 'D' (0x44)
				BN_num_bytes(key->p)+2+		//tag = 'P' (0x50)
				BN_num_bytes(key->q)+2+		//tag = 'Q' (0x51)
				BN_num_bytes(key->dmp1)+2+	//tag = (0x10)
				BN_num_bytes(key->dmq1)+2+	//tag = (0x11)
				BN_num_bytes(key->iqmp)+2;	//tag = (0x12)
		
		prvBuf = (unsigned char *)malloc(kLen);

		/* Copy d,p,q, etc.. in TLV format */
		if(prvBuf != NULL)
		{
			int l=0,k=0;
			
			/*--- d ---*/
			//printf("key->d : ");
			//BN_print_fp(stdout,key->d);
			//printf("\n");
			l=BN_num_bytes(key->d);
			prvBuf[k] = 0x44;
			k++;
			prvBuf[k] = l;
			k++;
			l = BN_bn2bin(key->d,&prvBuf[k]);
			if(l != prvBuf[k-1])
				printf("WARNNING: - d!\n");
			k += l;
			
			/*--- p ---*/
			//printf("key->p : ");
			//BN_print_fp(stdout,key->p);
			//printf("\n");
			l=BN_num_bytes(key->p);
			prvBuf[k] = 0x50;
			k++;
			prvBuf[k] = l;
			k++;
			l = BN_bn2bin(key->p,&prvBuf[k]);
			if(l != prvBuf[k-1])
				printf("WARNNING: - p!\n");
			k += l;
			
			/*--- q ---*/
			//printf("key->q : ");
			//BN_print_fp(stdout,key->q);
			//printf("\n");
			l=BN_num_bytes(key->q);
			prvBuf[k] = 0x51;
			k++;
			prvBuf[k] = l;
			k++;
			l = BN_bn2bin(key->q,&prvBuf[k]);
			if(l != prvBuf[k-1])
				printf("WARNNING: - q!\n");
			k += l;
			
			/*--- dmp1 ---*/
			//printf("key->dmp1 : ");
			//BN_print_fp(stdout,key->dmp1);
			//printf("\n");
			l=BN_num_bytes(key->dmp1);
			prvBuf[k] = 0x10;
			k++;
			prvBuf[k] = l;
			k++;
			l = BN_bn2bin(key->dmp1,&prvBuf[k]);
			if(l != prvBuf[k-1])
				printf("WARNNING: - dmp1!\n");
			k += l;
			
			/*--- dmq1 ---*/
			//printf("key->dmq1 : ");
			//BN_print_fp(stdout,key->dmq1);
			//printf("\n");
			l=BN_num_bytes(key->dmq1);
			prvBuf[k] = 0x11;
			k++;
			prvBuf[k] = l;
			k++;
			l = BN_bn2bin(key->dmq1,&prvBuf[k]);
			if(l != prvBuf[k-1])
				printf("WARNNING: - dmq1!\n");
			k += l;
			
			/*--- iqmp ---*/
			//printf("key->iqmp : ");
			//BN_print_fp(stdout,key->iqmp);
			//printf("\n");
			l=BN_num_bytes(key->iqmp);
			prvBuf[k] = 0x12;
			k++;
			prvBuf[k] = l;
			k++;
			l = BN_bn2bin(key->iqmp,&prvBuf[k]);
			if(l != prvBuf[k-1])
				printf("WARNNING: - iqmp!\n");
			k += l;
		}
		else
			fprintf(stderr,"ERROR: Can not allocate memory for prvBuf\n");

		//printf("DATA-TO-ENCRYPT:\n");
		//printBin(stdout,prvBuf,kLen);
		//RSA_print_fp(stdout,key,5);
		/* On output, kLen is length of encrypted data ! */
		cipherPrv = DES3_encrypt( prvBuf,&kLen);
		//printf("Size of encrypted kLen = %d\n",kLen);

		/* Encode cipherPrv */
		/* base64_encode mallocs for encBuf */
		encBuf = base64_encode(cipherPrv,kLen);
		
		/* Write Encoded and Encryptef Pv Key */
		write(fhPrv, encBuf, strlen(encBuf));
		
		free(prvBuf);
		/* malloc in DES3_Encrypt() */
		free(cipherPrv);
		/* malloc in base64_encode() */
		free(encBuf);
		
		close(fhPub);
		close(fhPrv);
	}

	//fprintf(stdout,"MAX_SIG: %d\n",RSA_size(key));

#endif
	
	config = fopen(filename, "r");
	if (config == NULL) {
		fprintf(stderr, "Could not open %s for reading\n", filename);
		exit(1);
	}
	while (1) 
	{
		char *ptr;
		char line[100];
		char *name, *value;

		fgets(line, 100, config);
		if (feof(config))
		{
			break;
		}
		
		linecount++;

		if ((ptr = strchr(line, '#')) != NULL) {
			*ptr = '\0';
		}
		if (*line == '\n' || *line == '\0')
			continue;

		if ((ptr = strchr(line, ';')) == NULL) {
			fprintf(stderr, "missing semi-colon on config file line %d\n",
				linecount);
			fclose(config);
			exit(1);
		}
		*ptr = '\0';

		if ((ptr = strchr(line, '=')) == NULL) {
			fprintf(stderr, "invalid syntax on config file line %d\n",
				linecount);
			fclose(config);
			exit(1);
		}
		*ptr = '\0';
		name = line;
		value = ptr + 1;

		if (strcasecmp(name, "DebugLevel") == 0) {
			int new_val = 0;
			if (Config.DebugLevel != -1)
				goto nextopt;
			new_val = atoi(value);
			if (new_val == 0 && *value != '0')
				goto invvalue;
			Config.DebugLevel = new_val;
		} else if (strcasecmp(name, "Log") == 0) {
			if (Config.Log != -1)
				goto nextopt;
			if (strcasecmp(value, "yes") ||
			    strcasecmp(value, "on")) {
				Config.Log = 1;
			} else if (strcasecmp(value, "no") ||
				   strcasecmp(value, "off")) {
				Config.Log = 0;
			} else goto invvalue;
		} else if (strcasecmp(name, "LogVerbosity") == 0) {
			int new_val = 0;
			if (Config.LogVerbosity != -1)
				goto nextopt;
			new_val = atoi(value);
			if (new_val == 0)
				goto invvalue;
			Config.LogVerbosity = new_val;
		} else if (strcasecmp(name, "LogFile") == 0) {
			if (strlen(Config.LogFile) > 0)
				goto nextopt;
			strcpy(Config.LogFile, value);
		} else if (strcasecmp(name, "NodeAddressGlobal") == 0) 
		{
			char *n, *p;
			n = strchr(value, '/');
			if (n == NULL) {
				Config.GlobalPrefixLen = 128;
			} else {
				p = n + 1;
				*n = '\0';
				Config.GlobalPrefixLen = strtoul(p, &p, 10);
				if (*p) Config.GlobalPrefixLen = 128;
			}

#ifndef _SECURE_IP_
			fprintf(stdout, "Using Global Address %s\n",value);
			if (inet_pton(AF_INET6, value, &Config.NodeAddrGlobal) <= 0) 
			{
				fprintf(stderr, "parse error, invalid IPv6 address on config file line %d\n",
					linecount);
				exit(1);
			}
#else
			/* Derive NodeAddrGlobal if network address exists */
			if(config_secIP(SCOPE_GLOBAL, p_mod, klb, value, &Config.NodeAddrGlobal)==0)
			{
				setGlobal = 1;
				//fprintf(stdout,"SecAODV Global Address: %x:%x:%x:%x:%x:%x:%x:%x\n",
				//		NIP6ADDR(&Config.NodeAddrGlobal));
			}
			else
			{
				if(p_mod != NULL)
				free(p_mod);
				exit(1);
			}
#endif
			
		} else if (strcasecmp(name, "NodeAddressSite") == 0) 
		{
			char *n, *p;
			n = strchr(value, '/');
			if (n == NULL) {
				Config.SitePrefixLen = 128;
			} else {
				p = n + 1;
				*n = '\0';
				Config.SitePrefixLen = strtoul(p, &p, 10);
				if (*p) Config.SitePrefixLen = 128;
			}
#ifndef _SECURE_IP_
			fprintf(stdout, "Using Site Address %s\n",value);
			if (inet_pton(AF_INET6, value, &Config.NodeAddrSite) <= 0) 
			{
				fprintf(stderr, "parse error, invalid IPv6 address on config file line %d\n",
					linecount);
				exit(1);
			}
#else
			/* Derive NodeAddrSite if network address exists 
			 */
			if(config_secIP(SCOPE_SITE, p_mod, klb, value, &Config.NodeAddrSite)==0)
			{
				setSite = 1;
				//fprintf(stdout,"SecAODV Site Address: %x:%x:%x:%x:%x:%x:%x:%x\n",
				//	NIP6ADDR(&Config.NodeAddrSite));
			}
#endif
		} else if (strcasecmp(name, "HostAddressGlobal") == 0)
		{
			char *n, *p;
			n = strchr(value, '/');
			if (n == NULL) {
				Config.HostPrefixLen = 128;
			} else {
				p = n + 1;
				*n = '\0';
				Config.HostPrefixLen = strtoul(p, &p, 10);
				if (*p) Config.HostPrefixLen = 128;
			}

			fprintf(stdout, "\tHost Address  = %s\n",value);
			
			if (inet_pton(AF_INET6, value, &Config.HostAddrGlobal) <= 0) 
			{
				fprintf(stderr, "parse error, invalid IPv6 address on config file line %d\n",
					linecount);
				exit(1);
			}
		} else if (strcasecmp(name, "InterfaceList") == 0) 
		{
			char *ifs = malloc(strlen(value));
			strcpy(ifs, value);
			if ((ptr = strchr(ifs, '"')) != NULL) 
			{
				ifs = ptr + 1;
				if ((ptr = strchr(ifs, '"')) == NULL) 
				{
					fprintf(stderr, "mismatching '\"' on config file line %d\n",
						linecount);
					fclose(config);
					exit(1);
				}
				*ptr = '\0';
			}
			if (strcasecmp(ifs, "all") == 0) 
			{
				struct if_nameindex *ifp, *p;
				struct iflist *iface;
				/* scan for all interfaces */
				ifp = if_nameindex();
				for (p = ifp; p->if_index != 0; p++) 
				{
					if (strcmp(p->if_name, "lo")) 
					{
						iface = malloc(sizeof(struct iflist));
						iface->IfIndex = p->if_index;
						strcpy(iface->Name, p->if_name);
						iface->SendHello = 0;
						iface->HelloInterval = 1000;
						Config.InterfaceList[p->if_index] = iface;
					}
				}
				if_freenameindex(ifp);
				goto nextopt;
			}
			ptr = strtok(ifs, " ");
			while (ptr) 
			{
				if(xdbg) fprintf(stdout, "token: %s %s\n", ptr,
					if_nametoindex(ptr) ? "found" : "not present");
				ptr = strtok(NULL, " ");
			}
			//free(ifs); //<- it core dumps! 
		} else if (strcasecmp(name, "Interface") == 0) 
		{
			struct iflist *iface;
			int ifi = 0;
			if(xdbg) fprintf(stdout, "searching for interface %s\n",value);
			if ((ifi = if_nametoindex(value)) != 0) 
			{
				if(xdbg) fprintf(stdout, "found specified interface %s\n",value);
				iface = malloc(sizeof(struct iflist));
				iface->IfIndex = ifi;
				strcpy(iface->Name, value);
				iface->SendHello = 0;
				iface->HelloInterval = 1000;
				Config.InterfaceList[ifi] = iface;
			} else {
				fprintf(stderr, "could not find specified interface %s\n",value);
				exit(1);
			}
			cur_ifi = ifi;
			
		} else if (strcasecmp(name, "SendHello") == 0) {
			if (cur_ifi == 0) 
			{
				fprintf(stderr, "invalid SendHello, no associated interface\n");
			}
			if ((strcasecmp(value, "no") == 0) ||
				   (strcasecmp(value, "off") == 0)) 
			{
				Config.InterfaceList[cur_ifi]->SendHello = 0;
			} else if ((strcasecmp(value, "yes") == 0) ||
			    (strcasecmp(value, "on") == 0)) 
			{
				Config.InterfaceList[cur_ifi]->SendHello = 1;
			}  else goto invvalue;
		} else if (strcasecmp(name, "HelloInterval") == 0) {
			char *p;
			unsigned long new_val;
			if (cur_ifi == 0) 
			{
				fprintf(stderr, "invalid HelloInterval, no associated interface\n");
			}
			p = value;
			new_val = strtol(p, &p, 10);
			if (*p) goto invvalue;
			Config.InterfaceList[cur_ifi]->HelloInterval = new_val;
			resetHelloParams(new_val);
		} else {
			fprintf(stderr, "invalid option [%s] on config file line %d\n",
				name, linecount);
			exit(1);
		}

		goto nextopt;
		
	invvalue:
		fprintf(stderr, "invalid value %s for %s on config file line %d\n",
			value, name, linecount);
		fclose(config);
	
	nextopt:
		if(xdbg) printf("line: %d\n ", linecount);
		;
	
	}
	
	fclose(config);

#if defined _SECURE_IP_	
	/* Run ifconfig */
	if((setGlobal) || (setSite))
	{
		int ret = 0;
		char buff[512];
		ret = system("ifconfig aodv0 down");
		//ret = system("ip link set aodv0 down");
		if(ret>=0)
		{
			//ret=system("ifconfig aodv0 up");
			ret = system("ip link set aodv0 up");
			//IMPORTANT: Make aodv0 the default
			if(ret>=0)
			{
				ret = system("ip -6 r add ::/0 dev aodv0");
				if(ret<0)
				{
					fprintf(stderr,"ERROR: System call error 2\n");
					exit(1);
				}
			}
			else
				fprintf(stderr,"ERROR: System call error 1\n");	
		}
		else
			fprintf(stderr,"ERROR: Interface aodv0 system call error\n");	
		if(ret<0)
		{
			fprintf(stderr,"ERROR: Could not set aodv0 interface\n");
			exit(1);
		}
		
		if(setGlobal)
		{
			char cmd[] = "ifconfig aodv0 inet6 add ";
			int len = strlen(cmd);
			
			//fprintf(stdout,"Adding addr %x:%x:%x:%x:%x:%x:%x:%x/128\n"
			//			,NIP6ADDR(&Config.NodeAddrGlobal));
			memset(buff,0,512);
			memcpy(buff,cmd,len);
			sprintf(&buff[len],"%x:%x:%x:%x:%x:%x:%x:%x/128",
					NIP6ADDR(&Config.NodeAddrGlobal));
			if(system(buff)<0)
			{
				exit(1);
			}
		}

		if(setSite)
		{
			char cmd[] = "ifconfig aodv0 inet6 add ";
			int len = strlen(cmd);
			
			//fprintf(stdout,"Adding addr %x:%x:%x:%x:%x:%x:%x:%x/128\n"
			//		,NIP6ADDR(&Config.NodeAddrSite));
			memset(buff,0,512);
			memcpy(buff,cmd,len);
			sprintf(&buff[len],"%x:%x:%x:%x:%x:%x:%x:%x/128",
					NIP6ADDR(&Config.NodeAddrSite));
			if(system(buff)<0)
			{
				exit(1);
			}
			
		}

	}
#endif

#ifdef _eAODVD_
	/* Decode the evilPubKey and set the evilRSA toinclude
	 * it in the messages */
	if(_eKey)
	{
		unsigned char decod[256];
		int len=0,k=0;

		memset(decod,0,256);
		len = base64_decode(evilKey,decod);

		if(xdbg) 
		{
			fprintf(stdout,"DECODED DATA [%d]\n",len);
			printBin(stdout,decod,len);
		}

		while(k<len)
		{
			if(decod[k] == 'E')
			{
				evp_exp_len = decod[++k];
				if(evp_exp_len >0)
				{
					evp_exp = (unsigned char *)malloc(evp_exp_len);
					memcpy(evp_exp,&decod[++k],evp_exp_len);
					k += evp_exp_len;
				}
			}

			if(decod[k] == 'M')
			{
				evp_mod_len = decod[++k];
				if(evp_mod_len > 0)
				{
					evp_mod = (unsigned char *)malloc(evp_mod_len);
					memcpy(evp_mod,&decod[++k],evp_mod_len);
					k += evp_mod_len;
				}
			}
		}
		
		if(xdbg) fprintf(stdout,"\n--->Constructing the evilRSA\n");
		evilRSA = RSA_new();

		evilRSA->e = BN_bin2bn(evp_exp, evp_exp_len, evilRSA->e);
		evilRSA->n = BN_bin2bn(evp_mod, evp_mod_len, evilRSA->n);
		
		if(xdbg)
		{
			BN_print_fp(stdout,evilRSA->e);
			fprintf(stdout,"\n");
			BN_print_fp(stdout,evilRSA->n);
		}

		if(xdbg) fprintf(stdout,"\n--->evilRSA done\n");
	}

#endif

}
