/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <pthread.h>

#include "config.h"
#include "ip6_queue.h"
#include "aodv6.h"


#include "messages.h"
#include "util.h"
#include "rt_table.h"
#include "kinterface.h"
#include "pktqueue.h"
#include "tqueue.h"
#include "popup.h"

#define DEBUG 1
#define DUMP 1

const char *ident = "AODVv6";
int sockfd = -1;

extern int xdbg;
FILE *ulogd;

/**
 * if_mc_group - join or leave multicast group
 * @sock: socket
 * @ifindex: interface to join/leave
 * @mc_addr: multicast address
 * @cmd: join/leave command
 *
 * Join/leave multicast group on interface.  cmd must be either
 * IPV6_JOIN_GROUP or IPV6_LEAVE_GROUP.  Also turns off local
 * multicast loopback. 
 **/
int if_mc_group(int sock, int ifindex, struct in6_addr *mc_addr, int cmd)
{
	unsigned int val = 0;
	struct ipv6_mreq mreq;
	int ret = 0;

	memset(&mreq, 0, sizeof(mreq));
	mreq.ipv6mr_interface = ifindex;
	ipv6_addr_copy(&mreq.ipv6mr_multiaddr, mc_addr);

	ret = setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 
			 &val, sizeof(int));

	if (ret < 0) return ret;

	return setsockopt(sock, IPPROTO_IPV6, cmd, &mreq, sizeof(mreq));
}

/*
 * Callback function for daemon_start signal(SIGCLD)
 */
void sig_child(int unused)
{
	int pid;
	union wait status;

	while ((pid = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0);
}

void sig_hup(int signal)
{
	rt_flush();
	return;
}

/*
 * Detach from any tty.
 */
void daemon_start(int ignsigcld)
{
	register int childpid, fd;

	if (getppid() == 1) goto out;

#ifdef SIGTTOU
	signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
	signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#endif

	if ((childpid = fork()) < 0)
		fprintf(stderr, "can't fork first child\n");
	else if (childpid > 0)
	{
		fprintf(stdout, "Forked child proccess with pid %d\n",childpid);
		exit(0);
	}

	if (setpgrp() == -1)
		fprintf(stderr, "can't change process group\n");
	if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
		ioctl(fd, TIOCNOTTY, (char *)NULL);
		close(fd);
	}

 out:
	for (fd = 0; fd < NOFILE; fd++) close(fd);
	errno = 0;

	chdir("/tmp");
	umask(0);

	if (ignsigcld) {
#ifdef SIGTSTP
		signal(SIGCLD, sig_child);
#else
		signal(SIGCLD, SIG_IGN);
#endif
	}
}


struct nl_handle nlh;

int nl_ip6_queue(void *arg)
{
	return ip6q_listen((struct nl_handle *)arg);
}

static int rtnl_rtmsg_listen(void *arg)
{
	struct rtnl_handle *rtnl;
	struct nlmsghdr *h;
	struct sockaddr_nl nl;
	struct iovec iov;
	char   buf[8192];
	struct msghdr msg = {
		(void*)&nl, sizeof(nl),
		&iov,	1,
		NULL,	0,
		0
	};
	int rcvd;

	rtnl = (struct rtnl_handle *)arg;

	memset(&nl, 0, sizeof(nl));
	nl.nl_family = AF_NETLINK;
	nl.nl_pid = 0;
	nl.nl_groups = 0;


	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	while (1) {
		rcvd = recvmsg(rtnl->fd, &msg, 0);

		if (rcvd < 0) continue;
		if (rcvd == 0) {
			perror("netlink");
			return -1;
		}
		for (h = (struct nlmsghdr *)buf; rcvd >= sizeof(struct nlmsghdr); ) {
			int err;
			int len = h->nlmsg_len;
			int l = len - sizeof(struct nlmsghdr);

			if (l < 0 || len > rcvd) {
				return -1;
			}

			err = rtnl_rt_sync_listener(&nl, h);
			if (err < 0)
				return err;

			rcvd -= NLMSG_ALIGN(len);
			h = (struct nlmsghdr *)((char *)h + NLMSG_ALIGN(len));
		}
		if (msg.msg_flags & MSG_TRUNC) {
			ulog(L_WARN | L_RTSYNC, "Message truncated\n");
			continue;
		}
		if (rcvd) {
			return -1;
		}
	}
}

/*
 * The main program
 */
#include "radix.h"
void rt_dump(void);
extern node_t *rtbl;
int main (int argc, char **argv)
#if 0 //---> 
{
	struct in6_addr nx = IPV6_ALL_MANET_NODES;
	struct in6_addr ip = { { { 0x21,0x03,0,0,0,0,0,0,0,0,0,0,0,0,0,0x1 } } };
	struct in6_addr ip2 = { { { 0x22,0x03,0,0,0,0,0,0,0,0,0,0,0,0,0,0x1 } } };
	struct in6_addr ip3 = { { { 0x23,0x03,0,0,0,0,0,0x1,0,0,0,0,0,0,0,0x2 } } };
	struct in6_addr ip4 = { { { 0x24,0x03,0,0,0,0,0,0x2,0,0,0,0,0,0,0,0x3 } } };
	struct in6_addr key;
	struct in6_addr key2;
	struct in6_addr key3;
	struct in6_addr key4;
	int ret = 0;
	//node_t *tree; 
	//node_t *n;
	struct rt_entry entry;
	struct rt_entry *found = NULL;

	memcpy(&entry.destination,&ip, sizeof(struct in6_addr));
	entry.plen = 128;
	entry.sequence = 0;
	entry.hopcount = entry.last_hopcount = 0;
	memcpy(&entry.nexthop,&nx, sizeof(struct in6_addr));
	entry.precursors = NULL;
	entry.lifetime = 1000;
	entry.flags = 0;
	entry.hello_interval = 2000;

	xdbg = 1;
	rtbl = create_tree();
	if(rtbl == NULL)
	{
		printf("Can't create tree\n");
		return -1;
	}

	ntoh_in6_addr(&ip, &key);
	ret = node_add(rtbl, &key, entry.plen, &entry, sizeof(struct rt_entry));

	rt_dump();

	printf("\n---> Add entry 2:\n"); 
	ntoh_in6_addr(&ip2, &key2);
	memcpy(&entry.destination,&ip2, sizeof(struct in6_addr));
	ret = node_add(rtbl, &key2, entry.plen, &entry, sizeof(struct rt_entry));

	rt_dump();

	printf("\n---> Add entry 3:\n"); 
	ntoh_in6_addr(&ip3, &key3);
	memcpy(&entry.destination,&ip3, sizeof(struct in6_addr));
	ret = node_add(rtbl, &key3, entry.plen, &entry, sizeof(struct rt_entry));

	rt_dump();
	printf("\n---> Add entry 4:\n"); 
	ntoh_in6_addr(&ip4, &key4);
	memcpy(&entry.destination,&ip4, sizeof(struct in6_addr));
	ret = node_add(rtbl, &key4, entry.plen, &entry, sizeof(struct rt_entry));

	rt_dump();

	found = (struct rt_entry *)node_lookup(rtbl, &key3, 128, 1);
	if(found)
		rt_entry_print(stdout,found);
	return ret;
}
#endif //<----
{
	struct sockaddr_in6 daemon;
	struct in6_addr if_ip = IN6ADDR_ANY_INIT;
	struct in6_addr mc_group = IPV6_ALL_MANET_NODES;
	struct rtnl_handle rtlh;
	pthread_t rt_listener;
	pthread_t nl_listener;
	pthread_attr_t attrs;
	struct if_nameindex *ifs, *p;
	struct sockaddr_in6 saddr;
	struct addr_list *list;
	int setint = 0, len, val = 1;
	int ret = 0;

	/* Set up the socket for popupUI */
	SetUI();

	/* Reset the sequence number */
	set_seqnum(0);

	memset(&Config, 0, sizeof(Config));
	Config.DebugLevel = -1;
	Config.Log = -1;
	Config.LogVerbosity = -1;

	if (config_cmdline(argc, argv) < 0)
		return -1;

	if (Config.Log <= 0) {
		Config.Log = 0;
		Config.LogVerbosity = 0;
	}

	if (Config.Log && strlen(Config.LogFile) == 0)
		sprintf(Config.LogFile, "/tmp/aodvd-%d.log", getpid());

	if (Config.DebugLevel == -1)
		Config.DebugLevel = 0;

	/* Autodetection, use all available interfaces */
	fflush(stderr);
	fprintf(stdout, "\tDebug level   = %d\n", Config.DebugLevel);
	fprintf(stdout, "\tLogging       = %s", Config.Log ? "on" : "off");
	if (Config.Log && Config.LogVerbosity) {
		fprintf(stdout, "\n\tVerbosity     = %d\n\tLogfile       = %s\n", 
		       Config.LogVerbosity, Config.LogFile);
		ulogd = fopen(Config.LogFile, "a");
		if (ulogd == NULL) {
			fprintf(stderr, "Could not open log file %s\n",
				Config.LogFile);
			exit(1);
		}
	} else {
		ulogd = NULL;
		fprintf(stdout, "\n");
	}

	/** For DEBUGGING only **/
#if 0
	Config.Log = 1;
	Config.LogVerbosity = L_EVERYTHING;
	ulogd = stdout;
#endif
	/****/

	/* dlog(DBGINFO, "Found interface(s): "); */
	ifs = if_nameindex();
	for (p = ifs; p->if_index != 0; p++) 
	{
		if (strcmp(p->if_name, "lo")) 
		{
			if (Config.InterfaceList[p->if_index]) 
			{
				fprintf(stdout, "\tInterface     = %s\n", Config.InterfaceList[p->if_index]->Name);
				fprintf(stdout, "\tSendHello     = %s\n", 
					(Config.InterfaceList[p->if_index]->SendHello == 1) ? "yes" : "no");
				fprintf(stdout, "\tHelloInterval = %ld\n\n", 
					Config.InterfaceList[p->if_index]->HelloInterval);
				//Why if setint!=0, RREP-RECV stops receiving Hellos???
				//setint = Config.InterfaceList[p->if_index]->HelloInterval;
			}
			/* dlog(DBGINFO, "%s ", p->if_name); */
		}
	}
	/* dlog(DBGINFO, "\n"); */
	p = NULL;

	openlog(ident, LOG_PID, LOG_USER);

	taskqueue_init();
	/* Initialize internal routing table */
	if (rt_init() < 0) {
		fprintf(stderr,": ERROR: Could not open netlink socket\n");
		syslog(LOG_ERR, "fatal: could not open netlink socket\n");
		exit(1);
	}

	/* register listener for routing table changes */
	fprintf(stdout,"---> Register listener for RT changes\n");
	memset(&rtlh, 0, sizeof(rtlh));

	rtlh.fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (rtlh.fd < 0) {
		return -1;
	}

	rtlh.local.nl_family = AF_NETLINK;
	if (bind(rtlh.fd, (struct sockaddr*)&rtlh.local, 
		 sizeof(struct sockaddr_nl)) < 0) {
		perror("bind");
		return -1;
	}

	len = sizeof(struct sockaddr_nl);
	if (getsockname(rtlh.fd, (struct sockaddr*)&rtlh.local, &len) < 0) {
		return -1;
	}

	pthread_attr_init(&attrs);
	pthread_attr_setdetachstate(&attrs, PTHREAD_CREATE_DETACHED);

	if (pthread_create(&rt_listener, &attrs, (void *)rtnl_rtmsg_listen, &rtlh)) {
		perror("pthread_create");
		syslog(LOG_ERR, "fatal: could not create rtnl_rtmsg_listen thread");
		exit(errno);
	}

	/* Add all local addresses */
	if (ifs && ifs->if_name) 
	{
		struct addr_list *a;
		char buf[INET6_ADDRSTRLEN];

		//dlog(DBGINFO, "Adding local address to routing table:\n");
		list = rtnl_local_addrs();
		for (a = list; a; a = a->next) 
		{
			if (rt_set_local(a) < 0) {
				syslog(LOG_ERR, "Could not set local "
					"address in routing table\n");
				exit(1);
			}
			if (xdbg) 
				fprintf(stdout, ": Add local addr %s to rt_table (ifindex=%d)\n", 
					inet_ntop(AF_INET6, &a->addr, buf, sizeof(buf)),
					a->ifindex);
			else
				fprintf(stdout, ": Add local addr %s to rt_table\n", 
					inet_ntop(AF_INET6, &a->addr, buf, sizeof(buf)));
			
			//dlog(DBGINFO, " %d: %s\n", a->ifindex, 
			//       inet_ntop(AF_INET6, &a->addr, buf, sizeof(buf)));
		}

	} else {
		syslog(LOG_ERR, "No interfaces found!\n");
		exit(-1);
	}

	if(xdbg) dlog(DBGINFO, "Using global address %x:%x:%x:%x:%x:%x:%x:%x\n", 
	       NIP6ADDR(&Config.NodeAddrGlobal));

	if_freenameindex(ifs);
	ifs = NULL;

	/* --- Initialize AODV socket --- */
	if (xdbg) fprintf(stdout, "Initializing AODV socket (%s:%d)\n",
					__FILE__,__LINE__);
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin6_family = AF_INET6;

	if ((sockfd = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP)) < 0 ) {
		//perror("socket");
		//syslog(LOG_ERR, "fatal: could not open datagram socket");
		if(xdbg) fprintf(stderr,"FATAL: could not open datagram socket\n");
		exit(errno);
	}

	memset(&daemon, 0, sizeof(daemon));
	daemon.sin6_family = AF_INET6;
	memcpy(&daemon.sin6_addr, &if_ip, sizeof(struct in6_addr));
	daemon.sin6_port = htons(AODVPORT);

	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (void *)&val,
                       sizeof val) < 0) {
		//perror("setsockopt");
		//syslog(LOG_ERR, "warning-1: could not setsockopt()");
		if(xdbg) fprintf(stderr,"ERROR 1:could not setsockopt()\n");
        }

	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &val, sizeof val) < 0) {
		//perror("setsockopt");
		//syslog(LOG_ERR, "warning-2: could not setsockopt()");
		if(xdbg) fprintf(stderr,"ERROR 2:could not setsockopt()\n");
	}

	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_HOPLIMIT, &val, sizeof val) < 0) {
		//perror("setsockopt");
		//syslog(LOG_ERR, "warning-3: could not setsockopt()");
		if(xdbg) fprintf(stderr,"ERROR 3:could not setsockopt()\n");
	}

	if (list) 
	{
		struct addr_list *a;
		for (a = list; a; a = a->next) 
		{
			if_mc_group(sockfd, a->ifindex, &mc_group, IPV6_JOIN_GROUP);
		}
	}

	if (bind(sockfd, (struct sockaddr *) &daemon, sizeof(daemon)) < 0) 
	{
#ifdef DEBUG_AODV
		switch (errno) 
		{
		case EBADF:
			syslog(LOG_ERR, "sockfd is not a valid descriptor");
			break;
		case EADDRNOTAVAIL:
			syslog(LOG_ERR, "address is not available");
			break;
		case EADDRINUSE:
			syslog(LOG_ERR, "address already in use");
			break;
		case EINVAL:
			syslog(LOG_ERR, "socket already bound to an address");
			break;
		case EACCES:
			syslog(LOG_ERR, "reserved port, user is not super-user");
			break;
		case ENOTSOCK:
			syslog(LOG_ERR, "sockfd is a descriptor for a file, not a socket");
			break;
		}
#endif
		//perror("bind");
		//syslog(LOG_ERR, "fatal: unable to bind to socket");
		fprintf(stderr,"FATAL: unable to bind to socket\n");
		closelog();
		close(sockfd);
		exit(errno);
	}


	/* Initialize userspace packet queue and task queue */
	if(xdbg) fprintf(stdout,"Initializing userspace task queue\n");
	ret = taskqueue_init();
	if(xdbg) fprintf(stdout,"Initializing userspace packet queue\n");
	/* returns always 0... */
	ret = pktqueue_init();

	/* Initialize netlink interface for queued packets and start
         * listener thread */
	if(xdbg) fprintf(stdout,"Initializing netlink interface\n");
	if (ip6q_init(&nlh, IP6Q_COPY_PACKET) < 0) 
	{
		//perror("ip6q_init");
		//syslog(LOG_ERR, "fatal: could not open netlink socket");
		fprintf(stderr,"FATAL:  could not open netlink socket\n");
		if (stat("/proc/net/ip6_queue", NULL) == -1 && errno == ENOENT) 
		{
			syslog(LOG_ERR, "ip6_queue.o is not loaded");
		}
		closelog();
		close(sockfd);
		exit(errno);
	}
	pthread_attr_init(&attrs);
	pthread_attr_setdetachstate(&attrs, PTHREAD_CREATE_DETACHED);
	
	if (pthread_create(&nl_listener, &attrs, (void *)nl_ip6_queue, &nlh)) 
	{
		//perror("pthread_create");
		//syslog(LOG_ERR, "fatal: could not create nl_listener thread");
		fprintf(stderr,"FATAL: could not create nl_listener thread\n");
		close(sockfd);
		exit(errno);
	}

	/* we should re-initialize if SIGHUP is received */
	signal(SIGHUP, sig_hup);
	/* we might want to do some cleanup if SIGTERM is received */
	signal(SIGTERM, sig_hup);

	/* detach from tty unless debug_level higher than 0 */
	if (Config.DebugLevel == 0)
	{
		daemon_start(1);
		fprintf(stdout, "AODVD started with DebugLevel 0 \n");
	}
	//syslog(LOG_INFO, "started");
	fprintf(stdout,"---> AODV Daemon - started...\n");

	aodv6_hello_start(0, HELLO_SEND, setint);

	/* main packet receive loop */

	fprintf(stdout,"---> Listening for received packets...\n");
	
	for (;;) {
		int rcvd;
		unsigned char buffer[IP_MAXPACKET];
		unsigned char msgbuffer[1024];
		struct iovec iov = { (void *)&buffer, IP_MAXPACKET };
		struct sockaddr_in6 addr;
		struct msghdr msg = {
			(void *)&addr, sizeof(addr),
			&iov, 1, 
			(void *)&msgbuffer, sizeof(msgbuffer), 
			0
		};
		unsigned char *data = NULL;
		int rval = 0;

		rcvd = recvmsg(sockfd, &msg, 0);
		data = (unsigned char *)msg.msg_iov->iov_base;

		/* multiplex AODV messages */
		switch (data[0]) {
		case MSG_RREQ:
			if(xdbg) fprintf(stdout,"%s: -- Recv RREQ\n",__FUNCTION__);
			rval = aodv6_rreq_recv(&msg, rcvd);
			break;
		case MSG_RREP:
			if(xdbg) fprintf(stdout,"%s: -- Recv RREP\n",__FUNCTION__);
			rval = aodv6_rrep_recv(&msg, rcvd);
			break;
		case MSG_RERR:
			if(xdbg) fprintf(stdout,"%s: -- Recv RERR\n",__FUNCTION__);
			rval = aodv6_rerr_recv(&msg, rcvd);
			break;
		case MSG_RREP_ACK:
			if(xdbg) fprintf(stdout,"%s: -- Recv RREP_ACK\n",__FUNCTION__);
			rval = aodv6_rrep_ack_recv(&msg, rcvd);
			break;
		default:
			/* unknown message */
			if(xdbg) fprintf(stdout,"%s: -- Recv UNKNOWN %d\n",__FUNCTION__, data[0]);
			rval = -E_UNKMSG;
		}
		if (rval == 0)
			continue;

		switch (rval) {
		case -E_MSGLEN:
			ulog(L_WARN | L_RECV, "Invalid message length.\n");
			break;
		case -E_UNKMSG:
			ulog(L_WARN | L_RECV, "Received unknown packet.\n");
			break;
		default:
			ulog(L_WARN | L_RECV, "Internal error (%d).\n", rval);
		}
	}

	close(sockfd);
	closelog();
	
#ifdef _SECURE_IP_
	if(key!=NULL)
		RSA_free(key);
#endif
	
	/* Close socket for PopupUI */
	ResetUI();
	
	return 0;
}
