/*
 *  AODV v6 Netfilter Module
 *
 *  Author:
 *  Antti Tuominen     <ajtuomin@tml.hut.fi>
 *
 *  $Id: ip6_nf_aodv.c,v 1.2 2004/05/06 04:08:02 michaela Exp $
 */

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv6.h>
#include <net/ipv6.h>
#include <net/ip6_fib.h>

//#define DEBUG_AODV 1

#define ACTIVE_ROUTE_TIMEOUT 4000 /* 3000 */
#define ELAPSE 2
#define DELETE_PERIOD ELAPSE * ACTIVE_ROUTE_TIMEOUT

int aodv6_rt_timeout = ACTIVE_ROUTE_TIMEOUT;
int aodv6_rt_stop_update = DELETE_PERIOD;

#if defined(MODULE) && LINUX_VERSION_CODE > 0x20115
MODULE_AUTHOR("Antti Tuominen <ajtuomin@tml.hut.fi>");
MODULE_DESCRIPTION("AODV for IPv6 Kernel Module");
MODULE_PARM(aodv6_rt_timeout, "i");
MODULE_LICENSE("GPL");
#endif

#define MODULE_ID "ip6_nf_aodv6"

#define SIOCVDEVSETINFO SIOCDEVPRIVATE
#define SIOCVDEVGETINFO (SIOCDEVPRIVATE + 1)

#define AODV_RREQ 1
#define AODV_RERR 2

#define NIP6ADDR(addr) \
        ntohs(((u16 *)addr)[0]), \
        ntohs(((u16 *)addr)[1]), \
        ntohs(((u16 *)addr)[2]), \
        ntohs(((u16 *)addr)[3]), \
        ntohs(((u16 *)addr)[4]), \
        ntohs(((u16 *)addr)[5]), \
        ntohs(((u16 *)addr)[6]), \
        ntohs(((u16 *)addr)[7])

#ifdef DEBUG_AODV
#define DEBUG_FN printk(KERN_DEBUG "ip6_nf_aodv6: === " __FUNCTION__ " ===\n")
#define DEBUG(x...) printk(KERN_DEBUG ## x)
#else
#define DEBUG_FN
#define DEBUG(x...)
#endif

struct aodv6_vdev_priv {
	struct net_device_stats stats;
	struct net_device *realdev;
};

/**
 * ms_to_jiffies - convert milliseconds to jiffies
 * @ms: milliseconds
 *
 * Converts time given in milliseconds into the kernel internal time
 * unit jiffies.
 **/
static inline unsigned long ms_to_jiffies(unsigned long ms)
{
	return ((ms * HZ) / 1000) + 1;
}

static inline void dump_skb(struct sk_buff *skb)
{
#ifdef DEBUG_AODV

	int s=0;
			
	if(skb)
	{
		printk(KERN_DEBUG "Len of actual data skb->data_len: %d\n",
							skb->data_len);
		printk(KERN_DEBUG "Buffer size skb->truesize: %d\n",skb->truesize);
		if(skb->data_len) {
			//Do we want to print skb->data?
			printk(KERN_DEBUG "TBD: Print skb->data\n");
		}
		
		s = (sizeof(struct ipv6hdr)<256)?sizeof(struct ipv6hdr):256;
		
		printk(KERN_DEBUG "skb->nh : nh.ipv6h size=%d\n", s);
		if(skb->nh.raw != NULL) 
		{
			struct ipv6hdr *ip6h;
			
			//Parse it as ipv6 header
			ip6h = (struct ipv6hdr *)skb->nh.raw;
			printk(KERN_DEBUG "nh.ipv6h->payload_len = %d \n",
					skb->nh.ipv6h->payload_len);
			printk(KERN_DEBUG "Sender addr %x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&skb->nh.ipv6h->saddr));
			printk(KERN_DEBUG "Destination addr %x:%x:%x:%x:%x:%x:%x:%x\n",
					NIP6ADDR(&skb->nh.ipv6h->daddr));
			printk(KERN_DEBUG "Hop Limit %d\n", skb->nh.ipv6h->hop_limit);
		}
		else
			printk(KERN_DEBUG "NULL nh.raw pointer?.\n");			

		s = (sizeof(struct ethhdr)<256) ? sizeof(struct ethhdr) : 256;
		printk(KERN_DEBUG "skb->mac : mac.ethernet size=%d\n", s );
		/*
		if(skb->mac.raw != NULL) {
			printk(KERN_DEBUG "mac.raw - OK\n");
		} else
			printk(KERN_DEBUG "mac.raw - NULL\n");
		*/
	}
#endif
}

/*****************************************************************************
 *
 * Dummy network interface functions
 *
 *****************************************************************************/

/*
 * In order to get skb's build even if there is no route to the
 * destination, we register a dummy network device aodv0 that always
 * holds routes to all destinations.  This way we trick
 * ip6_route_output() and later identify the packets to be queued to
 * user space.
 */

/**
 * aodv6_vdev_open - net_device open callback
 * @dev: net_device to open
 *
 * Called when someone calls net_device->open().
 **/
static int aodv6_vdev_open(struct net_device *dev)
{
	MOD_INC_USE_COUNT;
	DEBUG_FN;
	return 0;
}

/**
 * aodv6_vdev_stop - net_device stop callback
 * @dev: net_device to stop
 *
 * Called when net_device->stop is called.
 **/
static int aodv6_vdev_stop(struct net_device *dev)
{
	MOD_DEC_USE_COUNT;
	DEBUG_FN;
	return 0;
}

/**
 * aodv6_vdev_ioctl - handler for private ioctl's
 * @dev: net_device we are serving
 * @ifr: ioctl ifreq data pointer
 * @cmd: ioctl command
 *
 * When user calls ioctl with @cmd higher than %SIOCDEVPRIVATE
 * net_device->do_ioctl() gets called.  This is a simple handler for
 * those.  Ioctls are just caught but we do nothing.
 **/
static int aodv6_vdev_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	int retval;

	DEBUG_FN;
	switch (cmd) {
	case SIOCVDEVGETINFO:
		retval = 0;
		break;
	case SIOCVDEVSETINFO:
		retval = 0;
		break;
	default:
		retval = -EOPNOTSUPP;
	}
	return retval;
}

/**
 * aodv6_vdev_stats - return net_device_stats
 * @dev: net_device we are serving
 *
 * Net_device maintains some statistics in a &net_device_stats
 * structure.  Calling net_device->stats() returns a pointer to this
 * structure.  Returns device dependent statistics.
 **/
static struct net_device_stats *aodv6_vdev_stats(struct net_device *dev)
{
	DEBUG_FN;
	return &((struct aodv6_vdev_priv *)dev->priv)->stats;
}

/**
 * aodv6_xmit - send packets out of the device
 * @skb: packet to send
 * @dev: net_device to use
 *
 * If someone tries to transmit packets with aodv0 (i.e. calls
 * net_device->hard_start_xmit()) we call icmpv6_send() with
 * %ICMPV6_DEST_UNREACH to get proper error message.
 **/
static int aodv6_vdev_xmit(struct sk_buff *skb, struct net_device *dev)
{
	//IP6_INC_STATS(Ip6OutNoRoutes);
	
	DEBUG_FN;
	dump_skb(skb);
	
	icmpv6_send(skb, ICMPV6_DEST_UNREACH,
		    ICMPV6_ADDR_UNREACH, 0, skb->dev);
	kfree_skb(skb);
	return 0;
}

/**
 * aodv6_vdev_init - initialize net_device
 * @dev: net_device to initialize
 *
 * When a network device is registered net_device->init() gets called.
 * We do all necessary initialization here.
 **/
static int aodv6_vdev_init(struct net_device *dev)
{
	DEBUG_FN;

	dev->priv = kmalloc(sizeof(struct aodv6_vdev_priv), GFP_USER);
	if (!dev->priv) return -ENOMEM;
	memset(dev->priv, 0, sizeof(struct aodv6_vdev_priv));

	ether_setup(dev);

	dev->flags |= IFF_NOARP;

	dev->open = aodv6_vdev_open;
	dev->stop = aodv6_vdev_stop;
	dev->do_ioctl = aodv6_vdev_ioctl;
	dev->get_stats = aodv6_vdev_stats;
	dev->hard_start_xmit = aodv6_vdev_xmit;

	return 0;
}

static struct net_device aodv_vdev = {
	name: "aodv0",
	init: aodv6_vdev_init,
};

/**
 * aodv6_create_vdev - create aodv virtual network device
 *
 * Creates and registers aodv0 virtual network device.  Calling
 * register_netdev() invokes net_device->init().
 **/
static int aodv6_create_vdev(void)
{
	int err;

	DEBUG_FN;

	if ((err = register_netdev(&aodv_vdev))) {
		printk(KERN_WARNING "Error registering aodv_vdev: %i\n", err);
		return err;
	}
	return 0;
}

/**
 * aodv6_destroy_vdev - deregister and free network device
 *
 * When network device is no longer used we call this function.
 * Network device is unregistered and all dynamically allocated memory
 * is freed.
 **/
static void aodv6_destroy_vdev(void)
{
	DEBUG_FN;
	kfree(aodv_vdev.priv);
	unregister_netdev(&aodv_vdev);
}

static int aodv6_notify_listener(int msgtype, struct in6_addr *requester, 
			   struct in6_addr *target)
{
	DEBUG_FN;

	return 0;
}

/*****************************************************************************
 *
 * Netfilter hooks
 *
 *****************************************************************************/

/*
 * We have to register two hooks for intercepting packets with no
 * route: one for locally originated packets and another for packets
 * being forwarded through.
 */

/**
 * aodv6_nf_check_forward - netfilter hook
 * @hooknum: hook number we got called from
 * @skb: pointer to packet data pointer
 * @in: net_device packet came from
 * @out: net_device for going out
 * @okfn: function to call if we are ok
 *
 * Intercepts packets at the %NF_IP6_FORWARD hook.  If packet has
 * aodv0 as its outgoing device, we queue this packet.  In this case
 * node has received a data packet it does not know how to route.
 * Aodvd will get a notification and send a RERR to sender.
 * %NF_ACCEPT is returned to the packet queue so the packet gets
 * passed to aodv0 hard_start_xmit() for proper ICMP error reporting.
 **/
static unsigned int aodv6_nf_check_forward(
	unsigned int hooknum, struct sk_buff **skb, 
	const struct net_device *in, const struct net_device *out,
	int (*okfn) (struct sk_buff *))
{
#ifdef DEBUG_AODV
	struct ipv6hdr *hdr = (*skb) ? (*skb)->nh.ipv6h : NULL;
#endif

	DEBUG_FN;
#ifdef DEBUG_AODV
	printk(KERN_DEBUG "DEV_IN=%s DEV_OUT=%s\n",in->name, out->name);
#endif
	dump_skb(*skb);
	
	if (!strcmp(out->name, "aodv0")) 
	{
#ifdef DEBUG_AODV
		printk(KERN_DEBUG "Invalid forward to %x:%x:%x:%x:%x:%x:%x:%x\n",
		       NIP6ADDR((struct in6_addr *)&hdr->saddr));
#endif
		return NF_QUEUE;
	}

	return NF_ACCEPT;
}

/**
 * aodv6_nf_check_local - netfilter hook
 * @hooknum: hook number we got called from
 * @skb: pointer to packet data pointer
 * @in: net_device packet came from
 * @out: net_device for going out
 * @okfn: function to call if we are ok
 *
 * Intercepts locally originated packets at %NF_IP6_LOCAL_OUT hook.
 * If outgoing device is aodv0, we have generated a packet node
 * doesn't know how to route.  Packet is queued and aodvd is notified.
 * 
 * Aodvd tries to resolve the route and will return %NF_ACCEPT to the
 * queue.  Packets are marked for rerouting by setting %NFC_ALTERED in
 * skb->nfcache for route_me_harder().  If route_me_harder() selects
 * aodv0 as outgoing interface hard_start_xmit() will, again, make
 * sure we get a proper ICMP error.  Otherwise, packet is passed to
 * the real device.
 **/
static unsigned int aodv6_nf_check_local(
	unsigned int hooknum, struct sk_buff **skb, 
	const struct net_device *in, const struct net_device *out,
	int (*okfn) (struct sk_buff *))
{
#ifdef DEBUG_AODV
	struct ipv6hdr *hdr = (*skb) ? (*skb)->nh.ipv6h : NULL;
#endif

	DEBUG_FN;
	dump_skb(*skb);
	
	if (!strcmp(out->name, "aodv0")) {
#ifdef DEBUG_AODV
		printk(KERN_DEBUG "Queueing packet to %x:%x:%x:%x:%x:%x:%x:%x\n",
		       NIP6ADDR((struct in6_addr *)&hdr->daddr));
#endif
		(*skb)->nfcache |= NFC_ALTERED;
		return NF_QUEUE;
	}
	return NF_ACCEPT;
}

/**
 * aodv6_nf_update_route_use - netfilter hook
 * @hooknum: hook number we got called from
 * @skb: pointer to packet data pointer
 * @in: net_device packet came from
 * @out: net_device for going out
 * @okfn: function to call if we are ok
 *
 * Intercepts outgoing packets (both forwarded and locally originated)
 * after routing/rerouting at %NF_IP6_POST_ROUTING hook.  Whenever a
 * data packet is forwarded on a route the lifetime for this route is
 * updated.  Used route's lifetime is set to current time plus
 * @aodv6_rt_timeout (default: %ACTIVE_ROUTE_TIMEOUT) milliseconds.
 * See draft section 8.1.
 **/
static unsigned int aodv6_nf_update_route_use(
	unsigned int hooknum, struct sk_buff **skb, 
	const struct net_device *in, const struct net_device *out,
	int (*okfn) (struct sk_buff *))
{
	struct rt6_info *rt;
	struct sk_buff *pskb = *skb;

	DEBUG_FN;
	dump_skb(pskb);

	if (pskb == NULL) 
	{
#ifdef DEBUG_AODV
		printk("WARNNING: NULL pskb in "__FUNCTION__"\n");
#endif
		return NF_ACCEPT;
	}

	/* NOTE: pskb->dst is a dst_entry :
	struct dst_entry
	{
		struct dst_entry        *next;
		atomic_t		__refcnt;	// client references	
		int			__use;
		struct net_device       *dev;
		int			obsolete;
		int			flags;
	#define DST_HOST		1
		unsigned long		lastuse;
		unsigned long		expires;
		unsigned		mxlock;
		unsigned		pmtu;
		unsigned		window;
		unsigned		rtt;
		unsigned		rttvar;
		unsigned		ssthresh;
		unsigned		cwnd;
		unsigned		advmss;
		unsigned		reordering;
		unsigned long		rate_last;	// rate limiting for ICMP 
		unsigned long		rate_tokens;
		int			error;
		struct neighbour	*neighbour;
		struct hh_cache		*hh;
		int			(*input)(struct sk_buff*);
		int			(*output)(struct sk_buff*);
	#ifdef CONFIG_NET_CLS_ROUTE
		__u32			tclassid;
	#endif
		struct  dst_ops	        *ops;
		char			info[0];
	};

	-AND-
	struct rt6_info
	{
		union {
			struct dst_entry	dst;
			struct rt6_info		*next;
	} u;

	-ALSO-
	#define rt6i_dev			u.dst.dev
	#define rt6i_nexthop			u.dst.neighbour
	#define rt6i_expires			u.dst.expires
	*/

	rt = (struct rt6_info *)pskb->dst;

	/* We might want to do this is only to RTPROT_AODV routes but
	 * RTPROT_* is not implemented for IPv6 routes.
	 */

	/* MI - This approach keeps updating routes that are no longer valid!!!
	 * It should do it only after validating the route
	 */
	if(rt)
	{
		if (rt->rt6i_expires - rt->u.dst.lastuse <= ms_to_jiffies(aodv6_rt_stop_update))
		{		
			rt->rt6i_expires = jiffies + ms_to_jiffies(aodv6_rt_timeout);
#ifdef DEBUG_AODV
			printk("SADDR: %x:%x:%x:%x:%x:%x:%x:%x",NIP6ADDR(&pskb->nh.ipv6h->saddr));
			printk("  LASTUSE: %lu  EXPIRES: %lu",rt->u.dst.lastuse,rt->rt6i_expires);
			printk("  DIFF: %lu DEL_P: %ld\n",
					rt->rt6i_expires - rt->u.dst.lastuse, 
					ms_to_jiffies(aodv6_rt_stop_update));
		} else 
		{
			printk("SADDR: %x:%x:%x:%x:%x:%x:%x:%x",NIP6ADDR(&pskb->nh.ipv6h->saddr));
			printk("  CONTINUE WITHOUT RT UPDATE\n");
#endif		
		}
	}

	return NF_ACCEPT;
}

static struct nf_hook_ops fwd_intercept = {
	{NULL, NULL},
	aodv6_nf_check_forward,
	PF_INET6,
	NF_IP6_FORWARD,
	0
};

static struct nf_hook_ops local_out_intercept = {
	{NULL, NULL},
	aodv6_nf_check_local,
	PF_INET6,
	NF_IP6_LOCAL_OUT,
	0
};

static struct nf_hook_ops post_routing_intercept = {
	{NULL, NULL},
	aodv6_nf_update_route_use,
	PF_INET6,
	NF_IP6_POST_ROUTING,
	0
};


/*****************************************************************************
 *
 * Module stuff
 *
 *****************************************************************************/

int init_module(void)
{
	int ret = 0;
	printk(KERN_INFO "AODV for IPv6 Netfilter Module v0.1\n");
	/* register netfilter hooks */
	nf_register_hook(&fwd_intercept);
	nf_register_hook(&local_out_intercept);
	nf_register_hook(&post_routing_intercept);
	/* finally, create virtual dummy device */
	ret = aodv6_create_vdev();
	if (ret < 0) {
		/* create_vdev() failed, clean up and bail out */
		printk(KERN_ERR "aodv6_create_vdev failed\n");
		nf_unregister_hook(&local_out_intercept);
		nf_unregister_hook(&fwd_intercept);
		nf_unregister_hook(&post_routing_intercept);
	}
	return ret;
}

void cleanup_module(void)
{
	aodv6_destroy_vdev();
	nf_unregister_hook(&local_out_intercept);
	nf_unregister_hook(&fwd_intercept);
	nf_unregister_hook(&post_routing_intercept);
	printk(KERN_INFO "ip6_nf_aodv.o unloaded.\n");
}
