/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <linux/netfilter.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "kinterface.h"
#include "ip6_queue.h"
#include "pktqueue.h"
#include "aodv6.h"
#include "rt_table.h"
#include "util.h"

extern int xtr;
extern int xdbg;

int nlpid;
/**
 * ip6q_init - initialize netlink socket
 * @nlh: netlink socket handle
 * @mode: operation mode for communications
 *
 * Initializes netlink interface.  Selects @mode for passing
 * information to userspace.  @mode can be one of %IP6Q_COPY_NONE,
 * %IP6Q_COPY_META or %IP6Q_COPY_PACKET.  With %IP6Q_COPY_NONE the
 * packet is not copied to userspace at all but only a notification is
 * sent.  %IP6Q_COPY_META copies meta data associated with the packet
 * and %IP6Q_COPY_PACKET copies the whole packet.
 *
 * The ip6_queue uses %NETLINK_IP6_FW interface.
 **/
int ip6q_init(struct nl_handle *nlh, unsigned mode)
{
	int ret = 0;
	int sockaddr_len;
	ip6q_peer_msg_t m;
	struct sockaddr_nl nladdr;
	struct nlmsghdr hdr;
	struct iovec iov[2] = {{ &hdr, sizeof(hdr) }, { &m, sizeof(m) }};
	struct msghdr msg = {
		(void *)&nladdr, sizeof(nladdr),
		iov, 2, NULL, 0, 0
	};

	memset(nlh, 0, sizeof(nlh));

	nlh->fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_IP6_FW);
	if (nlh->fd < 0) {
		perror("Cannot open netlink socket");
		return -1;
	}
	else 
		if (xdbg) fprintf(stdout,"%s:%d: nlh->fd=%d\n",
							__FILE__, __LINE__,nlh->fd);
	

	memset(&nlh->local, 0, sizeof(nlh->local));
	nlh->local.nl_family = AF_NETLINK;

	if (bind(nlh->fd, (struct sockaddr *)&nlh->local, sizeof(nlh->local)) < 0) {
		perror("Cannot bind netlink socket");
		return -1;
	}

	sockaddr_len = sizeof(nlh->local);
	if (getsockname(nlh->fd, (struct sockaddr *)&nlh->local, 
											&sockaddr_len) < 0) 
	{
		perror("Cannot getsockname()");
		return -1;
	}
	nlh->seq = time(NULL);
	if(xdbg) fprintf(stdout,"%s:%d Sequence (nlh->seq): %d\n",
						__FILE__,__LINE__, nlh->seq);
	
	nlpid = getpid();
	if(xdbg) fprintf(stdout, "%s:%d PID(nlpid)=%d\n",
								__FILE__, __LINE__,nlpid);

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	memset(&m, 0, sizeof(m));
	m.msg.mode.value = mode;
	m.msg.mode.range = 0;
	hdr.nlmsg_len = NLMSG_LENGTH(sizeof(m));
	hdr.nlmsg_type = IP6QM_MODE;
	hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_REPLACE;
	hdr.nlmsg_pid = nlpid;
	hdr.nlmsg_seq = ++nlh->seq;
	if(xdbg) fprintf(stdout,"%s:%d Sequence still (nlh->seq): %d\n",
							__FILE__,__LINE__,nlh->seq);
	if(xdbg) fprintf(stdout, "%s:%d PID(nlpid)=%d\n",
							__FILE__,__LINE__,nlpid);
	ret = sendmsg(nlh->fd, &msg, 0);

	return ret;
}

/**
 * ip6q_send_verdict - send verdict to ip6_queue
 * @nlh: netlink socket handle
 * @verdict: verdict for a packet
 * @id: packet id this verdict applies to
 *
 * Send @verdict via the netlink socket to ip6_queue whether packet
 * @id should be dropped (%NF_DROP) or passed (%NF_ACCEPT).
 **/
int ip6q_send_verdict(struct nl_handle *nlh, unsigned int verdict, unsigned long id)
{
	ip6q_peer_msg_t m;
	struct sockaddr_nl nladdr;
	struct nlmsghdr hdr;
	struct iovec iov[2] = {{ &hdr, sizeof(hdr) }, { &m, sizeof(m) }};
	struct msghdr msg = {
		(void *)&nladdr, sizeof(nladdr),
		iov, 2, NULL, 0, 0
	};

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	memset(&m, 0, sizeof(m));
	m.msg.verdict.value = verdict;
	m.msg.verdict.id = id;
	hdr.nlmsg_len = NLMSG_LENGTH(sizeof(m));
	hdr.nlmsg_type = IP6QM_VERDICT;
	hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_REPLACE;
	hdr.nlmsg_pid = nlpid;
	hdr.nlmsg_seq = ++nlh->seq;

	return sendmsg(nlh->fd, &msg, 0);
}

extern struct nl_handle nlh;

void ip6q_accept(unsigned long id)
{
	ip6q_send_verdict(&nlh, NF_ACCEPT, id);
}

void ip6q_drop(unsigned long id)
{
	ip6q_send_verdict(&nlh, NF_DROP, id);
}

/**
 * ip6q_listen - netlink listener
 * @nlh: netlink socket handle
 *
 * Listens on the netlink socket @nlh.  Gets all packets flagged with
 * %NF_QUEUE from ip6_queue.
 **/
int ip6q_listen(struct nl_handle *nlh)
{
	ip6q_packet_msg_t *m;
	struct sockaddr_nl nladdr;
	struct nlmsghdr *hdr;
	struct ip6_hdr *ip6h;
	int len = 0;
	char buf[8192];
	struct iovec iov;
	struct msghdr msg = {
		(void *)&nladdr, sizeof(nladdr),
		&iov, 1, 
		NULL, 0, 
		0
	};

	if(xtr) 
		fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	nladdr.nl_pid = nlpid;
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	while (1) 
	{
		len = recvmsg(nlh->fd, &msg, 0);

		if (len < 0)
			continue;
		if (len == 0)
			return -1;

		hdr = (struct nlmsghdr *)buf;
		m = (ip6q_packet_msg_t *)(hdr + 1);
		ip6h = (struct ip6_hdr *)m->payload;
#ifdef TRACEPKT
		printf("\n\t-----> packet_id : %lu\n", m->packet_id);
		printf("data_len : %d\n", m->data_len);
		printf("src: %x:%x:%x:%x:%x:%x:%x:%x\n",
		       NIP6ADDR(&ip6h->ip6_src));
		printf("dst: %x:%x:%x:%x:%x:%x:%x:%x\n",
		       NIP6ADDR(&ip6h->ip6_dst));
#endif
#ifdef TEST
		ip6q_send_verdict(nlh, NF_DROP, m->packet_id);
#else
		/* if ip6_src own address, start RREQ and queue packet */
		if (rt_is_local_addr(&ip6h->ip6_src)) 
		{
			fprintf(stdout,"\t-----> It is our packet - \n\t\tqueue and call request_route().\n");
			if (pktqueue_insert(&ip6h->ip6_dst, m->packet_id) == 0)
				request_route(m->packet_id, &ip6h->ip6_dst, DIREC_BI);
		} else {
		/* else do RERR and drop packet */
			fprintf(stdout,"\t-----> src is NOT local address -\n");
			fprintf(stdout,"\t\t have no route to dst - send err & drop packet\n");
			no_forward_route(&ip6h->ip6_dst, &ip6h->ip6_src);
			ip6q_send_verdict(nlh, NF_DROP, m->packet_id);
		}
#endif
	}

	if(xtr) 
		fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return 0;
}

/**
 * ip6q_close - close netlink socket
 * @nlh: netlink socket handle
 *
 * Closes open netlink socket @nlh.
 **/
int ip6q_close(struct nl_handle *nlh)
{
	close(nlh->fd);
	return 0;
}

#ifdef TEST
/*
 * Test code.  Compile with -DTEST and -DTRACEPKT to see packets
 * passed to userspace.
 */
struct nl_handle nlh;

int main(int argc, char **argv)
{

	if (ip6q_init(&nlh, IP6Q_COPY_PACKET) < 0) {
		perror("sendmsg");
		return -1;
	}
	ip6q_listen(&nlh);
	return 0;
}
#endif
