/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "messages.h"


/**
 * create_rreq - create Route Request message
 * @flags: flags to set in RREQ
 * @hopcount: old hopcount
 * @rreq_id: RREQ ID
 * @dest: RREQ destination address
 * @dest_seq: destination sequence number
 * @orig: RREQ requester's source address
 * @orig_seq: source sequence number
 *
 * Creates AODV Route Request message.  This function allocates memory
 * and returns a pointer to a ready built Route Request.  Caller must
 * free this memory.  Valid @flags for RREQ are %F_JOIN, %F_REPAIR and
 * %F_GRATUITOUS.  IPv6 addresses must be in network byte-order.
 **/
struct aodv6_rreq *create_rreq(u32 flags, u8 hopcount, u32 rreq_id, 
			       struct in6_addr *dest, u32 dest_seq, 
			       struct in6_addr *orig, u32 orig_seq
#ifdef _SECURE_BOOT_
				   , u8 signIt
#endif
				   )
{
	struct aodv6_rreq *rreq;

	rreq = malloc(sizeof(struct aodv6_rreq));
	if (rreq == NULL)
		return NULL;
	memset(rreq, 0, sizeof(rreq));

	rreq->type = MSG_RREQ;
	rreq->flags |= (flags & F_JOIN) ? RREQ_F_JOIN : 0;
#if defined _SECURE_IP_ || defined _SECURE_BOOT_
	/* In secure mode no local repair is allowed so no 
	 * gratuitous unicast message to destination is needed either */
	rreq->flags |= 0;
	rreq->flags |= 0;
	rreq->flags |= RREQ_F_DEST_ONLY;
#else
	rreq->flags |= (flags & F_REPAIR) ? RREQ_F_REPAIR : 0;
	rreq->flags |= (flags & F_GRATUITOUS) ? RREQ_F_GRATUITOUS : 0;
#endif

	/* If destination seq number is not known, the flag MUST be set */
	if(dest_seq == 0)
		rreq->flags |= RREQ_F_UNKNOWN_SEQ;

	rreq->hopcount = hopcount;
	rreq->rreq_id = htonl(rreq_id);
	rreq->dest_seq = htonl(dest_seq);
	rreq->orig_seq = htonl(orig_seq);
	memcpy(&rreq->dest, dest, sizeof(struct in6_addr));
	memcpy(&rreq->orig, orig, sizeof(struct in6_addr));

#ifdef _SECURE_BOOT_
	if(signIt)
	{
		int ret = sign_msg(rreq, MSG_RREQ);
		
		if(ret<0)
			exit(1);
		
	}
#endif


	return rreq;
}

/**
 * create_rrep - create Route Reply message
 * @flags: flags to set in RREP
 * @prefix_len: prefix length if not host route
 * @hopcount: hops from orig to dest
 * @dest: requested destination address
 * @dest_seq: destination sequence number
 * @orig: requester's address
 * @lifetime: valid lifetime for this route
 * 
 * Creates AODV Route Reply message.  This function allocates memory
 * and returns a pointer to a ready built Route Reply.  Caller must
 * free this memory.  Valid @flags for RREP are %F_REPAIR and %F_ACK.
 * IPv6 addresses must be in network byte-order.
 **/
struct aodv6_rrep *create_rrep(u32 flags, u8 prefix_len, u8 hopcount,
			       struct in6_addr *dest, u32 dest_seq,
			       struct in6_addr *orig, u32 lifetime
#ifdef _SECURE_BOOT_
				   , int signIt
#endif
		)

{
	struct aodv6_rrep *rrep;
	//int xdbg=1; //use global

	rrep = malloc(sizeof(struct aodv6_rrep));
	if (rrep == NULL)
		return NULL;
	memset(rrep, 0, sizeof(struct aodv6_rrep));

	rrep->type = MSG_RREP;
#if defined _SECURE_BOOT_ || defined _SECURE_IP_
	rrep->flags |= 0;
#else
	rrep->flags |= (flags & F_REPAIR) ? RREP_F_REPAIR : 0;
#endif	
	rrep->flags |= (flags & F_ACK) ? RREP_F_ACK : 0;
	//MI- Why is prefix_len changed to 0 when 128?
	//rrep->prefix_len = prefix_len >= 128 ? 0 : prefix_len;
	rrep->prefix_len = prefix_len > 128 ? 0 : prefix_len;
	rrep->hopcount = hopcount;
	rrep->dest_seq = htonl(dest_seq);
	rrep->lifetime = htonl(lifetime);
	memcpy(&rrep->dest, dest, sizeof(struct in6_addr));
	memcpy(&rrep->orig, orig, sizeof(struct in6_addr));

#ifdef _SECURE_BOOT_
	if(signIt) 
	{
		/*
		 *  If *dest == *orig it means we have a hello message.
		 *  If we do not sign them, routing entry is added to RT without
		 *  verifying the IP and application can use it bypassing aodv
		 *  Will sign everything for now and see how slow it is
		 */
		//if(ipv6_addr_cmp(orig, dest))
		{
			int ret = 0;
		
			ret = sign_msg(rrep, MSG_RREP);

			if(ret<0)
			{
				if(xdbg) fprintf(stderr,"ERROR OCCURED DURING SIGNING\n");
				exit(1);
			}
		}
	}
#endif
	
	return rrep;
}

/**
 * create_rerr - create Route Error message.
 * @flags: flags to set in RERR
 * @destcount: number of invalidated destinations
 * @unreach_dest: list of unreachable destinations
 *
 * Creates AODV Route Error message.  This function allocates memory
 * and returns a pointer to a RERR with a list of unreachable
 * destinations.  Caller must free this memory later.  Valid @flags
 * for RERR is %F_NODELETE.  IPv6 addresses must be in network
 * byte-order.
 **/
struct aodv6_rerr *create_rerr(u32 flags, u8 destcount, 
			       struct dest *unreach_dest)
{
	struct aodv6_rerr *rerr;
	struct destination *dp;
	struct dest *dst;

	rerr = malloc(sizeof(struct aodv6_rerr) + 
		      destcount * sizeof(struct destination));
	if (rerr == NULL)
		return NULL;
	memset(rerr, 0, sizeof(struct aodv6_rerr) + 
	       destcount * sizeof(struct destination));
	dp = (struct destination *)(rerr + 1);
	rerr->type = MSG_RERR;
	rerr->flags |= (flags & F_NODELETE) ? RERR_F_NODELETE : 0;
	rerr->destcount = destcount;

	for (dst = unreach_dest; dst; dst = dst->next) {
		memcpy(dp, &dst->s_dest, sizeof(struct destination));
		dp = (struct destination *)(dp + 1);
	}

	return rerr;
}

/**
 * create_rrep_ack - create Route Reply Acknowledgment message
 *
 * Creates AODV Route Reply Acknowledgment message.  Allocates memory
 * and returns a pointer to a RREP-ACK.  Caller must free this memory
 * later.
 **/
struct aodv6_rrep_ack *create_rrep_ack(void)
{
	struct aodv6_rrep_ack *rrep_ack;

	rrep_ack = malloc(sizeof(struct aodv6_rrep_ack));
	if (rrep_ack == NULL)
		return NULL;
	rrep_ack->type = MSG_RREP_ACK;
	rrep_ack->reserved = 0;
	
	return rrep_ack;
}

/**
 * add_ext - add extension to Route Request or Route Reply message
 * @msg: RREQ or RREP to add extension to
 * @len: length of original message
 * @type: extension type
 * @length: length of extension
 * @value: extension data
 *
 * Appends a extension to RREQ or RREP message.  Allocates memory for
 * new message and returns new message length.  Caller must free
 * allocated memory.
 **/
int add_ext(void **msg, size_t len, u8 type, u8 length, u8 *value)
{
	unsigned char *p = *msg;
	struct extension *ext;

	if (p == NULL) return -1;

	p = realloc(p, len + sizeof(ext) + length);

	if (p == NULL) return -1;

	ext = (struct extension *)(p + len);
	ext->type = type;
	ext->length = length;
	memcpy(ext->data, value, length);
	
	return sizeof(struct extension) + len + length;
}

/**
 * dump_rreq - RREQ dump
 * @stream: output stream
 * @arg: message data
 *
 * Prints out Route Request message contained in @arg to output stream
 * specified by @stream.
 **/
void dump_rreq(FILE *out, void *arg)
{
	struct aodv6_rreq *rreq = (struct aodv6_rreq *)arg;

	fprintf(out, "\nType                        : Route Request (%d)\n", rreq->type);
	fprintf(out, "Flags                       : join(%d) repair(%d) gratuitous RREQ(%d)\n",
	       !!(rreq->flags & RREQ_F_JOIN), !!(rreq->flags & RREQ_F_REPAIR),
	       !!(rreq->flags & RREQ_F_GRATUITOUS));
	fprintf(out, "RREQ ID                     : %d\n", ntohl(rreq->rreq_id));
	fprintf(out, "Hop Count                   : %d\n", rreq->hopcount);
	fprintf(out, "Destination IP Address      : ");
	fprintf(out, "%x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&rreq->dest));
	fprintf(out, "Destination Sequence Number : %d\n", ntohl(rreq->dest_seq));
	fprintf(out, "Source IP Address           : ");
	fprintf(out, "%x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&rreq->orig));
	fprintf(out, "Source Sequence Number      : %d\n", ntohl(rreq->orig_seq));

#ifdef _SECURE_IP_ 
	fprintf(out, "Key length                  : %u\n", rreq->keyLen);
#endif
#ifdef _SECURE_BOOT_
	fprintf(out, "Signature length            : %u\n", rreq->signLen);
#endif
	
}

/**
 * dump_rrep - RREP dump
 * @stream: output stream
 * @arg: message data
 *
 * Prints out Route Reply message contained in @arg to output stream
 * specified by @stream.
 **/
void dump_rrep(FILE *out, void *arg)
{
	struct aodv6_rrep *rrep = (struct aodv6_rrep *)arg;

	fprintf(out, "\nType                        : Route Reply (%d)\n", rrep->type);
	fprintf(out, "Flags                       : repair(%d) acknowledgment(%d)\n",
	       !!(rrep->flags & RREP_F_REPAIR), !!(rrep->flags & RREP_F_ACK));
	fprintf(out, "Prefix Length               : %d\n", rrep->prefix_len);
	fprintf(out, "Hop Count                   : %d\n", rrep->hopcount);
	fprintf(out, "Destination IP Address      : ");
	fprintf(out, "%x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&rrep->dest));
	fprintf(out, "Destination Sequence Number : %d\n", ntohl(rrep->dest_seq));
	fprintf(out, "Source IP Address           : ");
	fprintf(out, "%x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&rrep->orig));
	fprintf(out, "Lifetime                    : %d\n", ntohl(rrep->lifetime));
	
#ifdef _SECURE_IP_ 
	fprintf(out, "Key length                  : %u\n", rrep->keyLen);
#endif
#ifdef _SECURE_BOOT_
	fprintf(out, "Signature length            : %u\n", rrep->signLen);
#endif

}

/**
 * dump_rreq - RERR dump
 * @stream: output stream
 * @arg: message data
 *
 * Prints out Route Error message contained in @arg to output stream
 * specified by @stream.
 **/
void dump_rerr(FILE *out, void *arg)
{
	struct aodv6_rerr *rerr = (struct aodv6_rerr *)arg;
	struct destination *dst = (struct destination *)(rerr + 1);
	int i;

	fprintf(out, "Type                        : Route Error (%d)\n", rerr->type);
	fprintf(out, "Flags                       : no delete(%d)",
	       !!(rerr->flags & RERR_F_NODELETE));
	fprintf(out, "Destination Count           : %d\n", rerr->destcount);
	for (i = 0; i < rerr->destcount; i++) {
		fprintf(out, "Unreachable Dest Address[%d] : ", i + 1);
		fprintf(out, "%x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(&dst->addr));
		fprintf(out, "Unreachable Dest SeqNum[%d]  : %d\n", i + 1, 
			ntohl(dst->seq));
		dst = dst + 1;
	}
	
}

/**
 * dump_rrep_ack - RREP-ACK dump
 * @stream: output stream
 * @arg: message data
 *
 * Prints out Route Reply Acknowledgment message contained in @arg to
 * output stream specified by @stream.
 **/
void dump_rrep_ack(FILE *out, void *arg)
{
	struct aodv6_rrep_ack *rrep_ack = (struct aodv6_rrep_ack *)arg;

	fprintf(out, "Type                        : Route Reply Acknowledgment (%d)\n",
	       rrep_ack->type);
}

/**
 * dump_mux - message dump multiplexer
 * @stream: output stream
 * @arg: message data
 * @type: message type
 *
 * Passes message data in @arg to correct dump function according to
 * @type.  Valid types are %MSG_RREQ, %MSG_RREP, %MSG_RERR and
 * %MSG_RREP_ACK.
 **/
void dump_mux(FILE *out, void *arg, int type)
{
	switch (type) {
	case MSG_RREQ:
		dump_rreq(out, arg);
		break;
	case MSG_RREP:
		dump_rrep(out, arg);
		break;
	case MSG_RERR:
		dump_rerr(out, arg);
		break;
	case MSG_RREP_ACK:
		dump_rrep_ack(out, arg);
		break;
	default:
		fprintf(out, "Unknown message type\n");
	}
}
