/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "util.h"
#include "kinterface.h"
#include "tqueue.h"

/*
 * Packets are queued inside kernel.  List of destination and packet
 * ids is maintained in userspace to allow aodvd to notify kernel
 * which packets should be dropped or passed. 
 */
struct pktqueue_hdr {
	struct in6_addr dest;		/* packet destination */
	struct pkt_record *first;	/* first packet */
	struct pkt_record *last;	/* last packet */
	struct pktqueue_hdr *next;
	struct pktqueue_hdr *prev;
};

struct pkt_record {
	unsigned long packet_id;	/* packet id */
	struct pkt_record *next;
};

struct pktqueue_hdr queue_head;

/**
 * pktqueue_init - initialize userspace packet queue
 *
 * Initializes userspace packet queue that keeps track of packets
 * queued by ip6_queue kernel queuing module.
 **/
int pktqueue_init(void)
{
	memset(&queue_head, 0, sizeof(queue_head));
	queue_head.next = queue_head.prev = &queue_head;
	return 0;
}

/**
 * pktqueue_insert - insert packet to queue
 * @dest: IPv6 destination address
 * @packet_id: identification number for the queued packet
 *
 * Insert packet with @packet_id in queue for destination @dest.
 * Packets queue in FIFO order per destination.
 **/
int pktqueue_insert(struct in6_addr *dest, unsigned long packet_id)
{
	struct pktqueue_hdr *q;
	struct pkt_record *new;
	int found = 0;

	for (q = queue_head.next; q != &queue_head; q = q->next) {
		if (ipv6_addr_cmp(dest, &q->dest) == 0) {
			found = 1;
			break;
		}
	}
	if (!found) {
		struct pktqueue_hdr *nh;
		nh = malloc(sizeof(struct pktqueue_hdr));
		if (nh == NULL) {
			ulog(L_ERR | L_MEM, "Failed to allocate memory.");
			return -ENOMEM;
		}
		memcpy(&nh->dest, dest, sizeof(struct in6_addr));
		nh->next = q;
		nh->prev = q->prev;
		nh->first = nh->last = NULL;
		q->prev->next = nh;
		q->prev = nh;
		q = nh;
	}

	new = malloc(sizeof(struct pkt_record));
	if (new == NULL) {
		ulog(L_ERR | L_MEM, "Failed to allocate memory.");
		return -ENOMEM;
	}

	new->packet_id = packet_id;
	new->next = NULL;

	if (q->last == NULL) {
		q->first = q->last = new;
	} else {
		q->last->next = new;
		q->last = new;
	}

	return found;
}

/**
 * pktqueue_remove - remove all packets for destination
 * @dest: IPv6 destination address
 *
 * Remove all packets for @dest and send verdict to ip6_queue for each
 * removed packet ID.
 **/
int pktqueue_remove(struct in6_addr *dest)
{
	struct pktqueue_hdr *q;

	for (q = queue_head.next; q != &queue_head; q = q->next) 
	{
		if (ipv6_addr_cmp(&q->dest, dest) == 0) 
		{
			struct pkt_record *p;

			q->prev->next = q->next;
			q->next->prev = q->prev;
			q->next = q->prev = NULL;
			p = q->first;
			clr_alarm(p->packet_id, ROUTE_DISC);
			while (p) {
				q->first = p->next;
				p->next = NULL;
				ip6q_accept(p->packet_id);
				free(p);
				p = q->first;
			}
			p = NULL;
			q->first = q->last = NULL;
			free(q);
			q = NULL;

			return 0;
		}
	}
	return -1;
}

/**
 * pktqueue_apply - apply arbitrary function to packets
 * @dest: IPv6 destination address
 * @func: function to apply
 *
 * Applies given func() for all packets queued for @dest.  @func must
 * be of type void func(unsigned long packet_id).
 **/
int pktqueue_apply(struct in6_addr *dest, void (*func)(unsigned long packet_id))
{
	struct pktqueue_hdr *q;

	for (q = queue_head.next; q != &queue_head; q = q->next) {
		if (ipv6_addr_cmp(&q->dest, dest) == 0) {
			struct pkt_record *p;
			p = q->first;
			while (p) {
				func(p->packet_id);
				p = p->next;
			}
			return 0;
		}
	}
	return -1;
}
