/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "radix.h"
#include "util.h"
#include "rt_table.h"



#ifdef DEBUG_AODV
int xtr=1;
int xdbg=1;
#else
int xtr=0;
int xdbg=0;
#endif


/*
 * setbit - set selected bit in key
 * @key: pointer to a key
 * @bit: bit to set
 *
 * Set bit at offset @bit from @key.
 **/
inline u_int32_t setbit(void *key, int bit)
{
	u_int32_t mask = 1 << (32 - bit % 32);
	u_int32_t ret = 0;
	int offset = (bit - 1) / 32;
	u_int32_t *akey = (u_int32_t *)key;

	if (key == NULL) return 0;

	ret = *(akey + offset) & mask;

	return (ret) ? 1 : 0; 
}

/**
 * node_alloc - allocate new tree node
 * @p: parent node for new node
 *
 * Allocates and initializes new tree &node.  Sets @parent to @p and
 * @left, @right and @data to %NULL.
 **/
inline node_t *node_alloc(node_t *p)
{
	node_t *n;

	n = malloc(sizeof(node_t));
	if (n == NULL) {
		perror(__FUNCTION__);
		return NULL;
	}

	n->parent = p;
	n->left = NULL;
	n->right = NULL;
	n->data = NULL;

	return n;
}

/**
 * node_free - free tree node
 * @tree: tree root
 * @n: node to free
 *
 * Frees node @n.  If @n is not a leaf or still has associated data,
 * returns %NULL.  Otherwise, frees tree node @n and returns its
 * parent.
 **/
inline node_t *node_free(node_t *tree, node_t *n)
{
	node_t *parent;

	if (tree == NULL || n == NULL || n == tree)
		return NULL;
	if (n->left || n->right || n->data)
		return NULL;
	parent = n->parent;
	free(n);

	return parent;
}

/**
 * create_tree - create new tree root
 *
 * Allocates a new tree root node of type node_t.  If allocation
 * fails, returns %NULL.  Otherwise, returns new root.  This just
 * calls node_alloc() with NULL parent.
 **/
node_t *create_tree(void)
{
	return node_alloc(NULL);
}

/**
 * tree_walker - walk through radix tree
 * @tree: tree root
 * @func: function to apply for each data entry
 * @data: extra data for func()
 *
 * Walks through the whole given radix tree.  Applies @func to every
 * data entry.  @func returns 0 if we should continue walking or
 * non-zero if we should stop, negative on error.  Extra user data may
 * be passed to @func with @data.  func() is of type typedef int
 * (*walker_func_t)(void *data, void *udata);
 * 
 * MI: Function modified...
 **/
int tree_walker(node_t *tree, walker_func_t func, void *data)
{
	int ret = 0;
	node_t *left;
	node_t *right;

	if (xtr) fprintf(stdout," === %s ===\n",__FUNCTION__);
	
	if (tree == NULL || func == NULL)
		return 0;

	left = tree->left;
	right = tree->right;

	if (tree->data) 
	{
		ret = func(tree->data, data);
		if (ret != 0)
			return ret;
	}
	
	if(left)
	{
		//printf("<--- \n");
		if ((ret = tree_walker(left, func, data)) != 0)
			return ret;
	}
	
	if (right)
	{
		//printf("--->\n");
		if ((ret = tree_walker(right, func, data)) != 0)
			return ret;
	}
	
	return 0;
}

/**
 * __node_lookup - find a node with key
 * @tree: tree root
 * @key: search key
 * @keylen: length of the search key
 * @wildcard: accept wildcard matches
 *
 * Searches radix tree for node with @key.  If @wildcard is set,
 * matches with data along the path are accepted.  This function
 * should not be called directly.
 *
 * MI: Function modified. Original not working at all.
 **/
node_t *__node_lookup(node_t *tree, void *key, int keylen, int wildcard)
{
	int bit;
	u_int32_t kb = 0, pkb = 0;
	node_t *n;
	struct in6_addr *pkey = NULL;
	int match;
	int entrcnt = 0;
	node_t *last = NULL;
	int bits_lastmatch = 0;

	n = tree;

	if (tree == NULL || key == NULL)
		return NULL;

	pkey = (struct in6_addr *)malloc(sizeof(struct in6_addr));
	
	if(xdbg)
	{
		struct in6_addr buf;
		hton_in6_addr(key,&buf);
		fprintf(stdout,"\n===> Looking for IP: %x:%x:%x:%x:%x:%x:%x:%x\n",
				NIP6ADDR(&buf));
	}
	
	while(n->data != NULL)
	{
		match = 1;
		++entrcnt;
		pkey = hton_in6_addr(&((struct rt_entry *)n->data)->destination, pkey);
		if(xdbg) fprintf (stdout,"Compare with: %x:%x:%x:%x:%x:%x:%x:%x\n",
				NIP6ADDR(&((struct rt_entry *)n->data)->destination));
		
		for (bit = 1; bit <= keylen; bit++) 
		{
			kb = setbit(key, bit); 
			pkb = setbit(pkey, bit);
			if(kb != pkb)
			{
				//If *wildcard* save what you have untill now
				if((wildcard) && (bits_lastmatch < bit-1) && match)
				{
					last = n;
					bits_lastmatch = bit-1;
				}
				match = 0;
			}

		} //for  

		if(match)
		{
			if (xdbg) 
			{
				fprintf(stdout,"Found match in AODV_RT: %x:%x:%x:%x:%x:%x:%x:%x\n",
				NIP6ADDR(&((struct rt_entry *)n->data)->destination));
			}
			break;
		}
		else //search down the tree...
		{
			//printf("Search down with kb = %d\n",kb);
			if(kb == 0)
			{
				if (n->left != NULL)
				{
					n = n->left;
					continue;
				}
				else if (!wildcard)
				{
					if(pkey)
						free(pkey);
					return NULL;
				}
			} else if (kb == 1)
			{
				if(n->right != NULL)
				{
					n = n->right;
					continue;
				}
				else if(!wildcard)
				{
					if (pkey)
						free(pkey);
					return NULL;
				}
			}
		}

		if (wildcard)
		{
			//Search randomly for a better match
			if ((n->left == NULL) && (n->right == NULL))
				break;
			else if (n->left != NULL)
			{
				n = n->left;
				continue;
			} else if (n->right != NULL)
			{
				n = n->right;
				continue;
			}
			
		}
	} //while

	if(pkey)
		free(pkey);
		
	if(wildcard)
		return last;
	else
		return n;
}

/**
 * node_add - add node with data to the radix tree
 * @tree: tree root
 * @key: search key
 * @keylen: length of the search key
 * @data: data associated to the node
 * @datalen: length of the data in bytes
 *
 * Creates a new node to the radix tree.  Given @data is stored at the
 * node.
 *
 * MI: Function modified. Original not working at all.
 **/
int node_add(node_t *tree, void *key, int keylen, void *data, int datalen)
{
	void *buf;
	int bit, kb;
	node_t *n;

	if (xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	
	if (tree == NULL || key == NULL || data == NULL) return -EINVAL;

	n = tree;

	while(n->data != NULL)
	{
		bit = keylen;
		kb = setbit(key,bit);
		if (kb == 0) 
		{
			if (n->left == NULL)
			{
				n->left = node_alloc(n);
				if (n->left == NULL)
					return -ENOMEM;
			}
			n = n->left;
		} else if (kb == 1) 
		{
			if (n->right == NULL) 
			{
				n->right = node_alloc(n);
				if (n->right == NULL)
					return -ENOMEM;
			}
			n = n->right;
		}
	}

	if (n->data != NULL)
		free(n->data);

	buf = malloc(datalen);
	if (buf == NULL) 
	{
		perror(__FUNCTION__);
		return -ENOMEM;
	}

	memcpy(buf, data, datalen);
	n->data = buf;

	if(xdbg)
	{
		struct rt_entry *ent = (struct rt_entry *)n->data;
		fprintf(stdout, "\tentry->dest: %x:%x:%x:%x:%x:%x:%x:%x/%d\n", 
				NIP6ADDR(&ent->destination), ent->plen);
		fprintf(stdout, "\tentry->nexthop: %x:%x:%x:%x:%x:%x:%x:%x\n",
				NIP6ADDR(&ent->nexthop));
		fprintf(stdout,"\tentry->lifetime[expires]: %lu[%lu] msec\n"
				,(unsigned long)ent->lifetime,(unsigned long)ent->lifetime-timems());
	}
	return 0;
}

/**
 * node_lookup - finds node data with key
 * @tree: tree root
 * @key: search key
 * @keylen: length of the search key
 * @wildcard: accept wildcard matches
 *
 * Searches radix tree for node with @key.  If @wildcard is set,
 * matches with data along the path are accepted.  Returns node data
 * entry rather than the actual node.
 **/
void *node_lookup(node_t *tree, void *key, int keylen, int wildcard)
{
	node_t *n;
	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (tree == NULL || key == NULL)
	{
		if (xdbg) 
			fprintf(stdout,"tree or key are NULL\n");
		return NULL;
	}

	if ((n = __node_lookup(tree, key, keylen, wildcard)) == NULL)
	{
		if (xdbg) 
			fprintf(stdout,"__node_lookup returned NULL\n");
		return NULL;
	}

	if (xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return n->data;
}

/**
 * node_modify - modify existing node in radix tree
 * @tree: tree root
 * @key: search key
 * @keylen: length of the search key
 * @data: data associated to the node
 * @datalen: length of the data in bytes
 *
 * Finds node with @key and modifies its data entry.  Given @data is
 * stored at the node.  If node is not found, returns negative.
 **/
int node_modify(node_t *tree, void *key, int keylen, void *data, int datalen)
{
	node_t *n;
	void *buf, *tmp;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	if ((n = __node_lookup(tree, key, keylen, 0)) == NULL)
		return -ENOENT;

	if (data == NULL)
		return -ENODATA;

	buf = malloc(datalen);
	if (buf == NULL) {
		perror(__FUNCTION__);
		return -ENOMEM;
	}

	memcpy(buf, data, datalen);
	tmp = n->data;
	n->data = buf;
	free(tmp);

	if (xdbg) fprintf(stdout,"Modified node's data (=== %s ===)\n", __FUNCTION__);
	return 0;
}

/**
 * node_delete - delete existing node from radix tree
 * @tree: tree root
 * @key: search key
 * @keylen: length of the search key
 *
 * Deletes a node with @key and @keylen from the radix tree.  Also
 * removes all intermediate nodes in the path.
 **/
void *node_delete(node_t *tree, void *key, int keylen) 
{
	node_t *n;
	void *data;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if ((n = __node_lookup(tree, key, keylen, 0)) == NULL)
		return NULL;

	data = n->data;
	n->data = NULL;

	while ((n = node_free(tree, n)) != NULL) ;

	return data;
}
