/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include "aodv6.h"
#include "rt_table.h"
#include "radix.h"
#include "tqueue.h"
#include "config.h"

struct rtnl_handle rth;
int rfd = -1;

struct rtnl_request {
	struct nlmsghdr n;
	struct rtmsg r;
	char buf[1024];
};

struct rtnl_genreq {
	struct nlmsghdr h;
	struct rtgenmsg m;
};

node_t *rtbl;

struct seqnum {
	u_int32_t sequence_number;
#if 0
	lock_t lock;
#endif
} seqnum;

struct in6_ifreq {
	struct in6_addr ifr6_addr;
	int ifr6_prefixlen;
};

extern int get_host_addr(struct in6_addr *addr);

void rt_dump(void);
void rt_expire(void *args);

/**
 * inc_get_seqnum - increment and get next sequence number
 *
 * Increments node's own sequence number by one and returns the new
 * sequence.
 **/
unsigned int inc_get_seqnum(void)
{
	u_int32_t sn;
	sn = ++seqnum.sequence_number;
	return sn;
}

/**
 * upd_get_seqnum - update and get next sequence number
 * @n: received sequence number
 *
 * Updates node's own sequence number to max(current, @n) and returns
 * new sequence number.  @n is the sequence number received in a Route
 * Request.
 **/
unsigned int upd_get_seqnum(u_int32_t n)
{
	u_int32_t sn;
	/* Sequence number MUST be incremented to avoid loops */
	sn = max(n, seqnum.sequence_number) + 1;
	seqnum.sequence_number = sn;

	return sn;
}

/**
 * get_seqnum - get current sequence number
 *
 * Returns current sequence number of this node.  Does not update
 * sequence.
 **/
unsigned int get_seqnum(void)
{
	u_int32_t sn;
	sn = seqnum.sequence_number;
	return sn;
}

/**
 * set_seqnum - set current sequence number
 *
 * Sets sequence number of this node.  
 **/
void set_seqnum(u_int32_t n)
{
	seqnum.sequence_number = n;
	return;
}


int rtmsg_add_rta(struct nlmsghdr *n, int maxlen, int type, void *data, int alen)
{
	int len = RTA_LENGTH(alen);
	struct rtattr *rta;

	if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
	{
		fprintf(stderr, "ERROR encountered in === %s ===\n",__FUNCTION__);
		return -1;
	}
	
	rta = (struct rtattr *)(((char *)n) + NLMSG_ALIGN(n->nlmsg_len));
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), data, alen);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
	return 0;
}

int rtmsg_query(struct rtnl_handle *rtnl, struct nlmsghdr *n, pid_t peer,
		unsigned groups, struct nlmsghdr *answer)
{
	int rcvd;
	unsigned seq;
	struct nlmsghdr *h;
	struct sockaddr_nl nl;
	struct iovec iov = { (void*)n, n->nlmsg_len };
	char   buf[8192];
	struct msghdr msg = {
		(void*)&nl, sizeof(nl),
		&iov,	1,
		NULL,	0,
		0
	};

	memset(&nl, 0, sizeof(nl));
	nl.nl_family = AF_NETLINK;
	nl.nl_pid = peer;
	nl.nl_groups = groups;
	n->nlmsg_seq = seq = ++rtnl->seq;
	if (answer == NULL)
		n->nlmsg_flags |= NLM_F_ACK;

	rcvd = sendmsg(rtnl->fd, &msg, 0);

	if (rcvd < 0) {
		perror("Cannot talk to rtnetlink");
		return -7;
	}

	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	while (1) {
		rcvd = recvmsg(rtnl->fd, &msg, 0);
		if (rcvd < 0) 
			continue;

		if (rcvd == 0) 
			return -8;

		for (h = (struct nlmsghdr *)buf; rcvd >= sizeof(*h); ) 
		{
			int len = h->nlmsg_len;

			if (len - sizeof(struct nlmsghdr) < 0 || len > rcvd)
				return -8;
			if (h->nlmsg_pid != rtnl->local.nl_pid)
				continue;

			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *err = (struct nlmsgerr*)NLMSG_DATA(h);
				int ret = -err->error;
				return ret;
			}

			if (answer) {
				memcpy(answer, h, h->nlmsg_len);
				return 0;
			}

			rcvd -= NLMSG_ALIGN(len);
			h = (struct nlmsghdr *)((char *)h + NLMSG_ALIGN(len));
		}
		if (msg.msg_flags & MSG_TRUNC) continue;

		if (rcvd) 
			return -10;
	}
}

static int nlmsg_add_route(struct sockaddr_nl *who, struct nlmsghdr *n)
{
	struct rtmsg *r = NLMSG_DATA(n);
	int len = n->nlmsg_len;
	struct rtattr *tb[RTA_MAX+1];
	struct in6_addr dest, nexthop;
	struct rt_entry *e;
	struct rtattr *rta;
	struct in6_ifreq *data;
	long expires = 0;
	int oif, metric;
	u32 timerid;

	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	len -= NLMSG_LENGTH(sizeof(*r));
	if (len < 0)
		return -1;

	if (n->nlmsg_type != RTM_NEWROUTE || r->rtm_type != RTN_UNICAST || 
	    r->rtm_table != RT_TABLE_MAIN)
		return 0;

	memset(tb, 0, sizeof(tb));

	rta = RTM_RTA(r);
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= RTA_MAX)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
	if (len) return -1;

	if (tb[RTA_DST] == NULL)
		return 0;

	memset(&dest, 0, sizeof(struct in6_addr));
	memcpy(&dest, RTA_DATA(tb[RTA_DST]), (r->rtm_dst_len+7)/8);

	if ((dest.s6_addr32[0] & htonl(0xFF000000)) == htonl(0xFF000000) ||
	    (dest.s6_addr32[0] & htonl(0xFE800000)) == htonl(0xFE800000))
		return 0;

	memset(&nexthop, 0, sizeof(nexthop));
	if (tb[RTA_GATEWAY])
		memcpy(&nexthop, RTA_DATA(tb[RTA_GATEWAY]), 16);

	oif = *(int*)RTA_DATA(tb[RTA_OIF]);
	metric = *(int*)RTA_DATA(tb[RTA_PRIORITY])/256;

	if (tb[RTA_CACHEINFO]) {
		struct rta_cacheinfo *ci = RTA_DATA(tb[RTA_CACHEINFO]);

		if (ci->rta_expires != 0)
		{
			expires = ci->rta_expires/100;
			if(xdbg) 
				fprintf(stdout, "Expiration set to %lu\n", expires);
		} 
	}

	e = malloc(sizeof(struct rt_entry));
	if (e == NULL)
		return -ENOMEM;
	memset(e, 0, sizeof(struct rt_entry));

	memcpy(&e->destination, &dest, sizeof(dest));
	memcpy(&e->nexthop, &nexthop, sizeof(nexthop));
	e->plen = r->rtm_dst_len;
	e->interface = oif;
	e->hopcount = metric;
	e->last_hopcount = 255;
	e->sequence = 1;
	/*
	 * MI - If not a cached route, expires=0, as initialized above
	 * This is the case for local IPs
	 * if expires=0, RT entry will be already expired 
	 * NOTE expires[sec] we need lifetime in millisecs
	 */
	e->lifetime = timems() + ((expires==0) ? LOCAL_ROUTE_LIFE : (expires*1000));

	if (r->rtm_protocol == RTPROT_BOOT) 
	{
		e->lifetime = timems() + 100;
		if(xdbg)
			fprintf(stdout,"INFO-ADD_RT: RTPROT_BOOT: e->lifetime = %u\n",
				e->lifetime);
	}

	data = malloc(sizeof(struct in6_ifreq));
	if (data == NULL) return -ENOMEM;

	memset(data, 0, sizeof(struct in6_ifreq));
	memcpy(&data->ifr6_addr, &e->destination, 
	       sizeof(struct in6_addr));
	data->ifr6_prefixlen = e->plen;

	//if(xdbg)
	//	rt_entry_print(stdout,e);

	rt_add(e, 0);

	timerid = in6_hash(&e->destination);
	add_alarm(e->lifetime - timems(), timerid, ROUTE_EXPIRE,
		  rt_expire, data);
	free(e);

	if (xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return 0;
}

static int nlmsg_add_addr(struct sockaddr_nl *nl, struct nlmsghdr *n, void *arg)
{
	struct addr_list **list;
	struct ifaddrmsg *ifa = NLMSG_DATA(n);
	struct rtattr *rta_tb[IFA_MAX+1];
	struct rtattr *rta;
	struct addr_list *a;
	int len;

	if (n->nlmsg_type != RTM_NEWADDR)
		return 0;
	if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifa)))
		return -1;

	if (ifa->ifa_scope & ~SCOPE_SITE)
		return 0;
	if (ifa->ifa_flags & (IFA_F_TENTATIVE | IFA_F_DEPRECATED))
		return 0;

	a = malloc(sizeof(struct addr_list));
	if (a == NULL)
		return -1;

	a->ifindex = ifa->ifa_index;
	memset(rta_tb, 0, sizeof(rta_tb));
	rta = IFA_RTA(ifa);
	len = n->nlmsg_len - NLMSG_LENGTH(sizeof(*ifa));
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= IFA_MAX)
			rta_tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
	if (len) return -1;

	if (!rta_tb[IFA_LOCAL])
		rta_tb[IFA_LOCAL] = rta_tb[IFA_ADDRESS];
	if (!rta_tb[IFA_ADDRESS])
		rta_tb[IFA_ADDRESS] = rta_tb[IFA_LOCAL];

	memcpy(&a->addr, RTA_DATA(rta_tb[IFA_ADDRESS]), sizeof(struct in6_addr));
	a->prefixlen = ifa->ifa_prefixlen;
	a->next = NULL;

	for (list = (struct addr_list **)arg; *list; list = &(*list)->next);
	*list = a;

	return 0;
}

/**
 * rtnl_local_addrs - get all routable addresses for this node
 *
 * Gets all site-local and global scope addresses for this node.
 * Addresses are stored in a linked list which is returned to the
 * caller.  Function allocates memory for the addresses and caller
 * should free the memory, if necessary.
 **/
struct addr_list *rtnl_local_addrs(void)
{
	struct rtnl_genreq msg;
	struct addr_list *alist = NULL;
	struct sockaddr_nl nl;
	char buf[8192];
	struct iovec iov = { 
		buf, 
		sizeof(buf)
	};

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	memset(&nl, 0, sizeof(nl));
	nl.nl_family = AF_NETLINK;

	memset(&msg, 0, sizeof(msg));
	msg.h.nlmsg_len = sizeof(msg);
	msg.h.nlmsg_type = RTM_GETADDR;
	msg.h.nlmsg_flags = NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST;
	msg.h.nlmsg_seq = rth.dump = ++rth.seq;
	msg.m.rtgen_family = AF_INET6;

	if (sendto(rth.fd, (void *)&msg, sizeof(msg), 0, (struct sockaddr *)&nl, sizeof(nl)) < 0) {
		perror("Cannot send dump request");
		exit(1);
	}

	while (1) {
		int rcvd;
		struct nlmsghdr *h;

		struct msghdr msg = {
			(void*)&nl, sizeof(nl),
			&iov, 1,
			NULL, 0,
			0
		};

		rcvd = recvmsg(rth.fd, &msg, 0);

		if (rcvd < 0) continue;
		if (rcvd == 0) return NULL;

		h = (struct nlmsghdr *)buf;
		while (NLMSG_OK(h, rcvd)) {
			int err;

			if (h->nlmsg_pid != rth.local.nl_pid ||
			    h->nlmsg_seq != rth.dump) {
				goto nextmsg;
			}

			if (h->nlmsg_type == NLMSG_DONE)
				return alist;
			if (h->nlmsg_type == NLMSG_ERROR)
				return NULL;

			err = nlmsg_add_addr(&nl, h, &alist);
			if (err < 0)
				return NULL;

		nextmsg:
			h = NLMSG_NEXT(h, rcvd);
		}
		if (msg.msg_flags & MSG_TRUNC) continue;

		if (rcvd) {
			exit(1);
		}
	}
}

/**
 * rtnl_get_saddr - get source address for a destination
 * @daddr: destination address
 * @saddr: source address
 *
 * Requests source address for a destination from the kernel routing
 * table.  Returns zero on success, and on error a negative value.
 **/
int rtnl_get_saddr(struct in6_addr *daddr, struct in6_addr *saddr)
{
	struct rtnl_request req;
	struct rtattr *tb[RTA_MAX+1];
	struct rtattr *rta;
	int len;

	if (daddr == NULL) return -EINVAL;
	if (saddr == NULL) return -EINVAL;

	memset(&req, 0, sizeof(struct rtnl_request));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = RTM_GETROUTE;
	req.r.rtm_family = AF_INET6;

	rtmsg_add_rta(&req.n, sizeof(req), RTA_DST, daddr, 16);

	if (rtmsg_query(&rth, &req.n, 0, 0, &req.n) < 0) {
		return -2;
	}

	len = req.n.nlmsg_len - NLMSG_LENGTH(sizeof(req.r));
	memset(tb, 0, sizeof(tb));

	rta = RTM_RTA(&req.r);
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= RTA_MAX)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
	if (len) return -1;

	if (*(int *)RTA_DATA(tb[RTA_OIF]) == 1) {
		return -1;
	}
	if (tb[RTA_SRC]) {
		memcpy(saddr, RTA_DATA(tb[RTA_SRC]), 
		       sizeof(struct in6_addr));
	}

	return 0;
}

/*
 * Since rtnetlink doesn't allow us to set a lifetime for a route, we
 * use the old ioctl interface instead.  This however, is a bit buggy
 * and requires us to give expiration time in _jiffies_.  I couldn't
 * figure out any other way to get the jiffies, than to read from the
 * /proc/uptime.  This is a very dirty hack, and if there is a way
 * around this, please let me know.  When a proper way is found,
 * get_jiffies() and rtioctl_rt_add() will be removed.
 */
unsigned long get_jiffies(void)
{
	unsigned long j, sec, jif;
	int fd = -1;
	char buffer[51], left[50];

	if ((fd = open("/proc/uptime", O_RDONLY)) < 0) {
		perror("open");
		return -1;
	}
	read(fd, &buffer, 50);
	close(fd);
	sscanf(buffer, "%lu.%lu %s", &sec, &jif, left);
	j = (sec * 100) + jif;

	return j;
}

#define RTF_EXPIRES 0x00400000

/**
 * rtioctl_rt_add - add route to kernel routing table
 * @destination: destination address
 * @plen: prefix length if not host route
 * @nexthop: address for next hop
 * @ifindex: interface index
 *
 * Adds route to @destination via @nexthop to kernel routing table
 * using ioctl.
 **/
int rtioctl_rt_add(struct in6_addr *destination, int plen,
		   struct in6_addr *nexthop, int ifindex,
		   unsigned long lifetime)
{
	struct in6_rtmsg rtmsg;
	int err = 0;
	
	if(xtr)	fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	
	if (xdbg)
	{
		fprintf(stdout, "\tdestination/plen: %x:%x:%x:%x:%x:%x:%x:%x/%d\n",
			NIP6ADDR(destination), plen);
		if(nexthop)
			fprintf(stdout, "\tnexthop %x:%x:%x:%x:%x:%x:%x:%x ifindex=%d\n", 
				NIP6ADDR(nexthop), ifindex);
		else
			fprintf(stdout, "\tnexthop (none) ifindex=%d\n",ifindex);
		fprintf(stdout, "\tlifetime: %d\n", (int)lifetime);
	}

	if(nexthop) 
	{
		if (( ipv6_addr_cmp(destination,&Config.NodeAddrGlobal) == 0) &&
			(ipv6_addr_cmp(nexthop,&Config.HostAddrGlobal) == 0))
		{
			//fprintf(stdout,"NOTE: NOT ADDED because would return errors!\n");
			return 0;
		}
	}
	
	memset(&rtmsg, 0, sizeof(struct in6_rtmsg));
	memcpy(&rtmsg.rtmsg_dst, destination, sizeof(struct in6_addr));

	rtmsg.rtmsg_flags = RTF_UP;
	if (plen == 128)
		rtmsg.rtmsg_flags |= RTF_HOST;
	if (nexthop) {
		rtmsg.rtmsg_flags |= RTF_GATEWAY;
		memcpy(&rtmsg.rtmsg_gateway, nexthop, sizeof(struct in6_addr));
	}
	rtmsg.rtmsg_metric = 1;
	rtmsg.rtmsg_dst_len = plen;
	rtmsg.rtmsg_metric = 256;
	rtmsg.rtmsg_flags |= RTF_EXPIRES;

	//MI- jiffies are centiseconds, lifetime is in milliseconds
	rtmsg.rtmsg_info = get_jiffies() + (lifetime / 10);
	rtmsg.rtmsg_ifindex = ifindex;

	err = ioctl(rfd, SIOCADDRT, &rtmsg);
	if (err < 0) 
	{
		if(xdbg) 
		{
			fprintf(stdout,"WARNING ioctl returned err = %d: but it looks like it updated the RT for\n", err);
			fprintf(stdout,"%x:%x:%x:%x:%x:%x:%x:%x via %x:%x:%x:%x:%x:%x:%x:%x\n", 
				NIP6ADDR(destination), NIP6ADDR(nexthop));
			perror("SIOCADDRT");
		}
		//return err;
	} 
	/* Success - add alarm to remove the entry when it expires */
	{
		struct in6_ifreq *data;
		u32 timerid = in6_hash(destination);
			
	 	data = malloc(sizeof(struct in6_ifreq));
		if (data != NULL)
		{
			memset(data, 0, sizeof(struct in6_ifreq));
			memcpy(&data->ifr6_addr, destination, 
			       sizeof(struct in6_addr));
			data->ifr6_prefixlen = plen;

			if(xdbg)
				fprintf(stdout,"Route will expire in %lu, set alarm to remove it.\n", lifetime);
			add_alarm(lifetime, timerid, ROUTE_EXPIRE,
				  rt_expire, data);
		}

	}

	if(xdbg) 
	{
		fprintf(stdout,"-> Updated static RT: ");
		fprintf(stdout,"%x:%x:%x:%x:%x:%x:%x:%x/%d ",NIP6ADDR(destination), plen);
		if(nexthop)
			fprintf(stdout,"via %x:%x:%x:%x:%x:%x:%x:%x", NIP6ADDR(nexthop));
		fprintf(stdout,"\n");
	}
	return 0;
}

/**
 * rtnl_rt_add - add route to kernel routing table
 * @destination: destination address
 * @plen: prefix length if not host route
 * @nexthop: address for next hop
 * @ifindex: interface index
 * @lifetime: route lifetime
 *
 * Adds route to @destination via @nexthop to kernel routing table
 * using rtnetlink.
 **/
int rtnl_rt_add(struct in6_addr *destination, int plen,
		struct in6_addr *nexthop, int ifindex,
		unsigned long lifetime)
{
#if THINGS_WORKED_AS_THEY_SHOULD

	struct rtnl_request req;
	char mxbuf[256];
	struct rtattr *mxrta = (void *)mxbuf;
	int r = 0;

	if (destination == NULL) return -EINVAL;
	if (plen < 0 || plen > 128) return -EINVAL;

	memset(&req, 0, sizeof(req));

	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE | NLM_F_EXCL;
	req.n.nlmsg_type = RTM_NEWROUTE;

	req.r.rtm_family = AF_INET6;
	req.r.rtm_dst_len = plen;
	req.r.rtm_table = RT_TABLE_MAIN;
	/* Linux IPv6 stack currently (=< 2.4.18) discards RTPROT
	 * info.  RTPROT_AODV is used for future. */
	req.r.rtm_protocol = RTPROT_AODV;
	req.r.rtm_scope = RT_SCOPE_UNIVERSE;
	req.r.rtm_type = RTN_UNICAST;
	/* See if we should set rtm_flags = RTM_F_NOTIFY; to get
         * notifications of route changes. */

	mxrta->rta_type = RTA_METRICS;
	mxrta->rta_len = RTA_LENGTH(0);

	ulog(L_INFO | L_RTADD, "== " __FUNCTION__ " ==\n");
	ulog(L_INFO | L_RTADD, "destination: %x:%x:%x:%x:%x:%x:%x:%x\n",
	       NIP6ADDR(destination));
	
	if(xtr) fprintf(stdout, "== " __FUNCTION__ " ==\n");
	if (xdbg) fprintf(stdout, "destination: %x:%x:%x:%x:%x:%x:%x:%x/%d\n",
	       NIP6ADDR(destination), plen);
	
	rtmsg_add_rta(&req.n, sizeof(req), RTA_DST, destination, 16);
	
	if (nexthop) 
	{
		rtmsg_add_rta(&req.n, sizeof(req), RTA_GATEWAY, nexthop, 16);
		fprintf(stdout, "gateway/nexthop: %x:%x:%x:%x:%x:%x:%x:%x\n",
		       NIP6ADDR(nexthop));
		ulog(L_INFO | L_RTADD, "gateway/nexthop: %x:%x:%x:%x:%x:%x:%x:%x\n",
		       NIP6ADDR(nexthop));
	} else
		fprintf(stdout, "gateway/nexthop: - NULL\n");
	
	if (ifindex > 1) 
	{
		rtmsg_add_rta(&req.n, sizeof(req), RTA_OIF, &ifindex, sizeof(int));
		fprintf(stdout,"oif(==ifindex==interface): %d\n",ifindex);
		ulog(L_INFO | L_RTADD, "oif: %d\n", ifindex);
	}
	
	/* check if we should set other attributes as well */
	if ((r = rtmsg_query(&rth, &req.n, 0, 0, NULL)) < 0) 
	{
		perror("rtmsg_query");
		return -2;
	}
	
	if(xtr) fprintf(stdout, "+++ %s +++\n",__FUNCTION__);
	return 0;

#else
	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	return rtioctl_rt_add(destination, plen, nexthop, ifindex, lifetime);
#endif
}

int rtnl_rt_host_add(struct in6_addr *destination, 
		     struct in6_addr *nexthop,
		     int ifindex, unsigned long lifetime)
{
	if(xtr) fprintf(stdout, "=== %s ===\n",__FUNCTION__);
	
	if (destination == NULL || nexthop == NULL) 
	{
		fprintf(stdout,"ERROR: dest or nexthop are NULL. Go back\n");
		return -EINVAL;
	}

	if (memcmp(destination, nexthop, sizeof(struct in6_addr)) == 0) 
	{
		//fprintf(stdout,"dest == nexthop add route with nexthop NULL.\n");
		return rtnl_rt_add(destination, 128, NULL, ifindex, lifetime);
	}

	return rtnl_rt_add(destination, 128, nexthop, ifindex, lifetime);
}


/**
 * rtnl_rt_update - update kernel routing table entry
 * @dest: destination address
 * @nh: next hop address
 * @oif: outgoing interface index
 *
 * Updates or creates a new kernel routing table entry for destination
 * @dest with @nh next hop through interface @oif.
 **/
int rtnl_rt_update(struct in6_addr *dest, struct in6_addr *nh, int oif,
		   unsigned long lifetime)
{
	struct rtnl_request req;
	struct rtattr *tb[RTA_MAX+1];
	struct rtattr *rta;
	int len, coif = 0, ret=0;
 
	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (dest == NULL) return -EINVAL;

	ulog(L_INFO | L_RTMOD, "updating route for oif %d\n", oif);
	memset(&req, 0, sizeof(struct rtnl_request));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = RTM_GETROUTE;
	req.r.rtm_family = AF_INET6;
	req.r.rtm_dst_len = 128;

	ret=rtmsg_add_rta(&req.n, sizeof(req), RTA_DST, dest, 16);
	
	if(ret!=0) 
		fprintf(stdout,"ERROR from rtmsg_add_rta: could not add dest to req.n\n");
	
	if (nh && !IN6_IS_ADDR_UNSPECIFIED(nh)) 
	{
		ret = rtmsg_add_rta(&req.n, sizeof(req), RTA_GATEWAY, nh, 16);
		if(ret!=0) fprintf(stdout,"ERROR from rtmsg_add_rta: could not add nexthop to req.n\n");
	}
		
	if (oif > 1)
	{
		ret=rtmsg_add_rta(&req.n, sizeof(req), RTA_OIF, &oif, sizeof(int));
		if(ret!=0) 
			fprintf(stdout,"ERROR from rtmsg_add_rta: could not add device's oif to req.n\n");
	}

	if (rtmsg_query(&rth, &req.n, 0, 0, &req.n) < 0) 
	{
		fprintf(stdout, "ERROR: rtmsg_query returned error.\n");
		perror("rtmsg_query");
	}

	len = req.n.nlmsg_len - NLMSG_LENGTH(sizeof(req.r));
	if (len < 0) {
		ulog(L_WARN | L_RTMOD, "invalid message length\n");
		return -1;
	}
	memset(tb, 0, sizeof(tb));

	rta = RTM_RTA(&req.r);
	while (RTA_OK(rta, len)) 
	{
		if (rta->rta_type <= RTA_MAX)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
	
	if (len) 
	{
		fprintf(stderr,"ERROR in %s len is != 0\n",__FUNCTION__);
		return -1;
	}

	coif = *(int *)RTA_DATA(tb[RTA_OIF]);

	if(xdbg) 
	{
		fprintf(stdout,"rtmsg_query returned:\n");
		fprintf(stdout, "\tDestination  %x:%x:%x:%x:%x:%x:%x:%x\n",
			NIP6ADDR(RTA_DATA(tb[RTA_DST])));
		//? fprintf(stdout, "\tNexthop      %x:%x:%x:%x:%x:%x:%x:%x\n",
		//	NIP6ADDR(RTA_DATA(tb[RTA_GATEWAY])));
		fprintf(stdout, "\tInterface (coif/oif)   %d/%d\n",coif,oif);
		//? fprintf(stdout, "\tLifetime     %lu\n", RTA_DATA(tb[]))
	}

	if ((coif <= 1 || coif != oif) && (oif > 0)) 
	{
		if(xdbg) 
		{
			fprintf(stdout,"-> No Kernel RT exists, creating new entry:\n");
			if(dest)
				fprintf(stdout,"\t%x:%x:%x:%x:%x:%x:%x:%x ", 
					NIP6ADDR(dest));
			if(nh)
				fprintf(stdout," -> %x:%x:%x:%x:%x:%x:%x:%x ",
					NIP6ADDR(nh));
			fprintf(stdout, "dev %d lifetime %lu\n", oif, lifetime);
		}
		ulog(L_INFO | L_RTMOD, "no old route to nh, creating new\n");
		ret = rtnl_rt_host_add(dest, nh, oif, lifetime);

	} else	//We do want to update the lifetime as per RFC!
	{
		if(xdbg)
		{
			fprintf(stdout,"--->Case: coif[=%d] > 1 or coif == oif[=%d]\n",coif,oif);
			fprintf(stdout,"\tNext: update lifetime to %lu\n", lifetime);
		}
		ret = rtnl_rt_add(dest, 128, nh, oif, lifetime);
	}
	
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return ret;
}

int rtnl_rt_lookup(struct rtnl_request *req,
		   struct in6_addr *destination, int plen);
/**
 * rtnl_rt_del - delete route from kernel routing table
 * @destination: route destination to delete
 * @plen: prefix length of destination
 * @nexthop: next hop address
 * @ifindex: interface index
 *
 * Delete route to @destination via @nexthop from kernel routing table
 * using rtnetlink.
 **/
int rtnl_rt_del(struct in6_addr *destination, int plen, 
		struct in6_addr *nexthop, int ifindex)
{
	struct rtnl_request req;

	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	//ulog(L_INFO | L_RTADD, "== " __FUNCTION__ "\n");
	if (destination == NULL) return -EINVAL;
	if (plen < 0 || plen > 128) return -EINVAL;

	//ulog(L_INFO | L_RTADD, __FUNCTION__ ": destination ok\n");
	memset(&req, 0, sizeof(req));

	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST | NLM_F_EXCL;
	req.n.nlmsg_type = RTM_DELROUTE;

	req.r.rtm_family = AF_INET6;
	req.r.rtm_dst_len = plen;
	req.r.rtm_src_len = 0;
	req.r.rtm_tos = 0;
	req.r.rtm_table = RT_TABLE_MAIN;
	
	//RT_SCOPE_NOWHERE means not existing destinations
	req.r.rtm_scope = RT_SCOPE_NOWHERE;

	//Building the request
	rtmsg_add_rta(&req.n, sizeof(req), RTA_DST, destination, 16);
	if (nexthop) {
		rtmsg_add_rta(&req.n, sizeof(req), RTA_GATEWAY, nexthop, 16);
	}
	if (ifindex > 1) {
		rtmsg_add_rta(&req.n, sizeof(req), RTA_OIF, &ifindex, sizeof(int));
	}

	if(xdbg)fprintf(stdout,"-sending query to delete entry from RT\n");
	//ulog(L_INFO | L_RTADD, __FUNCTION__ ": sending query\n");
	if (rtmsg_query(&rth, &req.n, 0, 0, NULL) < 0) 
	{
		if(xdbg) fprintf(stdout,"-failure\n");
		//ulog(L_INFO | L_RTADD, __FUNCTION__ ": failure\n");
		return -2;
	}
	if(xdbg) fprintf(stdout,"-success\n");
	//ulog(L_INFO | L_RTADD, __FUNCTION__ ": success\n");

	if(xdbg)
	{
		fprintf(stdout,
			"Entry %x:%x:%x:%x:%x:%x:%x:%x", NIP6ADDR(destination));
		if(nexthop)
			fprintf(stdout," via %x:%x:%x:%x:%x:%x:%x:%x", NIP6ADDR(nexthop));
		fprintf(stdout, " removed\n");
	}

	if (xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return 0;
}


/**
 * rtnl_rt_lookup - find a route from kernel routing table
 * @req: rtnetlink message
 * @destination: route destination address
 * @plen: prefix lenght
 *
 * Finds a route to @destination from kernel routing table using
 * rtnetlink.
 **/
int rtnl_rt_lookup(struct rtnl_request *req,
		   struct in6_addr *destination, int plen)
{
	if (req == NULL || destination == NULL) return -EINVAL;
	if (plen < 0 || plen > 128) return -EINVAL;

	memset(req, 0, sizeof(struct rtnl_request));
	req->n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req->n.nlmsg_flags = NLM_F_REQUEST;
	req->n.nlmsg_type = RTM_GETROUTE;
	req->r.rtm_family = AF_INET6;
	req->r.rtm_dst_len = plen;

	rtmsg_add_rta(&req->n, sizeof(req), RTA_DST, destination, 16);

	if (rtmsg_query(&rth, &req->n, 0, 0, &req->n) < 0) {
		free(req);
		req = NULL;
		return -2;
	}

	return 0;
}

/**
 * rtnl_rt_sync_listener - listen to RTNetlink messages
 * @nlh: netlink socket handle
 * @n: pointer to netlink message header
 *
 * Listens to routing table messages to learn changes in the kernel
 * routing table.  Used for syncing daemon routing table with
 * kernel's.
 **/
int rtnl_rt_sync_listener(struct sockaddr_nl *nlh, struct nlmsghdr *n)
{
	struct rtmsg *r = NLMSG_DATA(n);
	struct rtattr *tb[RTA_MAX+1];
	struct rtattr *rta;
	struct rta_cacheinfo cacheinfo;
	struct in6_addr addr;
	struct in6_addr *gw = NULL, *dst = NULL;
	int len = n->nlmsg_len;
	int event = n->nlmsg_type;
	int oif = 0;
	unsigned long lifetime;

	//ulog(L_INFO | L_RTSYNC, "== " __FUNCTION__ " ==\n");

	if (r->rtm_family != AF_INET6)
		return 0;

	if (r->rtm_table != RT_TABLE_MAIN)
		return 0;

	if (r->rtm_scope != RT_SCOPE_UNIVERSE)
		return 0;

	if (!NLMSG_OK(n, n->nlmsg_len))
		return 0;

	if (r->rtm_type != RTN_UNICAST)
		return 0;

	len -= NLMSG_LENGTH(sizeof(*r));
	memset(tb, 0, sizeof(tb));

	rta = RTM_RTA(r);
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= RTA_MAX)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
	if (len) return -1;

	memset(&addr, 0, sizeof(addr));

	if (tb[RTA_DST]) {
		dst = malloc(sizeof(struct in6_addr));
		memset(dst, 0, sizeof(struct in6_addr));
		memcpy(dst, RTA_DATA(tb[RTA_DST]), (r->rtm_dst_len+7)/8);
		if ((dst->s6_addr32[0] & htonl(0xFF000000)) == htonl(0xFF000000)) {
			free(dst);
			return 0;
		}
	}

	if (tb[RTA_GATEWAY])
		gw = RTA_DATA(tb[RTA_GATEWAY]);

	if (tb[RTA_OIF])
		oif = *(int*)RTA_DATA(tb[RTA_OIF]);

	memset(&cacheinfo, 0, sizeof(cacheinfo));
	if (tb[RTA_CACHEINFO]) {
		memcpy(&cacheinfo, RTA_DATA(tb[RTA_CACHEINFO]),
		       sizeof(struct rta_cacheinfo));
	}

	if (event == RTM_DELROUTE) 
	{
		rt_del(dst, r->rtm_dst_len);
	} else if (event == RTM_NEWROUTE) 
	{
		if (cacheinfo.rta_expires == 0)
		{
			fprintf(stdout,"Cache expired. Lifetime now = %d\n", activeRouteTimeout);
			//lifetime = ACTIVE_ROUTE_TIMEOUT;
			lifetime = activeRouteTimeout;
		}
		else
		{
			fprintf(stdout,"Chache NOT expired. Lifetime now = %d\n", cacheinfo.rta_expires);
			lifetime = cacheinfo.rta_expires;
		}

		rtnl_rt_add(dst, r->rtm_dst_len, gw, oif, lifetime);
	}

	return 0;
}

/**
 * rt_init - initialize routing
 *
 * Initializes AODV internal routing table and initializes netlink
 * socket for kernel routing table interface.
 **/
int rt_init(void)
{
	int len;
	struct rtnl_genreq msg;
	struct sockaddr_nl nl;
	char buf[8192];
	struct iovec iov = { 
		buf, 
		sizeof(buf)
	};

	if(xtr) fprintf (stdout,"=== %s ===\n",__FUNCTION__);

	/* open netlink socket */
	memset(&rth, 0, sizeof(rth));

	if ((rth.fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE)) < 0) {
		perror("socket");
		return -1;
	}

	if ((rfd = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) 
	{
		perror("socket");
		return -1;
	}
	if(xdbg) fprintf(stdout,"Kernel RT sock: %d\n",rfd);
	
	rth.local.nl_family = AF_NETLINK;
	if (bind(rth.fd, (struct sockaddr *)&rth.local, 
		 sizeof(struct sockaddr_nl)) < 0) 
	{
		perror("bind");
		return -1;
	}

	len = sizeof(struct sockaddr_nl);
	if (getsockname(rth.fd, (struct sockaddr*)&rth.local, &len) < 0) {
		perror("getsockname");
		return -1;
	}

	/* initialize radix tree root */
	if ((rtbl = create_tree()) == NULL)
		return -1;

	/* dump current kernel routing table to ours */
	memset(&nl, 0, sizeof(nl));
	nl.nl_family = AF_NETLINK;

	memset(&msg, 0, sizeof(msg));
	msg.h.nlmsg_len = sizeof(msg);
	msg.h.nlmsg_type = RTM_GETROUTE;
	msg.h.nlmsg_flags = NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST;
	msg.h.nlmsg_seq = rth.dump = ++rth.seq;
	msg.m.rtgen_family = AF_INET6;

	if (sendto(rth.fd, (void *)&msg, sizeof(msg), 0, (struct sockaddr *)&nl, sizeof(nl)) < 0) {
		perror("Cannot send dump request");
		exit(1);
	}

	while (1) {
		int rcvd;
		struct nlmsghdr *h;

		struct msghdr msg = {
			(void*)&nl, sizeof(nl),
			&iov, 1,
			NULL, 0,
			0
		};

		rcvd = recvmsg(rth.fd, &msg, 0);

		if (rcvd < 0) continue;
		if (rcvd == 0) return -1;

		h = (struct nlmsghdr *)buf;
		while (NLMSG_OK(h, rcvd)) {
			int err;

			if (h->nlmsg_pid != rth.local.nl_pid ||
			    h->nlmsg_seq != rth.dump) {
				goto nextmsg;
			}

			if (h->nlmsg_type == NLMSG_DONE)
				return 0;
			if (h->nlmsg_type == NLMSG_ERROR)
				return -1;

			err = nlmsg_add_route(&nl, h);

			if (err < 0) 
			{
				fprintf(stderr,"Kernel RT dump encountered error! Terminate it.\n");
				ulog(L_CRIT | L_RTSYNC, "Dump terminated\n");
				exit(1);
			}

		nextmsg:
			h = NLMSG_NEXT(h, rcvd);
		}

		if (msg.msg_flags & MSG_TRUNC) continue;

		if (rcvd) {
			exit(1);
		}
	}

	return 0;
}

/**
 * rt_is_local - check route flag if entry is for local address
 * @entry: routing table entry
 *
 * Checks if @entry is for a local address.
 **/
int rt_is_local(struct rt_entry *entry)
{
	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (entry == NULL) return -EINVAL;

	if(xdbg) fprintf(stdout,"returns %d=(entry->flag=%d)&(RT_F_DEST_LOCAL=%d)\n",
				(entry->flags)&(RT_F_DEST_LOCAL),entry->flags, RT_F_DEST_LOCAL);

	return (entry->flags & RT_F_DEST_LOCAL);
}

/**
 * rt_is_local_address - check if local address
 * @addr: address to check
 *
 * Checks if @addr is one of local node's addresses.
 **/
int rt_is_local_addr(struct in6_addr *addr)
{
	struct in6_addr key;
	struct rt_entry *e;
	//AODV's autoconfig link-local address fe80::200:ff:fe00:0/64
	struct in6_addr aodv_link = { { { 0xfe,0x80,0,0,0,0,0,0,0x02,0,0,0xff,0xfe,0,0,0 } } };

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	//fprintf(stdout, "AODV-link-addr is: %x:%x:%x:%x:%x:%x:%x:%x\n",
	//NIP6ADDR(&aodv_link));
	
	if (addr == NULL) return -EINVAL;

	ntoh_in6_addr(addr, &key);
	
	if(xdbg) 
		fprintf(stdout,"Compare %x:%x:%x:%x:%x:%x:%x:%x & %x:%x:%x:%x:%x:%x:%x:%x\n", 
			NIP6ADDR(addr), NIP6ADDR(&Config.HostAddrGlobal));

	if(ipv6_addr_cmp(&Config.HostAddrGlobal,addr) == 0)
	{
		if(xdbg) fprintf(stdout,"%x:%x:%x:%x:%x:%x:%x:%x is host's global address\n", NIP6ADDR(addr));
		return 1;
	}

	if(xdbg) 
		fprintf(stdout,"Compare %x:%x:%x:%x:%x:%x:%x:%x & %x:%x:%x:%x:%x:%x:%x:%x\n", 
			NIP6ADDR(addr), NIP6ADDR(&Config.NodeAddrGlobal));

	if(ipv6_addr_cmp(&Config.NodeAddrGlobal,addr) == 0)
	{
		if(xdbg) fprintf(stdout,"%x:%x:%x:%x:%x:%x:%x:%x is node's global address\n", NIP6ADDR(addr));
		return 1;
	}


	if(ipv6_addr_cmp(&aodv_link,addr) == 0)
	{
		if(xdbg) fprintf(stdout,"%x:%x:%x:%x:%x:%x:%x:%x is local-link\n", NIP6ADDR(addr));
		return 1;
	}

	if(xdbg) 
		fprintf(stdout,"\tCalling node_lookup for addr: %x:%x:%x:%x:%x:%x:%x:%x/128 in RT:\n", NIP6ADDR(addr));
	//rt_dump(); //in the dump the nexthop is wrong...

	
	e = (struct rt_entry *)node_lookup(rtbl, &key, 128, 0);
	
	if (e == NULL)
	{
		if(xdbg) fprintf(stdout,"No entry found, returning with 0\n");
		return 0;
	}

	if(xdbg) fprintf(stdout,"Found entry, check flag.\n");

	return rt_is_local(e);
}

/**
 * rt_set_local - set local addresses
 * @list: linked list of addresses
 *
 * Inserts a list of addresses as node's local addresses in the
 * routing table.
 **/
int rt_set_local(struct addr_list *list)
{
	struct in6_addr key;
	struct in6_addr host_ip;
	struct rt_entry *entry = malloc(sizeof(struct rt_entry));
	int ret=0;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (entry == NULL) return -ENOMEM;
	if (list == NULL) return -EINVAL;

	ret = get_host_addr(&host_ip);
	
	if((ret < 0) || (&host_ip == NULL))
	{
		fprintf(stdout,"ERROR: Can't determin host_ip\n");
		return -1;
	}
	
	ntoh_in6_addr(&list->addr, &key);

	memset(entry, 0, sizeof(struct rt_entry));
	ipv6_addr_copy(&entry->destination, &list->addr);
	
	//With nexthop=NULL should work, we tested, but with itself
	//not sure. The encountered problems might come from here...
	//ipv6_addr_copy(&entry->nexthop, &list->addr);
	ipv6_addr_copy(&entry->nexthop, &host_ip);

	entry->interface = list->ifindex;
	entry->hopcount = 0;
	entry->last_hopcount = 0;
	//Why is lifetime 0? When displayed = lifetime-timems()
	//entry->lifetime = 0;
	entry->lifetime = timems() + LOCAL_ROUTE_LIFE;
	entry->plen = 128;
	entry->flags |= RT_F_DEST_LOCAL;

	if (node_add(rtbl, &key, entry->plen, entry,
		     sizeof(struct rt_entry)) < 0) {
		ulog(L_CRIT | L_ROUTE, "Could not add entry to AODV routing table\n");
		free(entry);
		return -1;
	}

	if(xdbg) {
		fprintf(stdout,"\tAdded local entry to AODV RT:\n");
		//rt_entry_print(stdout,entry);
	}
	
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);

	free(entry);
	return 0;
}

/**
 * rt_expire - expire a routing table entry for destination
 * @args: address and prefix length
 *
 * Invalidates a route for destination.  Can be called from task
 * queue.
 **/
void rt_expire(void *args)
{
	struct in6_ifreq *r = (struct in6_ifreq *)args;
	struct rt_entry *rte;
	struct in6_addr key;
	int wild;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
		
	//ulog(L_INFO | L_RTDEL, "== " __FUNCTION__ "==\n");
	if (args == NULL) return;

	/*
	 * MI: This function is called by add_alarm and
	 * it happens to delete routes that just have been
	 * updated 
	 */
	memset(&key,0,sizeof(struct in6_addr));
	ntoh_in6_addr(&r->ifr6_addr, &key);
	wild = (r->ifr6_prefixlen <= 128) ? 0 : 1;
	
	if ((rte = (struct rt_entry *)
		     node_lookup(rtbl, &key, 128, wild)) == NULL)	
		return;

	if(xdbg)
		fprintf(stdout,"Route to %x:%x:%x:%x:%x:%x:%x:%x expires in %d.\n",
	       	NIP6ADDR(&r->ifr6_addr), (int)(rte->lifetime - timems()));

	/* Check lifetime left and invalidated RT*/
	if((rte->lifetime <= timems()) || (rte->hopcount == INFINITE_HOPS))
	{
		if(xdbg)
			fprintf(stdout,
					"Route to %x:%x:%x:%x:%x:%x:%x:%x/%d expired => deleted\n",
	       			NIP6ADDR(&r->ifr6_addr), r->ifr6_prefixlen);
		ulog(L_INFO | L_RTDEL, "route to %x:%x:%x:%x:%x:%x:%x:%x expired\n",
	    	   NIP6ADDR(&r->ifr6_addr));
	
		rt_del(&r->ifr6_addr, r->ifr6_prefixlen);
	}
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);

	return;
}

/**
 * rt_add - add entry to routing table
 * @entry: entry to add
 * @kernel: add to kernel table
 *
 * Adds a new route entry to both internal and kernel routing tables.
 **/
int rt_add(struct rt_entry *entry, int kernel)
{
	struct in6_addr key;
	int ret = 0;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (entry == NULL)
	{
		return -EINVAL;
	}

	ntoh_in6_addr(&entry->destination, &key);

	if ((ret = node_add(rtbl, &key, entry->plen, 
		     entry, sizeof(struct rt_entry))) < 0) 
	{
		if(xdbg) 
			fprintf(stdout, 
				"node_add() returned error for entry %x:%x:%x:%x:%x:%x:%x:%x/%d\n",
				NIP6ADDR(&entry->destination), entry->plen);
		rtnl_rt_del(&entry->destination, entry->plen,
			    &entry->nexthop, entry->interface);
		return -1;
	}
	
	if(xdbg) {
		fprintf(stdout, "added entry in AODV_RT for %x:%x:%x:%x:%x:%x:%x:%x/%d\n",
			NIP6ADDR(&entry->destination), entry->plen);
		//rt_entry_print(stdout,entry);
	}

	ulog(L_INFO, "added route to %x:%x:%x:%x:%x:%x:%x:%x\n",
	     NIP6ADDR(&entry->destination));

	if (kernel) 
	{
		if(xdbg) fprintf(stdout,"***** kernel=true =>trying to add host rt_entry: \n");
		
		if (rtnl_rt_host_add(&entry->destination,
				     &entry->nexthop, 
				     entry->interface,
				     entry->lifetime - timems()) < 0)
		{
			fprintf(stdout,"ERROR: rtnl_rt_host_add returned < 0\n");
			ret = -1;
		}
	}

	/* rtnl_rt_host_add sets the alarm in ioctl 
	 * RFC keeps expired entries in AODV but maybe, 
	 * in case rtnl_rt_host_add is not succesful, 
	 * we should clean AODV table
	 * Something like next: */
	/*
	if(!ret)
	{
		add_alarm(entry->lifetime - timems(), 
					in6_hash(&entry->destination), 
					ROUTE_EXPIRE, rt_expire, data);
	}
	*/

	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	
	return ret;
}

/**
 * rt_update - update existing route entry
 * @entry: entry to update
 * @kernel: update kernel table
 * @create: create new if not found
 *
 * Updates existing route entry in internal routing table. If entry
 * does not exist create new entry if @create is set.
 **/
int rt_update(struct rt_entry *entry, int kernel, int create)
{
	struct in6_addr key;
	int plen, ret=0;
	char op[] = "Update";
				
	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (entry == NULL) return -EINVAL;

	if (entry->plen == 0)
		plen = 128;
	else
		plen = entry->plen;
	
	if(plen != entry->plen)
		fprintf(stdout,"Prefix len was %d in rt_entry, will update with %d\n", entry->plen, plen);
	
	if(xdbg)
	{
		fprintf(stdout, "Updating AODV RT for %x:%x:%x:%x:%x:%x:%x:%x/%d\n",
			NIP6ADDR(&entry->destination), plen);
		//rt_entry_print(stdout,entry);
	}

	ntoh_in6_addr(&entry->destination, &key);
	//node_modify() returns 0 on success
	if (node_modify(rtbl, &key, plen, entry, sizeof(struct rt_entry)) != 0) 
	{
		if (create) 
		{
			/* //see bellow
			struct in6_ifreq *data;
			u32 timerid = in6_hash(&entry->destination);
			*/

			if (rt_add(entry, kernel) < 0)
				return -EINVAL;

		/*	//rt_add calls now add_alarm
		 	data = malloc(sizeof(struct in6_ifreq));
			if (data == NULL) return -ENOMEM;

			memset(data, 0, sizeof(struct in6_ifreq));
			memcpy(&data->ifr6_addr, &entry->destination, 
			       sizeof(struct in6_addr));
			data->ifr6_prefixlen = entry->plen;
			add_alarm(entry->lifetime - timems(), timerid, ROUTE_EXPIRE,
				  rt_expire, data);
		*/
			return 0;
		} else
			return -ENOENT;
	}
	
	/* Updates or creates a new kernel routing 
	 * table entry for destination 
	 *
	 * If RT does not exist, create one 
	 */
	if(rtnl_rt_exist(&entry->destination) != 0) /* returns 0 on success ! */
		memcpy(op,"Create", sizeof(op));
	
	if(xdbg) fprintf(stdout, "Next: %s Kernel RT for %x:%x:%x:%x:%x:%x:%x:%x/%d with lifetime %d\n", op,
			NIP6ADDR(&entry->destination), plen, (int)(entry->lifetime - timems()));
	
	ret = rtnl_rt_update(&entry->destination, 
						&entry->nexthop, 
						entry->interface,
		    			entry->lifetime - timems());
		
	if(xdbg)
	{
	/* -this info is printed in rtioctl_rt_add()
		if(ret==0) 
			fprintf(stdout, 
				"Success - %s kernel RT to %x:%x:%x:%x:%x:%x:%x:%x\n", op, 
	      		NIP6ADDR(&entry->destination));
	    else 	
			fprintf(stderr, 
				"ERROR - Could not %s kernel RT to %x:%x:%x:%x:%x:%x:%x:%x \
				\n\treturned error=%d\n", op, 
	      		NIP6ADDR(&entry->destination), ret);
	*/
	}
	
	/* Add alarm to clean AODV table */
	{
		struct in6_ifreq *data;
		u32 timerid = in6_hash(&entry->destination);
			
	 	data = malloc(sizeof(struct in6_ifreq));
		if (data != NULL)
		{
			memset(data, 0, sizeof(struct in6_ifreq));
			memcpy(&data->ifr6_addr, &entry->destination, 
			       sizeof(struct in6_addr));
			data->ifr6_prefixlen = entry->plen;
			add_alarm(entry->lifetime - timems(), timerid, ROUTE_EXPIRE,
				  rt_expire, data);
		}
	}

	ulog(L_INFO | L_RTMOD, "modified route to %x:%x:%x:%x:%x:%x:%x:%x\n", 
	       NIP6ADDR(&entry->destination));

	return ret;
}

int rt_expired(struct in6_addr *dest, int plen)
{
	struct rt_entry *e;
	struct in6_addr key;

	ntoh_in6_addr(dest, &key);

	e = (struct rt_entry *)node_lookup(rtbl, &key, plen, 0);
	if (e == NULL)
		return -1;
	e->last_hopcount = e->hopcount;
	e->hopcount = INFINITE_HOPS;
	node_modify(rtbl, &key, plen, e, sizeof(struct rt_entry));

	return 0;
}

/**
 * rt_del - delete route entry
 * @dest: destination to delete
 * @plen: prefix length
 *
 * Deletes route entry from internal and kernel routing table.
 **/
int rt_del(struct in6_addr *dest, int plen)
{
	struct rt_entry *e;
	struct in6_addr key;
	int len, ret=0;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
		
	if (dest == NULL) return -EINVAL;
	if (plen < 0 || plen > 128) return -EINVAL;

	if (plen == 0)
		len = 128;
	else
		len = plen;

	ntoh_in6_addr(dest, &key);

	e = (struct rt_entry *)node_lookup(rtbl, &key, len, 0);
	if (e == NULL)
	{
		return -1;
	}

	/* Case: dest an nexthop are identical 
	 * - this is NEVER true as long as aodv0 and 
	 *   eth0 have different addresses  
	 */
	if ((memcmp(&e->nexthop, dest, sizeof(struct in6_addr)) == 0)
	    || (e->nexthop.s6_addr32[0] == 0 &&
		e->nexthop.s6_addr32[1] == 0 &&
		e->nexthop.s6_addr32[2] == 0 &&
		e->nexthop.s6_addr32[3] == 0))
		return rtnl_rt_del(dest, len, NULL, e->interface);

	ret = rtnl_rt_del(dest, len, &e->nexthop, e->interface);

	if (ret == 0)
		ret = rt_expired(&e->destination, plen);
	
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);

	return ret;
	/*rt_entry_free(e);*/
}

/**
 * rt_remove - remove invalid routes
 * @entry: entry to be removed
 * @data: 
 *
 **/
static int rt_remove(void *entry, void *data)
{
	struct rt_entry *e = (struct rt_entry *)entry;
	struct in6_addr *nh = NULL;
	int plen;

	if (entry == NULL) return -EINVAL;

	if (e->plen == 0)
		plen = 128;
	else
		plen = e->plen;

	nh = (struct in6_addr *)&(e->nexthop);

	/* This does not make sense! Of course nh == e->nexthop
	 if ((memcmp(nh, &e->destination, sizeof(struct in6_addr)) == 0)
	    || (nh->s6_addr32[0] == 0 && nh->s6_addr32[1] == 0 &&
		nh->s6_addr32[2] == 0 && nh->s6_addr32[3] == 0))
		nh = NULL;
	*/
	rtnl_rt_del(&e->destination, plen, nh, e->interface);
	return 0;
}

/**
 * rt_flush - flush routing table
 *
 * Flushes all entries currently in the routing table.  Used for
 * runtime restarts.
 **/
int rt_flush(void)
{
	tree_walker(rtbl, rt_remove, NULL);
	return 0;
}

/**
 * find_invalid - invalidate entry with broken next hop
 * @entry: routing table entry
 * @data: broken next hop address
 *
 * This is a tree walker callback.  It invalidates routing table
 * entries that use address given in @data as their next hop.
 **/
static int find_invalid(void *entry, void *data)
{
	struct in6_addr *nexthop = (struct in6_addr *)data;
	struct rt_entry *e = (struct rt_entry *)entry;

	if (entry == NULL) return -EINVAL;
	if (data == NULL) return -EINVAL;

	if (memcmp(nexthop, &e->nexthop, sizeof(struct in6_addr) == 0)) 
	{
		struct precursor *p, *next;

		/* 
		 * Remove the entry from Kernel's RT
		 */
		rtnl_rt_del(&e->destination, e->plen, &e->nexthop, e->interface);
	
		/* 
		 * Clean the AODV RT
		 */

		/* route with invalidated nexthop */
		p = (struct precursor *)e->precursors;
		while (p) 
		{
			/* notifier(&p->node);*/
			next = p->next;
			free(p);
			p = next;
		}

		e->sequence++;
		e->hopcount = 255;
		memset(&e->nexthop, 0, sizeof(struct in6_addr));

		if(xdbg) fprintf(stdout,"---> %s: invalid nexthop for "
		     "%x:%x:%x:%x:%x:%x:%x:%x\n",__FUNCTION__,
		     NIP6ADDR(&e->destination));
	}

	return 0;
}

/**
 * rt_invalid_nexthop - mark routes with this next hop as invalid
 * @addr: next hop address
 *
 * Walks through the routing table and marks all routes using @addr as
 * next hop invalid.
 **/
int rt_invalid_nexthop(struct in6_addr *addr)
{
	if (addr == NULL) return -EINVAL;

	tree_walker(rtbl, find_invalid, addr);
	return 0;
}

/**
 * rt_get - get route entry
 * @dest: route destination to get
 * @plen: prefix length
 * @type: route entry type
 *
 * Finds route entry and returns a copy.  Valid @type values are
 * %RT_ACTIVE, %RT_KERNEL and %RT_ANY.
 **/
struct rt_entry *rt_get(struct in6_addr *dest, int plen, int type)
{
	struct rtnl_request req;
	struct rt_entry *route = NULL, *rte = NULL;
	struct in6_addr key;
	int wild;
	int _printIt = 0;
	/* print debug statements only for foreign addresses */
	if(xdbg) _printIt = ipv6_addr_cmp(dest, &Config.NodeAddrGlobal);

	if (dest == NULL) return NULL;
	if (plen < 0 || plen > 128) return NULL;

	//MI: This wildcard=1 is not working!!!
	//A search for 2003::12 will return 2003::13 as a match !
	//NOTE: Not working with wildcard=0 either....
	//wild = plen < 128 ? 0 : 1;  //<-original version
	wild = plen <= 128 ? 0 : 1;
	
	memset(&key,0,sizeof(struct in6_addr));
	ntoh_in6_addr(dest, &key);

	if(_printIt)
	{
		if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
    	if (xdbg) fprintf(stdout,
						"\tLooking in RT for addr: %x:%x:%x:%x:%x:%x:%x:%x/%d wild=%d\n",
						NIP6ADDR(dest), plen, wild);
		//fprintf(stdout,"\twith key:  %x:%x:%x:%x:%x:%x:%x:%x\n",NIP6ADDR(&key));
		//fprintf(stdout, "Current RT is: \n");
		//rt_dump();
	}
	
	switch (type) 
	{
	case RT_ACTIVE:
		/* active routes have finite hopcount */
		if(_printIt & xdbg) fprintf(stdout, "RT_TYPE = RT_ACTIVE\n");
		if ((route = (struct rt_entry *)
		     node_lookup(rtbl, &key, 128, wild)) == NULL)
		{
			if(_printIt & xdbg) 
				fprintf(stdout, "No route returned by node_lookup().\n");
			if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
			return NULL;
		}
		
		/* Expiration checked bellow */
		if (route->hopcount == INFINITE_HOPS)
		{
			if(_printIt & xdbg) 
				fprintf(stdout, "Route returned by node_lookup() is invalid.\n");
			if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
			return NULL;
		}
		break;
		
	case RT_ANY:
		/* return route even if invalidated */
		if(_printIt & xdbg) fprintf(stdout, "RT_TYPE = RT_ANY\n");
		if ((route = (struct rt_entry *)
		     node_lookup(rtbl, &key, 128, wild)) == NULL)
		{
			if(_printIt & xdbg) 
			{
				fprintf(stdout,"No route found by node_lookup().\n");
				if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
			}
			return NULL;
		}
		break;
	case RT_KERNEL:
		if(_printIt & xdbg) fprintf(stdout, "RT_TYPE = RT_KERNEL\n");
		if(rtnl_rt_lookup(&req, dest, plen) != 0)
			return NULL;
		break;
	default:
		break;
	}

	if (route == NULL)
	{
		if(_printIt & xdbg) fprintf(stdout,"node_lookup returned route=NULL, return NULL\n");
		return NULL;
	}

	/* Check if the route is expired and invalidate it if so */
	if((route->lifetime - timems()) <= 0)
	{
		route->last_hopcount = route->hopcount;
		route->hopcount = 255;
		node_modify(rtbl, &key, plen, route, sizeof(struct rt_entry));
		return NULL;
	}

	rte = malloc(sizeof(struct rt_entry));
	memset(rte, 0, sizeof(*rte));
	memcpy(rte, route, sizeof(struct rt_entry));
	route = NULL;

	if(_printIt)
	{
		if(xdbg) 
		{
			fprintf(stdout,"Returning AODV rt_entry : \n");
			rt_entry_print(stdout,rte);
		}
		if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	}

	return rte;
}

/**
 * rtnl_rt_exist - check if route for an address exists
 * @dest: destination to check
 *
 * Queries the kernel routing table, whether a route exists for this
 * destination @dest.  Returns 0 on success, positive non-zero if not
 * found, and negative on error.
 **/
int rtnl_rt_exist(struct in6_addr *dest)
{
	struct rtnl_request req;
	struct rtattr * tb[RTA_MAX+1], *rta;
	char buffer[IF_NAMESIZE];
	int len;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	if (dest == NULL) return -EINVAL;

	memset(&req, 0, sizeof(struct rtnl_request));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = RTM_GETROUTE;
	req.r.rtm_family = AF_INET6;
	req.r.rtm_dst_len = 128;

	rtmsg_add_rta(&req.n, sizeof(req), RTA_DST, dest, 16);

	if (rtmsg_query(&rth, &req.n, 0, 0, &req.n) < 0) {
		return -2;
	}

	len = req.n.nlmsg_len - NLMSG_LENGTH(sizeof(req.r));
	if (len < 0)
		return -1;
	memset(tb, 0, sizeof(tb));

	rta = RTM_RTA(&req.r);
	while (RTA_OK(rta, len)) 
	{
		if (rta->rta_type <= RTA_MAX)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
	if (len) return -1;

	if (*(int *)RTA_DATA(tb[RTA_OIF]) == 1)
		return 1;	//route not found

	if (strcmp("aodv0", if_indextoname(*(int *)RTA_DATA(tb[RTA_OIF]), buffer)) == 0)
	{
		return 1; //route not found
	}

	return 0;
}

/**
 * rt_get_nexthop - get nexthop for a destination
 * @dest: route destination
 * @nexthop: pointer where to store the result
 *
 * Looks up next hop for @destination.  If destination has no valid
 * route, return negative.
 **/
int rt_get_nexthop(struct in6_addr *dest, struct in6_addr *nexthop)
{
	struct rt_entry *rte;
	struct in6_addr key;

	if(xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);

	if (dest == NULL || nexthop == NULL) return -EINVAL;

	if(xdbg) fprintf(stdout,"For destination %x:%x:%x:%x:%x:%x:%x:%x\n", NIP6ADDR(dest));

	//if ((rte = (struct rt_entry *)node_lookup(rtbl, dest, 128, 0)) == NULL)
	ntoh_in6_addr(dest, &key);
	if ((rte = (struct rt_entry *)node_lookup(rtbl, &key, 128, 0)) == NULL)
	{
		if (xdbg) 
		{
			fprintf(stdout,"\troute not found...\n");
			//rt_entry_print(stdout, rt_get(dest,128,RT_ANY));
		}
		return -1;
	}
	
	if(&rte->nexthop) {
		if(xdbg) fprintf(stdout,"\tnexthope is: %x:%x:%x:%x:%x:%x:%x:%x [hcount=%d]\n",
				NIP6ADDR(&rte->nexthop), rte->hopcount);
	}
	
	if (rte->hopcount == INFINITE_HOPS)
	{
		fprintf(stdout, "\tRoute invalidated (infinite_hops)\n");
		return -1;
	}

	memset(nexthop,0,sizeof(struct in6_addr));
	memcpy(nexthop, &rte->nexthop, sizeof(struct in6_addr));
	return 0;
}

/**
 * rt_entry_precursor_search - searches precursor list for addr
 * @entry: entry where to search
 * @addr: precursor address
 * Returns 1 if addr not found among precursors and 0 otherwise.
 */
int rt_entry_precursor_search(struct rt_entry *entry, struct in6_addr *addr)
{
	struct rt_entry *rte = NULL;
	struct in6_addr key;
	struct precursor *pre = NULL;
	struct in6_addr *pip = NULL;
	
	if (xtr) fprintf(stdout,"=== %s ===\n",__FUNCTION__);
	
	if (entry == NULL) return -EINVAL;
	if (addr == NULL) return -EINVAL;
	
	ntoh_in6_addr(&entry->destination, &key);
	rte = (struct rt_entry *)node_lookup(rtbl, &key, 128, 0);
	
	if (rte == NULL) return -ENODATA;
	
	//rt_entry_print(stdout,rte);
	
	if(rte->precursors != NULL)
		pre = (struct precursor *)rte->precursors;
	else
		return 1; //No precursor exists
	
	while (pre != NULL) 
	{
		if((pip = &pre->node) != NULL)
		{
			if(ipv6_addr_cmp(pip,addr) == 0)
			{
				if (xdbg) fprintf(stdout,
						"Precursor %x:%x:%x:%x:%x:%x:%x:%x exists\n",
						NIP6ADDR(addr));
				if (xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
				return 0;
			}
			else
				pre = pre->next;
				
		}	
	}

	if (xdbg) fprintf(stdout,"Precursor %x:%x:%x:%x:%x:%x:%x:%x does not exists.\n",
						NIP6ADDR(addr));
	if (xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);

	return 1; /* address is not among precursors */
}

/**
 * rt_entry_precursor_add - add precursor to route entry
 * @entry: entry where to add
 * @addr: precursor address
 *
 * Prepends new precursor @addr to a routing entry.  When a RREP is
 * sent or forwarded recipient is added to the precursor list for that
 * As per RFC 3561 Section 2, route entry:
 * [Route-destination] via [sender of RREP] precursor [Route originator]
 * "If the RREP has a nonzero prefix length, then the originator of the RREQ 
 * which solicitated the RREP information is included among the precursors..."
 **/
int rt_entry_precursor_add(struct rt_entry *entry, struct in6_addr *addr)
{
	struct precursor *prec = malloc(sizeof(struct precursor));
	struct rt_entry *rte = NULL;
	struct in6_addr key;

	if (entry == NULL || addr == NULL) return -EINVAL;

	if (prec == NULL) return -ENOMEM;

	ntoh_in6_addr(&entry->destination, &key);
	rte = (struct rt_entry *)node_lookup(rtbl, &key, 128, 0);
	if (rte == NULL) return -ENOENT;
	
	/* Must check if address is already inserted... */
	if(rt_entry_precursor_search(rte,addr))
	{
		memcpy(&prec->node, addr, sizeof(struct in6_addr));
		prec->next = rte->precursors;
		rte->precursors = prec;
	}
	return 0;
}

/**
 * rt_entry_precursor_destroy - remove all precursors
 * @entry: routing entry
 *
 * Destroys whole precursor list for @entry.
 **/
int rt_entry_precursor_destroy(struct rt_entry *entry)
{
	struct rt_entry *rte = NULL;
	struct in6_addr key;
	struct precursor *pre, *next;

	if (entry == NULL) return -EINVAL;

	ntoh_in6_addr(&entry->destination, &key);
	rte = (struct rt_entry *)node_lookup(rtbl, &key, 128, 0);
	if (rte == NULL) return -1;

	pre = (struct precursor *)rte->precursors;
	while (pre) {
		next = pre->next;
		free(pre);
		pre = next;
	}

	return 0;
}

/**
 * rt_entry_free - free routing table entry
 * @entry: entry to free
 *
 * Frees all memory held by routing table entry @entry, including
 * precursor list.
 **/
void rt_entry_free(struct rt_entry *entry)
{
	struct precursor *pre, *next;

	if (entry == NULL) return;

	pre = (struct precursor *)entry->precursors;
	while (pre) {
		next = pre->next;
		free(pre);
		pre = next;
	}
	free(entry);
	entry = NULL;
}

/**
 * rt_entry_print - print routing table entry
 * @out: output stream
 * @e: entry to print
 *
 * Prints out routing table entry.  For debugging use.
 **/
void rt_entry_print(FILE *out, struct rt_entry *e)
{
	struct precursor *p;
	char buf[INET6_ADDRSTRLEN];

	if (out == NULL)
		out = stdout;

	if (e == NULL)
	{
		fprintf(out,"WARNNING: NULL rt_entry *e\n");
		return;
	}

/*
struct rt_entry 
{
	struct in6_addr destination;  	// IPv6 address for destination    *
	u8 plen;		       			// Prefix length (hostroute = 128) *
	int interface;		       		// Interface identifier            *
	u32 sequence;		       		// Destination sequence number     *
	u8 hopcount;		       		// Number of hops to destination   *
	u8 last_hopcount;	       		// Hop count from last RREP        *
	struct in6_addr nexthop;       	// IPv6 address for the next hop   *
	struct precursor *precursors;  	// List of nodes we sent RREPs     *
	u32 lifetime;		       		// Valid lifetime for this route   *
	u32 flags;		       			// Routing flags                   *
	u32 hello_interval;	       		// Interval to wait for Hellos     *
	//struct rt_entry *next; 
};
*/
	fprintf(out, "  Destination    : %s\n",
		inet_ntop(AF_INET6, &e->destination, buf, sizeof(buf)));
	fprintf(out, "  Prefix length  : %d\n", e->plen);
	fprintf(out, "  Dest seq#      : %d\n", e->sequence);
	fprintf(out, "  Hop count      : %d\n", e->hopcount);
	fprintf(out, "  Last hop count : %d\n", e->last_hopcount);
	fprintf(out, "  Next hop       : %s\n",
		inet_ntop(AF_INET6, &e->nexthop, buf, sizeof(buf)));
	if(e->lifetime<=timems())
		fprintf(out, "  Lifetime       : expired (%d)\n", (int)(e->lifetime - timems()));
	else
		fprintf(out, "  Lifetime       : %lu\n", e->lifetime - timems());
	fprintf(out, "  Flags          : %d\n", e->flags);
	fprintf(out, "  Hello interval : %d\n", e->hello_interval);
	fprintf(out, "  Precursors     :\n");
	if (e->precursors == NULL) 
	{
		fprintf(out, "\t(none)\n");
		return;
	}
	for (p = e->precursors; p; p = p->next) 
	{
		fprintf(out, "\t%s\n",
			inet_ntop(AF_INET6, &p->node, buf, sizeof(buf)));
	}
}

/**
 * rt_entry_oneline - print routing table entry to a stream
 * @data: routing table entry
 * @stream: output stream
 *
 * Prints a routing table entry @data to an output stream @stream
 * (e.g. stdout).  Useful for debugging.
 **/
int rt_entry_oneline(void *data, void *stream)
{
	FILE *out = (FILE *)stream;
	struct rt_entry *e = (struct rt_entry *)data;
	char buf[INET6_ADDRSTRLEN];

	if (data == NULL || stream == NULL) return -EINVAL;

	//fprintf(out, "%-36s/%3d %03d(%03d) %-36s %05lu\n",
	fprintf(out, "%s/%-3d hopcount:%03d(last_hopcount:%03d) nexthop:%s lifetime:%lu\n",
		inet_ntop(AF_INET6, &e->destination, buf, sizeof(buf)),
		e->plen, e->hopcount, e->last_hopcount,
		inet_ntop(AF_INET6, &e->nexthop, buf, sizeof(buf)),
		e->lifetime - timems());

	return 0;
}

/**
 * rt_dump - dump all routing table entries to STDOUT
 *
 * Prints whole routing table on standard output, as one line entries.
 * Useful only for debugging.
 **/
void rt_dump(void)
{
	tree_walker(rtbl, rt_entry_oneline, stdout);
}
