/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#ifndef __RT_TABLE_H__
#define __RT_TABLE_H__ 1

#include <netinet/in.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include "util.h"

struct rtnl_handle
{
	int			fd;
	struct sockaddr_nl	local;
	struct sockaddr_nl	peer;
	__u32			seq;
	__u32			dump;
};

#define RT_ACTIVE	0x1
#define RT_ANY		0x2
#define RT_KERNEL	0x4

/*
 * linux/rtnetlink.h defines some routing protocols for use with
 * struct rtmsg rtm_protocol.  Kernel does not use these for anything,
 * but they are stored in the routing table for userspace programs.
 */
#define RTPROT_MANET	20
#define RTPROT_AODV	RTPROT_MANET + 1

#define INFINITE_HOPS	255

/*
 * Neighbouring node for which we've generated RREPs.  Should be
 * notified when route becomes invalid.
 */
struct precursor {
	struct in6_addr node;
	struct precursor *next;
};

typedef void (*rt_notifier_t)(struct in6_addr *);

/*
 * Entry in the AODV routing table.
 */
struct rt_entry {
	struct in6_addr destination;   /* IPv6 address for destination    */
	u8 plen;		       /* Prefix length (hostroute = 128) */
	int interface;		       /* Interface identifier            */
	u32 sequence;		       /* Destination sequence number     */
	u8 hopcount;		       /* Number of hops to destination   */
	u8 last_hopcount;	       /* Hop count from last RREP        */
	struct in6_addr nexthop;       /* IPv6 address for the next hop   */
	struct precursor *precursors;  /* List of nodes we sent RREPs     */
	u32 lifetime;		       /* Valid lifetime for this route   */
	u32 flags;		       /* Routing flags                   */
	u32 hello_interval;	       /* Interval to wait for Hellos     */
/*	struct rt_entry *next; */
};

#define RT_F_LOCAL_REPAIR 0x00010000
#define RT_F_DEST_LOCAL   0x01000000

/*
 * Address list item
 */
struct addr_list {
	struct addr_list *next;
	int ifindex;
	struct in6_addr addr;
	int prefixlen;
};

/*
 * Initialize routing
 */
int rt_init(void);

int rtnl_rt_sync_listener(struct sockaddr_nl *nlh, struct nlmsghdr *n);

/*
 * Add local address to routing table (kernel rt not touched).
 */
int rt_set_local(struct addr_list *addr);

int rt_flush(void);

/*
 * Return list of addresses for this interface
 */
struct addr_list *rtnl_local_addrs(void);

int rtnl_rt_update(struct in6_addr *dest, struct in6_addr *nh,
		   int oif, unsigned long lifetime);
int rtnl_rt_exist(struct in6_addr *dest);

int rtnl_rt_host_add(struct in6_addr *destination, 
		     struct in6_addr *nexthop,
		     int ifindex, unsigned long lifetime);
int rtnl_rt_del(struct in6_addr *destination, int plen, 
		struct in6_addr *nexthop, int ifindex);

/*
 * Get latest sequence number for this node.
 */
void set_seqnum(u_int32_t n);
unsigned int get_seqnum(void);
unsigned int inc_get_seqnum(void);
unsigned int upd_get_seqnum(u_int32_t n);

/*
 * Determine if route local.
 */
int rt_is_local(struct rt_entry *entry);

/*
 * Determine if addr local address.
 */
int rt_is_local_addr(struct in6_addr *addr);

/*
 * Invalidate routes with this next hop.
 */
int rt_invalid_nexthop(struct in6_addr *addr);

/*
 * Add routing table entry.
 */
int rt_add(struct rt_entry *entry, int kernel);

/*
 * Update existing entry. If entry does not exist create new entry if
 * 'create' is set. 
 */
int rt_update(struct rt_entry *entry, int kernel, int create);

/*
 * Delete routing table entry.
 * Delete from Kernel RT and invalidate/expire
 * the AODV entry by setting hopcount to infinity 
 */
int rt_del(struct in6_addr *dest, int plen);

/*
 * Check if routing table entries from AODV RT.
 */
void rt_expire(void *args);

/*
 * Find routing table entry.  Returns a copy.
 */
struct rt_entry *rt_get(struct in6_addr *dest, int plen, int type);

/*
 * Get next hop for destination with valid route.
 */
int rt_get_nexthop(struct in6_addr *dest, struct in6_addr *nexthop);

/*
 * Free routing table entry
 */
void rt_entry_free(struct rt_entry *entry);

int rt_entry_precursor_add(struct rt_entry *entry, struct in6_addr *prec);

int rt_entry_precursor_destroy(struct rt_entry *entry);

int rt_entries_precursor_delete(struct in6_addr *addr);

void rt_entry_print(FILE *out, struct rt_entry *e);

#endif /* __RT_TABLE_H__ */
