/*
 *	HUT AODV for IPv6
 *	Timer queue
 *
 *	Author: Antti Tuominen <ajtuomin@tml.hut.fi>
 *
 *	$Id: tqueue.c,v 1.2 2004/05/06 04:08:02 michaela Exp $
 *
 *	Copyright (c) Antti J. Tuominen and Helsinki University of
 *	Technology.  See COPYING for license information.
 */

#define _GNU_SOURCE 1

#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include "tqueue.h"

#define timerafter(a, b) timercmp(a,b,<)
#define timerbefore(a, b) timercmp(a,b,>)

#define DEBUG DEBUG_AODV

extern int xtr;

struct tq_elem {
	struct tq_elem *next;
	struct tq_elem *prev;
	unsigned long key;		/* identifier for task  */
	unsigned char type;		/* type of task         */
	struct timeval expires;		/* expire time for task */
	pid_t pid;			/* who queued this task */
	int (*task)(void *args);	/* pointer to task      */
	void *data;			/* data for task        */
};

static struct tq_elem tqueue;
static pthread_spinlock_t spinlock;

static void expire(int signal);

/**
 * taskqueue_init - initialize task queue
 *
 * Initializes task queue.
 **/
int taskqueue_init(void)
{
	int ret = 0;
	ret = pthread_spin_init(&spinlock, 0);
	memset(&tqueue, 0, sizeof(tqueue));
	tqueue.next = tqueue.prev = &tqueue;
	tqueue.expires.tv_sec = 
		tqueue.expires.tv_usec = (~0);

	return ret;
}

/**
 * taskqueue_destroy - destroy task queue
 *
 * Destroys task queue and frees all entries.
 **/
void taskqueue_destroy(void)
{
	struct tq_elem *p, *t;

	signal(SIGALRM, SIG_IGN);
	pthread_spin_destroy(&spinlock);
	for (p = tqueue.next; p != &tqueue; p = t) {
		t = p->next;
		if (p->data)
			free(p->data);
		free(p);
	}
}

/**
 * sorted_insert - insert queue item sorted by expiration time
 * @tqi: task queue entry to insert
 *
 * Inserts @tqi to task queue sorted by expiration time.
 **/
static inline int sorted_insert(struct tq_elem *tqi)
{
	struct tq_elem *cur;

	for (cur = tqueue.next; cur != &tqueue; cur = cur->next) {
		if (timerbefore(&cur->expires, &tqi->expires))
			break;
	}
	cur->prev->next = tqi;
	tqi->prev = cur->prev;
	cur->prev = tqi;
	tqi->next = cur;

	return tqi->prev == &tqueue ? 0 : 1;
}

/**
 * update_alarm - update alarm to next expiration time
 *
 * Sets alarm to next expiration time in task queue.
 **/
static int update_alarm(void)
{
	struct tq_elem *tqi;
	struct timeval now;
	struct itimerval exp;

	tqi = tqueue.next;

	if (tqi == &tqueue) {
		return 0; /* queue empty */
	}

	gettimeofday(&now, NULL);
	memset(&exp, 0, sizeof(exp));
	timersub(&tqi->expires, &now, &exp.it_value);

	signal(SIGALRM, expire);
	if (timerafter(&now, &tqi->expires)) {
		setitimer(ITIMER_REAL, &exp, NULL); /* set new timer */
	} else {
		kill(tqi->pid, SIGALRM); /* already expired */
	}

	return 0;
}

/**
 * clr_alarm - clear alarms for a destination
 * @id: alarm identifier
 * @type: task type
 *
 * Clears all alarms for @dest.
 **/
int clr_alarm(unsigned long id, unsigned char type)
{
	struct tq_elem *cur;
	sigset_t bsigset, oldsigset;

	pthread_spin_lock(&spinlock);
	if (tqueue.next == &tqueue) {
		pthread_spin_unlock(&spinlock);
		return 0;
	}

	sigemptyset(&bsigset);
	sigaddset(&bsigset, SIGALRM);
	sigprocmask(SIG_BLOCK, &bsigset, &oldsigset);

	for (cur = tqueue.next; cur != &tqueue; cur = cur->next) {
		if (cur->key == id && cur->type == type) {
			cur->prev->next = cur->next;
			cur->next->prev = cur->prev;
			cur->next = cur->prev = NULL;
			if (cur->data)
				free(cur->data);
			free(cur);
			break;
		}
	}

	update_alarm();
	pthread_spin_unlock(&spinlock);
	sigprocmask(SIG_SETMASK, &oldsigset, NULL);

	return 0;
}

/**
 * add_alarm - add new task with alarm to task queue
 * @ms: expires in ms milliseconds
 * @key: search key
 * @type: task type
 * @task: task to execute on expiry
 * @data: data for task
 *
 * Adds @task to task queue with search @key.  Task will expire in @ms
 * milliseconds.  Task @data is stored with the entry.
 **/
int add_alarm(unsigned long ms, unsigned long key, unsigned char type,
	      void (*task)(void *args), void *data)
{
	struct tq_elem *tqi;
	struct timeval now;
	sigset_t bsigset, oldsigset;

	if((type != ROUTE_DISC)||(type != ROUTE_EXPIRE))
		xtr=0;

	if(xtr) fprintf(stdout,"=== %s === task type: %d\n", __FUNCTION__,type);

	tqi = (struct tq_elem *)malloc(sizeof(struct tq_elem));
	if (tqi == NULL) {
		fprintf(stderr,"ERROR in %s(): Can't malloc\n",__FUNCTION__);
		return -1;
	}
	clr_alarm(key, type);

	sigemptyset(&bsigset);
	sigaddset(&bsigset, SIGALRM);
	sigprocmask(SIG_BLOCK, &bsigset, &oldsigset);

	tqi->task = (void *)task;
	tqi->data = data;
	tqi->key = key;
	tqi->type = type;
	tqi->pid = getpid();
	gettimeofday(&now, NULL);
	tqi->expires.tv_sec = now.tv_sec + (ms / 1000);
	tqi->expires.tv_usec = now.tv_usec + ((ms % 1000) * 1000);

	pthread_spin_lock(&spinlock);
	sorted_insert(tqi);
	update_alarm();
	pthread_spin_unlock(&spinlock);

	sigprocmask(SIG_SETMASK, &oldsigset, NULL);
	
	if(xtr) fprintf(stdout,"+++ %s +++\n",__FUNCTION__);
	return 0;
}

/**
 * expire - expire task
 * @signal: UNIX signal
 *
 * Called on %SIGALRM.  Expires next task in queue if expiration time
 * less than current time.
 **/
static void expire(int signal)
{
	struct tq_elem *cur;
	struct timeval now;

	gettimeofday(&now, NULL);
	pthread_spin_lock(&spinlock);
	cur = tqueue.next;

	if (cur == &tqueue) {
		pthread_spin_unlock(&spinlock);
		return;
	}

	cur->prev->next = cur->next;
	cur->next->prev = cur->prev;
	cur->next = cur->prev = NULL;
	pthread_spin_unlock(&spinlock);

	(*cur->task)(cur->data);
	if (cur->data)
		free(cur->data);
	free(cur);

	pthread_spin_lock(&spinlock);
	update_alarm();
	pthread_spin_unlock(&spinlock);
}
