/*
 *	SecAODV IPv6 Project
 *
 *	See NIST-COPYRIGHT and COPYING for license.
 */

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <netinet/ip6.h>

#include "config.h"

/**
 * timems - current time in milliseconds
 *
 * Returns current time of day in milliseconds.
 **/
unsigned long timems(void)
{
	unsigned long ms;
	struct timeval now;

	gettimeofday(&now, NULL);
	ms = now.tv_sec * 1000;
	ms += now.tv_usec / 1000;

	return ms;
}

/**
 * in6_hash32 - 32bit hash from IPv6 address
 * @addr: IPv6 address
 *
 * Calculates a 32bit hash key from an 128bit IPv6 address @addr.
 **/
u_int32_t in6_hash(struct in6_addr *addr)
{
	return addr->s6_addr32[0] ^ addr->s6_addr32[1] ^
		addr->s6_addr32[2] ^ addr->s6_addr32[3];
}

/**
 * print_in6addr - print IPv6 address
 * @addr: address to print
 *
 * Prints an IPv6 address.
 **/
void print_in6addr(FILE *f, const struct in6_addr *addr)
{
	if (addr == NULL) {
		fprintf(stderr, "%s addr == NULL\n", __FUNCTION__);
		return;
	}
	if(f == NULL)
		f = stdout;
	
	fprintf(f,"%x:%x:%x:%x:%x:%x:%x:%x\n",
	       (int)ntohs(addr->s6_addr16[0]), (int)ntohs(addr->s6_addr16[1]),
	       (int)ntohs(addr->s6_addr16[2]), (int)ntohs(addr->s6_addr16[3]),
	       (int)ntohs(addr->s6_addr16[4]), (int)ntohs(addr->s6_addr16[5]),
	       (int)ntohs(addr->s6_addr16[6]), (int)ntohs(addr->s6_addr16[7]));
}


/**
 * ntoh_in6_addr - convert IPv6 address from network to host byte order
 * @addr: address to convert
 * @buf: buffer where to store converted address
 *
 * Converts an IPv6 address from network byte order to host byte
 * order.  
 **/
struct in6_addr *ntoh_in6_addr(struct in6_addr *addr, struct in6_addr *buf)
{
	int i;

	memset(buf, 0, sizeof(struct in6_addr));
	if (addr == NULL) {
		fprintf(stderr, "%s addr == NULL\n", __FUNCTION__);
		return NULL;
	}
	if (buf == NULL) {
		perror(__FUNCTION__);
		return NULL;
	}

	for (i = 0; i < 4; i++)
		buf->s6_addr32[i] = ntohl(addr->s6_addr32[i]);

	return buf;
}

/**
 * hton_in6_addr - convert IPv6 address from host to network byte order
 * @addr: address to convert
 * @buf: buffer where to store converted address
 *
 * Converts an IPv6 address from host byte order to network byte
 * order.  Returns newly allocated network byte order address.  User
 * should free() the memory after use.
 **/
struct in6_addr *hton_in6_addr(struct in6_addr *addr, struct in6_addr *buf)
{
	int i;

	if (addr == NULL) {
		fprintf(stderr, "%s addr == NULL\n",__FUNCTION__);
		return NULL;
	}
	if (buf == NULL) {
		perror(__FUNCTION__);
		return NULL;
	}

	for (i = 0; i < 4; i++)
		buf->s6_addr32[i] = htonl(addr->s6_addr32[i]);

	return buf;
}

/**
 * dprint - print a debug message
 * @level: debug level where this message should be printed
 * @fmt: printf() style format
 *
 * Prints a debug message to stderr.  If current debug level is higher
 * or equal to @level, the message gets printed.  Uses normal printf()
 * style format.
 **/
void dlog(int level, const char* fmt, ...)
{
	char s[1024];
	va_list args;

	if (Config.DebugLevel < level)
		return;

	va_start(args, fmt);
	vsprintf(s, fmt, args);
	fprintf(stderr, "%s", s);
	va_end(args);
}

int ulogif(int level)
{
	return (Config.LogVerbosity & level);
}

extern FILE *ulogd;

void ulog(int level, const char *fmt, ...)
{
	char s[1024];
	va_list args;

	if (!(Config.LogVerbosity & level) || ulogd == NULL)
		return;

	va_start(args, fmt);
	vsprintf(s, fmt, args);
	fprintf(ulogd, "%s", s);
	va_end(args);
}
